Package: alabaster.matrix
Title: Load and Save Artifacts from File
Version: 1.12.0
Date: 2025-03-04
Authors@R: person("Aaron", "Lun", role=c("aut", "cre"), email="infinite.monkeys.with.keyboards@gmail.com")
License: MIT + file LICENSE
Description: Save matrices, arrays and similar objects into file
        artifacts, and load them back into memory. This is a more
        portable alternative to serialization of such objects into RDS
        files. Each artifact is associated with metadata for further
        interpretation; downstream applications can enrich this
        metadata with context-specific properties.
Depends: alabaster.base
Imports: methods, BiocGenerics, S4Vectors, DelayedArray (>= 0.33.3),
        S4Arrays, SparseArray (>= 1.5.22), rhdf5 (>= 2.47.1),
        HDF5Array, Matrix, Rcpp
Suggests: testthat, knitr, BiocStyle, chihaya, BiocSingular,
        ResidualMatrix
LinkingTo: Rcpp
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Encoding: UTF-8
biocViews: DataImport, DataRepresentation
Config/pak/sysreqs: libssl-dev libnode-dev zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 13:00:04 UTC
RemoteUrl: https://github.com/bioc/alabaster.matrix
RemoteRef: RELEASE_3_23
RemoteSha: 66f650beff91c543d36e03d848655e15b0b07bd5
NeedsCompilation: yes
Packaged: 2026-04-29 13:48:19 UTC; root
Author: Aaron Lun [aut, cre]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 13:50:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(.createRawArraySeed)
export(.extractArrayDimnames)
export(AmalgamatedArray)
export(AmalgamatedArraySeed)
export(DelayedMask)
export(ReloadedArray)
export(ReloadedArraySeed)
export(altReloadDelayedObject)
export(altReloadDelayedObjectFunction)
export(altStoreDelayedObject)
export(altStoreDelayedObjectFunction)
export(componentNames)
export(createExternalSeedDedupSession)
export(extractComponents)
export(loadArray)
export(loadWrapperArray)
export(preserveDelayedOperations)
export(readArray)
export(readDelayedArray)
export(readSparseMatrix)
export(recycleHdf5Files)
export(registerReloadDelayedObjectFunction)
export(reloadDelayedObject)
export(reloadDelayedObjectFunctionRegistry)
export(storeDelayedObject)
export(writeSparseMatrix)
exportClasses(AmalgamatedArray)
exportClasses(AmalgamatedArraySeed)
exportClasses(AmalgamatedMatrix)
exportClasses(DelayedMask)
exportClasses(ReloadedArray)
exportClasses(ReloadedArraySeed)
exportClasses(ReloadedMatrix)
exportClasses(WrapperArray)
exportClasses(WrapperArraySeed)
exportMethods(DelayedArray)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(extract_array)
exportMethods(extract_sparse_array)
exportMethods(is_sparse)
exportMethods(matrixClass)
exportMethods(path)
exportMethods(saveObject)
exportMethods(stageObject)
exportMethods(storeDelayedObject)
import(DelayedArray)
import(alabaster.base)
import(methods)
import(rhdf5)
importClassesFrom(DelayedArray,DelayedAbind)
importClassesFrom(DelayedArray,DelayedArray)
importClassesFrom(DelayedArray,DelayedMatrix)
importClassesFrom(DelayedArray,DelayedSetDimnames)
importClassesFrom(DelayedArray,DelayedUnaryOp)
importClassesFrom(Matrix,Matrix)
importClassesFrom(Matrix,dgCMatrix)
importClassesFrom(Matrix,dsparseMatrix)
importClassesFrom(SparseArray,SVT_SparseMatrix)
importFrom(BiocGenerics,"type<-")
importFrom(BiocGenerics,path)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,type)
importFrom(DelayedArray,DelayedArray)
importFrom(DelayedArray,RegularArrayGrid)
importFrom(DelayedArray,blockApply)
importFrom(DelayedArray,colAutoGrid)
importFrom(DelayedArray,currentViewport)
importFrom(DelayedArray,getAutoBlockLength)
importFrom(DelayedArray,getAutoBlockSize)
importFrom(DelayedArray,isPristine)
importFrom(DelayedArray,matrixClass)
importFrom(DelayedArray,new_DelayedArray)
importFrom(DelayedArray,rowAutoGrid)
importFrom(DelayedArray,seed)
importFrom(HDF5Array,H5SparseMatrixSeed)
importFrom(HDF5Array,HDF5Array)
importFrom(HDF5Array,HDF5ArraySeed)
importFrom(HDF5Array,getHDF5DumpChunkDim)
importFrom(HDF5Array,getHDF5DumpCompressionLevel)
importFrom(HDF5Array,writeHDF5Array)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(S4Arrays,acbind)
importFrom(S4Arrays,arbind)
importFrom(S4Arrays,extract_array)
importFrom(S4Arrays,is_sparse)
importFrom(S4Vectors,new2)
importFrom(SparseArray,extract_sparse_array)
importFrom(SparseArray,nzvals)
importFrom(SparseArray,nzwhich)
importFrom(SparseArray,read_block_as_sparse)
importFrom(alabaster.base,.loadObject)
importFrom(alabaster.base,.stageObject)
importFrom(alabaster.base,.writeMetadata)
importFrom(alabaster.base,acquireFile)
importFrom(alabaster.base,acquireMetadata)
importFrom(alabaster.base,addMissingPlaceholderAttributeForHdf5)
importFrom(alabaster.base,loadObject)
importFrom(alabaster.base,saveObject)
importFrom(alabaster.base,transformVectorForHdf5)
importFrom(rhdf5,H5Fclose)
importFrom(rhdf5,H5Fopen)
importFrom(rhdf5,H5Gclose)
importFrom(rhdf5,H5Gopen)
importFrom(rhdf5,H5Sunlimited)
importFrom(rhdf5,h5createDataset)
importFrom(rhdf5,h5createFile)
importFrom(rhdf5,h5createGroup)
importFrom(rhdf5,h5ls)
importFrom(rhdf5,h5read)
importFrom(rhdf5,h5readAttributes)
importFrom(rhdf5,h5write)
importFrom(rhdf5,h5writeAttribute)
importFrom(rhdf5,h5writeDataset)
useDynLib(alabaster.matrix, .registration=TRUE)

# Save array-like objects to file

|Environment|Status|
|---|---|
|[BioC-release](https://bioconductor.org/packages/release/bioc/html/alabaster.matrix.html)|[![Release OK](https://bioconductor.org/shields/build/release/bioc/alabaster.matrix.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/alabaster.matrix/)|
|[BioC-devel](https://bioconductor.org/packages/devel/bioc/html/alabaster.matrix.html)|[![Devel OK](https://bioconductor.org/shields/build/devel/bioc/alabaster.matrix.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/alabaster.matrix/)|

The **alabaster.matrix** package implements methods for saving and loading matrix- or array-like objects under the **alabaster** framework.
It provides a language-agnostic method for serializing data in arrays or abstractions thereof.
To get started, install the package and its dependencies from Bioconductor:

```r
# install.packages("BiocManager")
BiocManager::install("alabaster.matrix")
```

We can then save a variety of matrices and arrays to file.
For example, a sparse matrix can be saved to a HDF5 file in compressed sparse column format:

```r
library(Matrix)
y <- rsparsematrix(1000, 100, density=0.05)

# Saving it to a directory.
library(alabaster.matrix)
tmp <- tempfile()
saveObject(y, tmp)

# Reading it as a file-backed matrix.
roundtrip <- readObject(tmp)
roundtrip
## <1000 x 100> sparse ReloadedMatrix object of type "double":
##           [,1]   [,2]   [,3] ...  [,99] [,100]
##    [1,]      0      0      0   .      0      0
##    [2,]      0      0      0   .      0      0
##    [3,]      0      0      0   .      0      0
##    [4,]      0      0      0   .      0      0
##    [5,]      0      0      0   .      0      0
##     ...      .      .      .   .      .      .
##  [996,]      0      0      0   .      0      0
##  [997,]      0      0      0   .      0      0
##  [998,]      0      0      0   .      0      0
##  [999,]      0      0      0   .      0      0
## [1000,]      0      0      0   .      0      0

# Coerce this back into an in-memory sparse matrix:
inmemory <- as(roundtrip, "dgCMatrix")
```

We can also handle [`DelayedArray`](https://bioconductor.org/packages/DelayedArray) objects, possibly with preservation of delayed operations.
This uses the [**chihaya** specification](https://github.com/ArtifactDB/chihaya) to represent delayed operations inside a HDF5 file.

```r
library(DelayedArray)
y <- DelayedArray(rsparsematrix(1000, 100, 0.05))
y <- log1p(abs(y) / 1:100) # adding some delayed ops.

# Default method saves without preserving delayed operations.
tmp <- tempfile()
saveObject(y, tmp)
readObjectFile(tmp)$type
## [1] "compressed_sparse_matrix"

# But we can enable the delayed'ness explicitly, if so desired.
tmp2 <- tempfile()
saveObject(y, tmp2, delayedarray.preserve.ops=TRUE)
readObjectFile(tmp2)$type
## [1] "delayed_array"

roundtrip <- readObject(tmp2)
roundtrip
## <1000 x 100> sparse ReloadedMatrix object of type "double":
##           [,1]   [,2]   [,3] ...       [,99]      [,100]
##    [1,]      0      0      0   .           0           0
##    [2,]      0      0      0   .           0           0
##    [3,]      0      0      0   .           0           0
##    [4,]      0      0      0   .           0           0
##    [5,]      0      0      0   .           0           0
##     ...      .      .      .   .           .           .
##  [996,]      0      0      0   . 0.000000000 0.007368618
##  [997,]      0      0      0   . 0.000000000 0.000000000
##  [998,]      0      0      0   . 0.000000000 0.000000000
##  [999,]      0      0      0   . 0.000000000 0.000000000
## [1000,]      0      0      0   . 0.000000000 0.000000000
``` 

#' Read a dense array from disk
#'
#' Read a dense high-dimensional array from its on-disk representation.
#' This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
#'
#' @param path String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a dense array.
#' @param metadata Named list of metadata for this object, see \code{\link{readObject}} for more details.
#' @param ... Further arguments, ignored.
#' 
#' @return A dense file-backed \linkS4class{ReloadedArray}.
#'
#' @seealso
#' \code{"\link{saveObject,array-method}"}, to create the directory and its contents.
#'
#' @author Aaron Lun
#'
#' @examples
#' arr <- array(rpois(10000, 10), c(50, 20, 10))
#' dimnames(arr) <- list(
#'    paste0("GENE_", seq_len(nrow(arr))),
#'    letters[1:20],
#'    NULL
#' )
#'
#' dir <- tempfile()
#' saveObject(arr, dir)
#' readObject(dir)
#' 
#' @export
#' @aliases
#' loadArray
#' @importFrom BiocGenerics type<-
#' @importFrom HDF5Array HDF5Array
readArray <- function(path, metadata, ...) {
    fpath <- file.path(path, "array.h5")

    details <- local({
        fhandle <- H5Fopen(fpath, flags="H5F_ACC_RDONLY")
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, "dense_array")
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

        type <- h5_read_attribute(ghandle, "type")
        transposed <- h5_read_attribute(ghandle, "transposed", check=TRUE, default=0L)

        if (type == "vls") {
            dhandle <- H5Dopen(ghandle, "pointers")
            on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        } else {
            dhandle <- H5Dopen(ghandle, "data")
            on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        }

        placeholder <- h5_read_attribute(dhandle, "missing-value-placeholder", check=TRUE, default=NULL)
        ndims <- H5Sget_simple_extent_dims(H5Dget_space(dhandle))$rank
        names <- load_names(ghandle, ndims)

        list(type=type, names=names, transposed=(transposed != 0L), placeholder=placeholder)
    })

    if (details$type == "vls") {
        # TODO: perhaps write a delayed version of this. 
        out <- h5_read_vls_array(fpath, "dense_array/pointers", "dense_array/heap", missing.placeholder=details$placeholder, native=!details$transposed)
        details$type <- "string"
        details$placeholder <- NULL
        details$transposed <- TRUE
    } else {
        out <- HDF5Array(filepath=fpath, name="dense_array/data")
        if (type(out) == "raw") { # ... so that placeholders are correctly substituted.
            type(out) <- "integer"
        }
    }

    if (!is.null(details$names)) {
        dimnames(out) <- rev(details$names)
    }
    if (!details$transposed) { # All R-based HDF5 bindings will automatically transpose, so we need to untranspose if `transposed=FALSE`.
        out <- t(out)
    }
    if (!is.null(details$placeholder)) {
        out <- DelayedMask(out, placeholder=details$placeholder)
        out <- DelayedArray(out)
    }
    intended.type <- from_array_type(details$type)
    if (type(out) != intended.type) {
        type(out) <- intended.type
    }

    ReloadedArray(path=path, seed=out)
}

##############################
######### OLD STUFF ##########
##############################

#' @export
loadArray <- function(info, project) {
    seed <- .createRawArraySeed(info, project=project, names=TRUE)
    DelayedArray(seed)
}

# This tests the stageObject generic.
# library(testthat); library(alabaster.matrix); source("test-stage-matrix.R")

experiment <- "rnaseq"
assay <- "counts"

mat <- matrix(rpois(10000, 10), ncol=10)
rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
colnames(mat) <- letters[1:10]

test_that("stageObject works as expected for the default method", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    info <- stageObject(mat, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "hdf5_dense_array")
    expect_identical(info$array$type, "integer")

    # Metadata is valid.
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    # Round tripping.
    mat2 <- loadArray(info, project=dir)
    expect_identical(DelayedArray::type(mat2), "integer")
    expect_equal(sum(mat2), sum(mat))
    expect_identical(rownames(mat), rownames(mat2))
    expect_identical(colnames(mat), colnames(mat2))
})

test_that("stageObject works as expected without dimnames", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    dimnames(mat) <- NULL
    info <- stageObject(mat, dir, file.path(experiment, assay))

    # Metadata is valid.
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    # Round-tripping.
    mat2 <- loadArray(info, project=dir)
    expect_equal(sum(mat2), sum(mat))
    expect_null(rownames(mat2))
    expect_null(colnames(mat2))
})

test_that("stageObject works with sparse matrices", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    mat <- as(matrix(rpois(10000, 1), ncol=10), "dgCMatrix")
    rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
    colnames(mat) <- letters[1:10]

    info <- stageObject(mat, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "sparse_matrix")
    expect_identical(info$array$type, "number")

    # Metadata is valid.
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    # Round tripping.
    mat2 <- loadArray(info, project=dir)
    expect_equal(sum(mat2), sum(mat))
    expect_identical(rownames(mat), rownames(mat2))
    expect_identical(colnames(mat), colnames(mat2))

    # Without dimnames.
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    mat <- as(matrix(rpois(10000, 1), ncol=10), "dgCMatrix")
    info <- stageObject(mat, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "sparse_matrix")
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    mat2 <- loadArray(info, project=dir)
    expect_null(rownames(mat2))
    expect_null(colnames(mat2))
})

test_that("stageObject works with character matrices and missing values", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    mat <- matrix(sample(LETTERS, 100, replace=TRUE), 25, 4)
    mat[1] <- "NA"
    mat[100] <- NA_character_

    info <- stageObject(mat, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "hdf5_dense")

    # Metadata is valid.
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    # Round-tripping.
    mat2 <- loadArray(info, project=dir)
    expect_identical(mat, as.matrix(mat2))
})

test_that("stageObject works with logical matrices", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    info <- stageObject(mat > 0, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "hdf5_dense_array")
    expect_identical(info$array$type, "boolean")

    mat2 <- loadArray(info, project=dir)
    expect_identical(DelayedArray::type(mat2), "logical")
    expect_identical(mat > 0, as.matrix(mat2))

    # Also with sparse counts.
    assay2 <- "counts2"
    mat <- Matrix::rsparsematrix(52, 14, 0.1) > 0

    info <- stageObject(mat, dir, file.path(experiment, assay2))
    expect_match(info$`$schema`, "hdf5_sparse_matrix")
    expect_identical(info$array$type, "boolean")

    mat2 <- loadArray(info, project=dir)
    expect_identical(DelayedArray::type(mat2), "logical")
    expect_identical(as.matrix(mat), as.matrix(mat2))
})

test_that("stageObject works with DelayedMatrices (naive)", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    # Works in the pristine case, where it just extracts the seed.
    library(DelayedArray)
    mat <- as(matrix(rpois(10000, 1), ncol=10), "dgCMatrix")
    x <- DelayedArray(mat)

    info <- stageObject(x, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "sparse_matrix")
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    mat2 <- loadArray(info, project=dir)
    expect_equal(unname(as.matrix(mat2)), unname(as.matrix(x)))
    expect_null(rownames(mat2))
    expect_null(colnames(mat2))

    # Works with assigned dimnames.
    y <- x
    dimnames(y) <- list(seq_len(nrow(y)), LETTERS[1:10])

    info <- stageObject(y, dir, file.path(experiment, "other"))
    expect_error(alabaster.base::.writeMetadata(info, dir), NA)

    mat2 <- loadArray(info, project=dir)
    expect_equal(as.matrix(mat2), as.matrix(y))
})

staging.options <- new.env()
staging.options$preserve.delayed <- FALSE
staging.options$recycle.hdf5 <- FALSE

#' Preserve delayed operations during staging
#'
#' Preserve delayed operations via \pkg{chihaya} when staging a \linkS4class{DelayedArray} with \code{\link{stageObject}}.
#'
#' @param preserve Whether to preserve delayed operations using the \pkg{chihaya} specification.
#'
#' @return Logical scalar indicating whether delayed operations are to be preserved by the DelayedArray method.
#' If \code{preserve} is supplied, it is used to set this scalar, and the \emph{previous} value of the scalar is invisibly returned.
#'
#' @details
#' By default, any DelayedArray in \code{\link{stageObject}} will be saved as a new dense array or sparse matrix.
#' However, if this option is enabled, DelayedArrays will instead be saved in the \pkg{chihaya} specification,
#' where the delayed operations are themselves stored in the HDF5 file (see \url{https://artifactdb.github.io/chihaya/} for details).
#'
#' The \pkg{chihaya} specification is more complicated to parse but can be helpful in reducing disk usage.
#' One simple example is to avoid sparsity-breaking or integer-to-float operations by storing their delayed representations in the file.
#' If the seed matrix is derived from some immutable reference location, advanced users can even store links to that location instead of duplicating the seed data.
#'
#' @author Aaron Lun
#'
#' @examples
#' preserveDelayedOperations()
#' old <- preserveDelayedOperations(TRUE)
#' preserveDelayedOperations()
#' preserveDelayedOperations(old)
#' 
#' @export
preserveDelayedOperations <- function(preserve) {
    prev <- staging.options$preserve.delayed
    if (missing(preserve)) {
        prev
    } else {
        staging.options$preserve.delayed <- preserve
        invisible(prev)
    }
}

#' Recycle existing HDF5 files
#'
#' Re-use existing files in HDF5-backed arrays rather than reserializing them in \code{\link{stageObject}}.
#'
#' @param recycle Whether to recycle existing files for HDF5-backed DelayedArrays.
#' 
#' @return Logical scalar indicating whether HDF5 files are to be reused.
#' If \code{recycle} is supplied, it is used to set this scalar, and the \emph{previous} value of the scalar is invisibly returned.
#'
#' @details
#' If this options is enabled, \code{stageObject} will attempt to link/copy existing files for any HDF5-backed DelayedArray instances -
#' most specifically, \linkS4class{HDF5Array} objects and \linkS4class{H5SparseMatrix} objects using the 10X format.
#' This avoids re-serialization of the data for faster staging.
#' It also allows advanced users to add their own customizations into the HDF5 file during staging, as long as they do not interfere with \code{\link{loadArray}}.
#'
#' By default, this option is disabled as the properties of the existing file are not known in the general case.
#' In particular, the file might contain other groups/datasets that are irrelevant, and use up extra disk space if copied; or confidential, and should not be stored in the staging directory.
#' Users should only enable this option if they have full control over the generation and contents of the backing HDF5 files.
#'
#' Also note that any dimnames on \code{x} will be ignored during recycling.
#'
#' @author Aaron Lun
#'
#' @examples
#' recycleHdf5Files()
#' old <- recycleHdf5Files(TRUE)
#' recycleHdf5Files()
#' recycleHdf5Files(old)
#' 
#' @export
recycleHdf5Files <- function(recycle) {
    prev <- staging.options$recycle.hdf5
    if (missing(recycle)) {
        prev
    } else {
        staging.options$recycle.hdf5 <- recycle
        invisible(prev)
    }
}

# This tests the stageObject generic for arrays.
# library(testthat); library(alabaster.matrix); source("test-stage-array.R")

experiment <- "rnaseq"
assay <- "counts"

arr <- array(rpois(10000, 10), c(50, 20, 10))
dimnames(arr) <- list(
   paste0("GENE_", seq_len(nrow(arr))),
   letters[1:20],
   NULL
)

test_that("stageObject works as expected for the default method", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    info <- stageObject(arr, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "hdf5_dense_array")

    # Round tripping.
    arr2 <- loadArray(info, project=dir)
    expect_equal(sum(arr2), sum(arr))
    expect_identical(rownames(arr), rownames(arr2))
    expect_identical(colnames(arr), colnames(arr2))

    # Checking that metadata save works.
    expect_error(alabaster.base::.writeMetadata(info, dir=dir), NA)
})

test_that("stageObject works as expected for NA values", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    storage.mode(arr) <- "integer"
    {
        info <- stageObject(arr, dir, file.path(experiment, "assay-0"))
        fpath <- file.path(odir, "assay-0/array.h5")
        out <- rhdf5::h5readAttributes(fpath, "data")
        expect_false("missing-value-placeholder" %in% names(out))

        arr2 <- loadArray(info, project=dir)
        expect_identical(DelayedArray::type(arr2), "integer")
        expect_equal(arr, as.array(arr2))
    }

    arr[1] <- NA
    {
        info <- stageObject(arr, dir, file.path(experiment, "assay-1"))
        fpath <- file.path(odir, "assay-1/array.h5")
        out <- rhdf5::h5readAttributes(fpath, "data")
        expect_identical(out[["missing-value-placeholder"]], NA_integer_)

        arr2 <- loadArray(info, project=dir)
        expect_identical(DelayedArray::type(arr2), "integer")
        expect_equal(arr, as.array(arr2))
    }

    storage.mode(arr) <- "double"
    arr[2] <- NaN
    arr[3] <- Inf
    arr[4] <- -Inf
    {
        info <- stageObject(arr, dir, file.path(experiment, "assay-2"))
        fpath <- file.path(odir, "assay-2/array.h5")
        out <- rhdf5::h5readAttributes(fpath, "data")
        #expect_identical(out[["missing-value-placeholder"]], alabaster.matrix:::lowest_double())

        arr2 <- loadArray(info, project=dir)
        expect_identical(DelayedArray::type(arr2), "double")
        expect_equal(arr, as.array(arr2))
    }
})

test_that("stageObject works as expected without dimnames", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    dimnames(arr) <- NULL
    info <- stageObject(arr, dir, file.path(experiment, assay))

    # Round-tripping.
    arr2 <- loadArray(info, project=dir)
    expect_equal(sum(arr2), sum(arr))
    expect_null(rownames(arr2))
    expect_null(colnames(arr2))
})

test_that("stageObject works with DelayedArrays", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    # Works in the pristine case.
    library(DelayedArray)
    dimnames(arr) <- NULL
    x <- DelayedArray(arr)

    info <- stageObject(x, dir, file.path(experiment, assay))
    expect_match(info$`$schema`, "hdf5_dense_array")

    arr2 <- loadArray(info, project=dir)
    expect_equal(unname(as.array(arr2)), unname(as.array(x)))
    expect_null(rownames(arr2))
    expect_null(colnames(arr2))

    # Works with assigned dimnames.
    y <- x
    dimnames(y) <- list(as.character(seq_len(50)), NULL, LETTERS[1:10])

    info <- stageObject(y, dir, file.path(experiment, "other"))
    arr2 <- loadArray(info, project=dir)
    expect_equal(as.array(arr2), as.array(y))
})

test_that("stageObject works with DelayedArrays with preserved operations", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    library(DelayedArray)
    x <- DelayedArray(arr)
    x <- log2(x / runif(nrow(x)) + 1)

    old <- preserveDelayedOperations(TRUE)
    info <- stageObject(x, dir, file.path(experiment, assay))
    preserveDelayedOperations(old)
    expect_match(info$`$schema`, "hdf5_delayed_array")
    expect_error(alabaster.base::.writeMetadata(info, dir=dir), NA)

    arr2 <- loadArray(info, project=dir)
    expect_equal(unname(as.array(arr2)), unname(as.array(x)))
})

test_that("stageObject works with DelayedArrays with local references", {
    dir <- tempfile()
    odir <- file.path(dir, experiment)
    dir.create(odir, recursive=TRUE)

    info <- stageObject(arr, dir, paste0(experiment, "/raw"))
    alabaster.base::.writeMetadata(info, dir=dir)
    raw.path <- info$path

    # Adding delayed operations.
    library(DelayedArray)
    x <- DelayedArray(arr)
    x <- log2(x / runif(nrow(x)) + 1)

    old <- preserveDelayedOperations(TRUE)
    info <- stageObject(x, dir, file.path(experiment, assay))
    preserveDelayedOperations(old)

    expect_match(info$`$schema`, "hdf5_delayed_array")
    expect_error(alabaster.base::.writeMetadata(info, dir=dir), NA)

    # Manually replacing the seed with a local reference.
    fpath <- file.path(dir, info$path)
    rhdf5::h5delete(fpath, "data/seed/seed/seed")
    rhdf5::h5createGroup(fpath, "data/seed/seed/seed")
    rhdf5::h5write(dim(arr), fpath, "data/seed/seed/seed/dimensions")
    rhdf5::h5write("integer", fpath, "data/seed/seed/seed/type")
    rhdf5::h5write(raw.path, fpath, "data/seed/seed/seed/path")

    {
        fhandle = rhdf5::H5Fopen(fpath, "H5F_ACC_RDWR")
        ghandle = rhdf5::H5Gopen(fhandle, "data/seed/seed/seed")
        rhdf5::h5writeAttribute("array", ghandle, "delayed_type", asScalar=TRUE)
        rhdf5::h5writeAttribute("custom alabaster local array", ghandle, "delayed_array", asScalar=TRUE)
        rhdf5::H5Gclose(ghandle)
        rhdf5::H5Fclose(fhandle)
    }

    # Now seeing if we can load it.
    arr2 <- loadArray(info, project=dir)
    expect_equal(unname(as.array(arr2)), unname(as.array(x)))
})

test_that("stageObject recycles existing HDF5Arrays", {
    dir <- tempfile()
    dir.create(dir, recursive=TRUE)

    library(HDF5Array)
    x <- as(matrix(runif(1000), 10, 100), "HDF5Array")
    rhdf5::h5write("foo", path(x), "bar") # injecting a little something to distinguish it from a rewritten file.

    {
        info <- stageObject(x, dir, "dense")
        expect_false(identical(file.size(path(x)), file.size(file.path(dir, info$path)))) # negative control that triggers the fallback.
        arr2 <- loadArray(info, project=dir)
        expect_equal(unname(as.array(arr2)), unname(as.array(x)))
    }

    old <- recycleHdf5Files(TRUE)
    on.exit(recycleHdf5Files(old))

    {
        info <- stageObject(x, dir, "dense2")
        expect_match(info$`$schema`, "hdf5_dense_array")
        expect_error(alabaster.base::.writeMetadata(info, dir=dir), NA)
        expect_identical(file.size(path(x)), file.size(file.path(dir, info$path))) # same file.

        arr2 <- loadArray(info, project=dir)
        expect_equal(unname(as.array(arr2)), unname(as.array(x)))
    }

    {
        y <- Matrix::rsparsematrix(20, 50, density=0.1)
        tmp <- tempfile(fileext=".h5")
        rhdf5::h5createFile(tmp)
        rhdf5::h5createGroup(tmp, "foo")
        rhdf5::h5write(y@p, tmp, "foo/indptr")
        rhdf5::h5write(y@i, tmp, "foo/indices")
        rhdf5::h5write(y@x, tmp, "foo/data")
        rhdf5::h5write(dim(y), tmp, "foo/shape")
        x <- H5SparseMatrix(tmp, "foo")

        info <- stageObject(x, dir, "sparse")
        expect_match(info$`$schema`, "hdf5_sparse_matrix")
        expect_error(alabaster.base::.writeMetadata(info, dir=dir), NA)
        expect_identical(file.size(path(x)), file.size(file.path(dir, info$path)))

        arr2 <- loadArray(info, project=dir)
        expect_equal(unname(as.array(arr2)), unname(as.array(x)))
    }
})

test_that("reading arrays work with non-default NA placeholders", {
    dir <- tempfile()
    dir.create(dir, recursive=TRUE)

    # Dense case.
    {
        x <- matrix(rpois(1000, 2), 100, 10)
        storage.mode(x) <- "integer"

        info <- stageObject(x, dir, "dense")
        alabaster.base::addMissingPlaceholderAttributeForHdf5(
            file.path(dir, info$path), 
            info$hdf5_dense_array$dataset, 
            0L
        )

        arr2 <- loadArray(info, project=dir)
        ref <- as.matrix(x)
        ref[ref==0L] <- NA
        expect_identical(ref, as.matrix(arr2))
    }

    # Sparse case.
    {
        x <- rsparsematrix(100, 20, 0.5)
        first <- x@x[1]

        info <- stageObject(x, dir, "sparse")
        alabaster.base::addMissingPlaceholderAttributeForHdf5(
            file.path(dir, info$path), 
            paste0(info$hdf5_sparse_matrix$group, "/data"), 
            first
        )

        arr2 <- loadArray(info, project=dir)
        ref <- as.matrix(x)
        ref[ref==first] <- NA
        expect_identical(ref, as.matrix(arr2))
    }
})

#' Write a sparse matrix
#'
#' Writes a sparse matrix to file in a compressed sparse format.
#'
#' @param x A sparse matrix of some sort.
#' This includes sparse \linkS4class{DelayedMatrix} objects.
#' @param file String containing a path to the HDF5 file.
#' The file is created if it is not already present.
#' @param name String containing the name of the group to store \code{x}.
#' @param chunk Integer scalar specifying the chunk size for the indices and values.
#' @param column Logical scalar indicating whether to store as compressed sparse column format.
#' @param tenx Logical scalar indicating whether to use the 10X compressed sparse column format.
#' @param guess.integer Logical scalar specifying whether to guess an appropriate integer type from \code{x}.
#'
#' @details
#' This writes a sparse matrix to file in various formats:
#' \itemize{
#' \item \code{column=TRUE} and \code{tenx=FALSE} uses H5AD's \code{csr_matrix} format.
#' \item \code{column=FALSE} and \code{tenx=FALSE} uses H5AD's \code{csc_matrix} format.
#' \item \code{tenx=TRUE} uses 10X Genomics' HDF5 matrix format.
#' }
#' For the first two formats, the apparent transposition is deliberate, because columns in R are interpreted as rows in H5AD.
#' This allows us to retain consistency the interpretation of samples (columns in R, rows in H5AD) and features (vice versa).
#' Constructors for classes like \linkS4class{H5SparseMatrix} will automatically transpose so no extra work is required.
#'
#' If \code{guess.integer=TRUE}, we attempt to save \code{x}'s values into the smallest type that will accommodate all of its values. 
#' If \code{x} only contains unsigned integers, we will attempt to save either 8-, 16- or 32-bit unsigned integers.
#' If \code{x} contains signed integers, we will fall back to 32-bit signed integers.
#' For all other values, we will fall back to double-precision floating point values.
#'
#' We attempt to save \code{x}'s indices to unsigned 16-bit integers if the relevant dimension of \code{x} is small enough.
#' Otherwise we will save it as an unsigned 32-bit integer.
#'
#' @return
#' A \code{NULL} invisibly.
#' The contents of \code{x} are written to \code{name} in \code{file}.
#' 
#' @author Aaron Lun
#'
#' @examples
#' library(Matrix)
#' x <- rsparsematrix(100, 20, 0.5)
#' tmp <- tempfile(fileext=".h5")
#' writeSparseMatrix(x, tmp, "csc_matrix")
#' writeSparseMatrix(x, tmp, "csr_matrix", column=FALSE)
#' writeSparseMatrix(x, tmp, "tenx_matrix", tenx = TRUE)
#' 
#' rhdf5::h5ls(tmp)
#' library(HDF5Array)
#' H5SparseMatrix(tmp, "csc_matrix")
#' H5SparseMatrix(tmp, "csr_matrix")
#' H5SparseMatrix(tmp, "tenx_matrix")
#' @export
#' @importFrom rhdf5 h5createGroup h5createFile
writeSparseMatrix <- function(x, file, name, chunk=10000, column=TRUE, tenx=FALSE, guess.integer=TRUE) {
    if (!file.exists(file)) {
        h5createFile(file)
    }
    .write_CS_matrix(file, name, x, chunk_dim=chunk, by_column=column, use_tenx=tenx, guess_type=guess.integer)
    invisible(NULL)
}

#' @import methods
#' @importFrom rhdf5 h5write h5createGroup H5Gopen H5Gopen H5Fopen H5Fclose 
#' h5writeAttribute h5createDataset h5writeDataset H5Sunlimited H5Gclose
#' @importFrom BiocGenerics type
#' @importFrom SparseArray read_block_as_sparse
#' @importFrom DelayedArray colAutoGrid rowAutoGrid
#' @importClassesFrom Matrix dgCMatrix
.write_CS_matrix <- function(file, name, mat, chunk_dim = 10000, by_column=TRUE, use_tenx=FALSE, guess_type=TRUE) {
    handle <- H5Fopen(file)
    on.exit(H5Fclose(handle))
    h5createGroup(handle, name)

    if (use_tenx) {
        by_column <- TRUE # must be column major.
        h5writeDataset(dim(mat), handle, paste0(name, "/shape"))
    } else {
        if (by_column) {
            format <- "csr_matrix"
        } else {
            format <- "csc_matrix"
        }

        ghandle <- H5Gopen(handle, name)
        on.exit(H5Gclose(ghandle), add = TRUE, after = FALSE)

        h5writeAttribute(format, ghandle, "encoding-type", variableLengthString=TRUE, encoding="UTF8", asScalar=TRUE)
        h5writeAttribute("0.1.0", ghandle, "encoding-version", variableLengthString=TRUE, encoding="UTF8", asScalar=TRUE)
        h5writeAttribute(rev(dim(mat)), ghandle, "shape")
    }

    # Scan through and guess the best types.
    details <- .extract_sparse_details(mat)
    any.neg <- details$negative
    extreme <- details$extreme
    any.nonint <- details$non.integer
    count <- details$count
    has.missing <- details$missing

    if (any.nonint) {
        type <- "DOUBLE"
    } else if (extreme >= 2^31) {
        # Don't attempt to use unsigned 32-bit ints; values between 2^31
        # and 2^32 will fail to be converted to R's signed integers.
        type <- "DOUBLE"
    } else if (has.missing) {
        # Don't bother choosing a smaller type, we need to encode NA_integer_ as -2^31.
        type <- "INT32" 
    } else if (any.neg) {
        type <- "INT32"
    } else {
        # We don't try to save it as a byte for now, because that gets
        # interpreted by rhdf5 as a raw vector... not helpful.
        if (extreme < 2^16) {
            type <- "UINT16"
        } else {
            # Again, don't attempt to use unsigned 32-bit ints.
            type <- "INT32"
        }
    }

    htype <- paste0("H5T_NATIVE_", type)

    if (by_column) {
        max.index <- nrow(mat)
    } else {
        max.index <- ncol(mat)
    }

    if (max.index < 2^16) {
        itype <- "H5T_NATIVE_UINT16"
    } else {
        itype <- "H5T_NATIVE_UINT32"
    }

    chunk_dim <- min(details$count, chunk_dim)

    h5createDataset(
        handle,
        paste0(name, "/data"),
        dims = details$count,
        H5type = htype,
        chunk = chunk_dim
    )

    is_integer <- type != "DOUBLE"
    transformer <- identity
    if (has.missing) {
        if (is_integer) {
            transformer <- as.integer
        } else {
            transformer <- as.double
        }
        addMissingPlaceholderAttributeForHdf5(file, paste0(name, "/data"), transformer(NA))
    }

    h5createDataset(
        handle,
        paste0(name, "/indices"),
        dims = details$count,
        H5type = itype,
        chunk = chunk_dim
    )

    if (by_column) {
        p <- .dump_column_sparse_matrix(
            mat, 
            handle, 
            data.path=paste0(name, "/data"), 
            index.path=paste0(name, "/indices"), 
            start=NULL,
            transformer=transformer
        )
    } else {
        p <- .dump_row_sparse_matrix(
            mat, 
            handle, 
            data.path=paste0(name, "/data"), 
            index.path=paste0(name, "/indices"), 
            start=NULL,
            transformer=transformer
        )
    }
    indptrs <- c(0, cumsum(as.double(p)))

    iname <- paste0(name, "/indptr")
    h5createDataset(
        handle,
        iname,
        dims = length(indptrs),
        H5type = "H5T_NATIVE_UINT64"
    )

    h5writeDataset(indptrs, handle, iname)
}

####################################################

setGeneric(".extract_sparse_details", function(x) standardGeneric(".extract_sparse_details"))

.check_for_missing_value <- function(x) {
    if (anyNA(x)) {
        if (is.double(x)) {
            if (sum(is.na(x)) > sum(is.nan(x))) {
                return(TRUE)
            }
        } else {
            return(TRUE)
        }
    }
    return(FALSE)
}

#' @importClassesFrom Matrix dsparseMatrix
#' @importFrom BiocGenerics type
#' @importFrom DelayedArray getAutoBlockLength
setMethod(".extract_sparse_details", "dsparseMatrix", function(x) {
    chunksize <- getAutoBlockLength(type(x))

    any.nonint <- FALSE
    i <- 1L
    while (i <= length(x@x)) {
        end <- i + chunksize - 1L
        current <- x@x[i:min(end, length(x@x))]
        if (any(current%%1!=0, na.rm=TRUE)) {
            any.nonint <- TRUE
            break
        } 
        i <- end + 1L
    }

    has.missing <- FALSE
    i <- 1L
    while (i <= length(x@x)) {
        end <- i + chunksize - 1L
        current <- x@x[i:min(end, length(x@x))]
        if (.check_for_missing_value(current)) {
            has.missing <- TRUE
            break
        }
        i <- end + 1L
    }

    limits <- range(x@x, na.rm=TRUE)
    list(
        missing = has.missing,
        negative = limits[1] < 0, 
        extreme = max(abs(limits), na.rm=TRUE),
        non.integer = any.nonint,
        count = length(x@x)
    )
})

#' @importClassesFrom SparseArray SVT_SparseMatrix
setMethod(".extract_sparse_details", "SVT_SparseMatrix", function(x) {
    # Can't be bothered adding support for a soft-deprecated function.
    if (x@.svt_version > 0L) {
        stop("SVT_SparseMatrix objects of version >= 1 are not yet supported")
    }

    limits <- range(unlist(lapply(x@SVT, function(x) range(x[[2]], na.rm=TRUE))), na.rm=TRUE)
    non.int <- any(vapply(x@SVT, function(x) any(x[[2]]%%1 != 0, na.rm=TRUE), TRUE), na.rm=TRUE)
    has.missing <- any(vapply(x@SVT, function(x) .check_for_missing_value(x[[2]]), TRUE), na.rm=TRUE)
    count <- sum(vapply(x@SVT, function(x) length(x[[1]]), 0L), na.rm=TRUE)

    list(
        missing = has.missing,
        negative = limits[1] < 0, 
        extreme = max(abs(limits), na.rm=TRUE),
        non.integer = non.int,
        count = count
    )
})

#' @importClassesFrom DelayedArray DelayedSetDimnames
setMethod(".extract_sparse_details", "DelayedSetDimnames", function(x) .extract_sparse_details(x@seed))

.combine_extracted_details <- function(collected) {
    list(
        missing = any(vapply(collected, function(y) y$missing, TRUE), na.rm=TRUE),
        negative = any(vapply(collected, function(y) y$negative, TRUE), na.rm=TRUE),
        extreme = max(unlist(lapply(collected, function(y) y$extreme)), na.rm=TRUE),
        non.integer = any(vapply(collected, function(y) y$non.integer, TRUE), na.rm=TRUE),
        count = sum(unlist(lapply(collected, function(y) y$count)), na.rm=TRUE)
    )
}

#' @importClassesFrom DelayedArray DelayedAbind
setMethod(".extract_sparse_details", "DelayedAbind", function(x) {
    collected <- lapply(x@seeds, .extract_sparse_details)
    .combine_extracted_details(collected)
})

#' @importClassesFrom DelayedArray DelayedMatrix
setMethod(".extract_sparse_details", "DelayedMatrix", function(x) .extract_sparse_details(x@seed))

#' @importFrom SparseArray nzvals
.extract_sparse_details_fragment <- function(sparse) {
    vals <- nzvals(sparse)
    any.negative <- any(vals < 0, na.rm=TRUE)
    any.nonint <- any(vals != round(vals), na.rm=TRUE)
    extreme.val <- max(abs(vals), na.rm=TRUE)
    has.missing <- .check_for_missing_value(vals)

    list(
        missing = has.missing,
        negative = any.negative, 
        non.integer = any.nonint, 
        extreme = extreme.val, 
        count = length(vals)
    )
}

#' @importFrom DelayedArray blockApply
setMethod(".extract_sparse_details", "ANY", function(x) {
    collected <- blockApply(x, FUN = .extract_sparse_details_fragment, as.sparse = TRUE) # any grid works here.
    .combine_extracted_details(collected)
})

####################################################

setGeneric(".dump_column_sparse_matrix", function(x, handle, index.path, data.path, start, transformer) {
    standardGeneric(".dump_column_sparse_matrix")
})

#' @importFrom rhdf5 h5writeDataset
#' @importClassesFrom Matrix dgCMatrix
setMethod(".dump_column_sparse_matrix", "dgCMatrix", function(x, handle, index.path, data.path, start, transformer) {
    if (is.null(start)) {
        h5writeDataset(transformer(x@x), handle, data.path)
        h5writeDataset(x@i, handle, index.path)
    } else {
        index <- list(as.double(start) + seq_along(x@x))
        h5writeDataset(transformer(x@x), handle, data.path, index=index)
        h5writeDataset(x@i, handle, index.path, index=index)
    }
    diff(x@p)
})

.sanitize_start <- function(start) {
    if (is.null(start)) {
        0
    } else {
        as.double(start) # avoid integer overflow
    }
}

#' @importFrom rhdf5 h5writeDataset
#' @importFrom BiocGenerics type
#' @importClassesFrom SparseArray SVT_SparseMatrix
#' @importFrom DelayedArray getAutoBlockSize
setMethod(".dump_column_sparse_matrix", "SVT_SparseMatrix", function(x, handle, index.path, data.path, start, transformer) {
    if (x@.svt_version > 0L) {
        stop("SVT_SparseMatrix objects of version >= 1 are not yet supported")
    }

    # Processing things in chunks to reduce the number of HDF5 calls.
    chunksize <- min(getAutoBlockLength("integer"), getAutoBlockLength(type(x)))
    column.counts <- vapply(x@SVT, function(y) length(y[[1]]), 0L)

    start <- .sanitize_start(start)
    cached <- 0
    last.cleared <- 0L

    for (i in seq_len(ncol(x))) {
        cached <- cached + column.counts[i]

        if (cached >= chunksize || i == ncol(x)) {
            targets <- x@SVT[(last.cleared + 1L):i]
            all.i <- unlist(lapply(targets, function(y) y[[1]]))
            all.d <- unlist(lapply(targets, function(y) y[[2]]))

            index <- list(start + seq_along(all.i))
            h5writeDataset(all.i, handle, index.path, index = index)
            h5writeDataset(transformer(all.d), handle, data.path, index = index)

            last.cleared <- i
            cached <- 0
            start <- start + length(all.i)
        }
    }

    column.counts
})

#' @importClassesFrom DelayedArray DelayedSetDimnames
setMethod(".dump_column_sparse_matrix", "DelayedSetDimnames", function(x, handle, index.path, data.path, start, transformer) {
    .dump_column_sparse_matrix(x@seed, handle, index.path, data.path, start, transformer)
})

#' @importClassesFrom DelayedArray DelayedMatrix
setMethod(".dump_column_sparse_matrix", "DelayedMatrix", function(x, handle, index.path, data.path, start, transformer) {
    .dump_column_sparse_matrix(x@seed, handle, index.path, data.path, start, transformer)
})

#' @importClassesFrom DelayedArray DelayedAbind
setMethod(".dump_column_sparse_matrix", "DelayedAbind", function(x, handle, index.path, data.path, start, transformer) {
    if (x@along != 2L) {
        return(callNextMethod()) # goes to the ANY method.
    }

    start <- .sanitize_start(start)
    collected <- vector("list", length(x@seeds))
    for (s in seq_along(x@seeds)) {
        current <- .dump_column_sparse_matrix(x@seeds[[s]], handle, index.path, data.path, start=start, transformer=transformer)
        collected[[s]] <- current
        start <- start + sum(as.double(current))
    }

    unlist(collected)
})

#' @importFrom SparseArray read_block_as_sparse
#' @importFrom DelayedArray colAutoGrid
setMethod(".dump_column_sparse_matrix", "ANY", function(x, handle, index.path, data.path, start, transformer) {
    start <- .sanitize_start(start)
    grid <- colAutoGrid(x)
    out <- vector("list", length(grid))

    for (i in seq_along(grid)) {
        block <- read_block_as_sparse(x, grid[[i]])
        cout <- .blockwise_sparse_writer(
            block, 
            start, 
            file=handle, 
            transformer=transformer,
            index.path=index.path, 
            data.path=data.path, 
            by_column=TRUE
        )
        
        out[[i]] <- cout$number
        start <- cout$last
    }

    unlist(out)
})

#' @importFrom SparseArray read_block_as_sparse
#' @importFrom DelayedArray rowAutoGrid
.dump_row_sparse_matrix <- function(x, handle, index.path, data.path, start, transformer) {
    start <- .sanitize_start(start)
    grid <- rowAutoGrid(x)
    out <- vector("list", length(grid))

    for (i in seq_along(grid)) {
        block <- read_block_as_sparse(x, grid[[i]])
        cout <- .blockwise_sparse_writer(
            block, 
            start, 
            transformer=transformer, 
            file=handle, 
            index.path=index.path, 
            data.path=data.path, 
            by_column=FALSE
        )

        out[[i]] <- cout$number
        start <- cout$last
    }

    unlist(out)
}

#' @importFrom SparseArray nzwhich nzvals
#' @importFrom rhdf5 h5writeDataset
.blockwise_sparse_writer <- function(block, last, transformer, file, index.path, data.path, by_column) {
    nzdex <- nzwhich(block, arr.ind=TRUE)
    if (by_column) {
        primary <- nzdex[, 2]
        secondary <- nzdex[, 1]
        ndim <- ncol(block)
    } else {
        primary <- nzdex[, 1]
        secondary <- nzdex[, 2]
        ndim <- nrow(block)
    }
    v <- nzvals(block)

    o <- order(primary, secondary)
    primary <- primary[o]
    secondary <- secondary[o]
    v <- v[o]

    newlast <- last + length(primary)
    index <- list(last + seq_along(primary))
    h5writeDataset(secondary - 1L, file, index.path, index = index)
    h5writeDataset(transformer(v), file, data.path, index = index)

    list(number=tabulate(primary, ndim), last=newlast)
}

# This tests the sparse readers/writers.
# library(testthat); library(alabaster.matrix); source("test-sparse.R")

library(Matrix)
library(SparseArray)

test_that("writing to a sparse matrix works as expected", {
    for (i in 1:3) {
        x <- rsparsematrix(100, 20, 0.5)
        if (i == 2) {
            x <- DelayedArray(x) * 1 # force use of the block method.
        } else if (i == 3) {
            x <- BiocGenerics::cbind(DelayedArray(x), DelayedArray(rsparsematrix(100, 10, 0.5)))
        } else if (i == 4) {
            x <- as(x, "SVT_SparseMatrix")
        } else if (i == 5) {
            x <- BiocGenerics::cbind(DelayedArray(as(x, "SVT_SparseMatrix")), DelayedArray(as(rsparsematrix(100, 10, 0.5), "SVT_SparseMatrix")))
        }

        tmp <- tempfile(fileext=".h5")
        writeSparseMatrix(x, tmp, "csc_matrix")
        if (i <= 3) {
            writeSparseMatrix(x, tmp, "csr_matrix", column=FALSE)
        }
        writeSparseMatrix(x, tmp, "tenx_matrix", tenx=TRUE)

        library(HDF5Array)
        expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(x)))
        if (i <= 3) {
            expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csr_matrix"))), unname(as.matrix(x)))
        }
        expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "tenx_matrix"))), unname(as.matrix(x)))
    }
})

test_that("writing to a sparse matrix works with tiny chunks", {
    for (i in 1:2) {
        x <- rsparsematrix(100, 20, 0.5)
        if (i == 2) {
            x <- DelayedArray(x) * 1 # force use of the block method.
        } else if (i == 3) {
            x <- as(x, "SVT_SparseMatrix")
        }

        tmp <- tempfile(fileext=".h5")
        setAutoBlockSize(max(dim(x))*8)
        writeSparseMatrix(x, tmp, "csc_matrix")
        if (i <= 2) {
            writeSparseMatrix(x, tmp, "csr_matrix", column=FALSE)
        }

        library(HDF5Array)
        expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(x)))
        if (i <= 2) {
            expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csr_matrix"))), unname(as.matrix(x)))
        }
    }

    setAutoBlockSize()
})

get_type <- function(tmp, path) {
    fhandle <- H5Fopen(tmp)
    on.exit(H5Fclose(fhandle))
    dhandle <- H5Dopen(fhandle, path)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    .Call("_getDatatypeName", H5Dget_type(dhandle), PACKAGE = "rhdf5")
}

test_that("writing to a sparse matrix works with guessed type", {
    set.seed(1000)
    for (i in 1:3) {
        core <- function() round(rsparsematrix(100, 20, 0.5))
        if (i == 1) {
            FUN <- function(f) f(core())
        } else if (i == 2) {
            FUN <- function(f) DelayedArray(f(core())) * 1 # force use of the block method.
        } else if (i == 3) {
            FUN <- function(f) BiocGenerics::cbind(DelayedArray(f(core())), DelayedArray(f(round(rsparsematrix(100, 10, 0.5)))))
        } else if (i == 4) {
            FUN <- function(f) as(f(core()), "SVT_SparseMatrix")
        }

        # Signed:
        tmp <- tempfile(fileext=".h5")
        y <- FUN(identity)
        writeSparseMatrix(y, tmp, "csc_matrix")
        expect_match(get_type(tmp, "csc_matrix/data"), "I32")
        expect_equal(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(y)))

        # Unsigned short
        tmp <- tempfile(fileext=".h5")
        y <- FUN(function(x) abs(x * 1000))
        writeSparseMatrix(y, tmp, "csc_matrix")
        expect_match(get_type(tmp, "csc_matrix/data"), "U16")
        expect_equal(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(y)))

        # Unsigned word
        tmp <- tempfile(fileext=".h5")
        y <- FUN(function(x) abs(x * 1000000))
        writeSparseMatrix(y, tmp, "csc_matrix")
        expect_match(get_type(tmp, "csc_matrix/data"), "I32")
        expect_equal(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(y)))

        # Unsigned long
        tmp <- tempfile(fileext=".h5")
        y <- FUN(function(x) abs(x * 10000000000))
        writeSparseMatrix(y, tmp, "csc_matrix")
        expect_match(get_type(tmp, "csc_matrix/data"), "F64")
        expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(y)))
    }
})

test_that("writing to a sparse matrix works with guessed index type for block method", {
    set.seed(1000)
    for (i in 1:2) {
        for (big in c(TRUE, FALSE)) {
            nr <- if (big) 100000 else 100

            x <- round(rsparsematrix(nr, 20, 0.5))
            if (i == 2) {
                x <- DelayedArray(x) * 1 # force use of the block method.
            } else if (i == 3) {
                x <- as(x, "SVT_SparseMatrix")
            }

            tmp <- tempfile(fileext=".h5")
            writeSparseMatrix(x, tmp, "csc_matrix")
            expected.type <- if (big) "U32" else "U16"
            expect_match(get_type(tmp, "csc_matrix/indices"), expected.type)
        }
    }
})

test_that("writing to a sparse matrix works with NA values", {
    library(HDF5Array)
    x <- rsparsematrix(100, 20, 0.5)

    # No NA's added.
    {
        tmp <- tempfile(fileext=".h5")
        writeSparseMatrix(x, tmp, "csc_matrix")
        expect_null(rhdf5::h5readAttributes(tmp, "csc_matrix/data")[["missing-value-placeholder"]])
    }

    # Double-precision mode.
    x@x[1] <- NA
    {
        tmp <- tempfile(fileext=".h5")
        writeSparseMatrix(x, tmp, "csc_matrix")
        expect_identical(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(x)))
        expect_identical(rhdf5::h5readAttributes(tmp, "csc_matrix/data")[["missing-value-placeholder"]], NA_real_)
    }

    # Integer mode.
    x <- round(x)
    {
        tmp <- tempfile(fileext=".h5")
        writeSparseMatrix(x, tmp, "csc_matrix")
        expect_equal(unname(as.matrix(H5SparseMatrix(tmp, "csc_matrix"))), unname(as.matrix(x)))
        expect_identical(rhdf5::h5readAttributes(tmp, "csc_matrix/data")[["missing-value-placeholder"]], NA_integer_)
    }
})

.onLoad <- function(libname, pkgname) {
    registerReadObjectFunction("dense_array", readArray)
    registerReadObjectFunction("compressed_sparse_matrix", readSparseMatrix)
    registerReadObjectFunction("delayed_array", readDelayedArray)
}

.onUnload <- function(libname, pkgname) {
    registerReadObjectFunction("dense_array", NULL)
    registerReadObjectFunction("compressed_sparse_matrix", NULL)
    registerReadObjectFunction("delayed_array", NULL)
}

#' Array loading utilities
#'
#' Utilities for loading an array saved by \code{\link{stageObject}}. 
#'
#' @param info A named list of metadata for this array.
#' @param project Any argument accepted by the acquisition functions, see \code{?\link{acquireFile}}. 
#' By default, this should be a string containing the path to a staging directory.
#' @param names Logical scalar indicating whether the seed should be annotated with dimnames (if available).
#' @param path String containing the path to the file containing said array.
#' @param group String containing the name of the group with the dimnames.
#' @param ndim Integer scalar specifying the number of dimensions.
#'
#' @return \code{.createRawArraySeed} returns a seed that can be used in the \code{\link{DelayedArray}} constructor.
#'
#' \code{.extractArrayDimnames} returns a list of character vectors or \code{NULL}, containing the dimnames.
#'
#' @details
#' For \code{.createArraySeed}, the array should be one of:
#' \itemize{
#' \item \code{hdf5_dense_array}
#' \item \code{hdf5_sparse_matrix}
#' \item \code{hdf5_delayed_array}
#' \item \code{amalgamated_array}
#' }
#'
#' For delayed arrays, the file may contain a seed array with the \code{"custom alabaster local array"} type.
#' This should have a \code{path} dataset containing a relative path to another array in the same \code{project}, which is loaded and used as the seed for this delayed array.
#' Callers can overwrite this behavior by setting \code{"custom alabaster local array"} in the \code{knownArrays} from \pkg{chihaya} before calling \code{.createRawArraySeed}.
#' 
#' For \code{.extractArrayDimnames}, \code{path} is expected to be a HDF5 file with a group specified by \code{group}.
#' Each child of this group is a string dataset named after a (0-indexed) dimension, containing the names for that dimension.
#'
#' @author Aaron Lun
#' @examples
#' # Staging an array as an example:
#' dir <- tempfile()
#' dir.create(dir)
#' mat <- array(rpois(10000, 10), c(50, 20, 10))
#' meta <- stageObject(mat, dir, "whee")
#'
#' # Loading it back as a DelayedArray seed:
#' .createRawArraySeed(meta, project=dir)
#'
#' @export
#' @name createRawArraySeed
#' @importFrom HDF5Array HDF5ArraySeed H5SparseMatrixSeed
#' @importFrom rhdf5 h5read h5readAttributes
#' @importFrom alabaster.base loadObject acquireFile acquireMetadata
.createRawArraySeed <- function(info, project, names = TRUE) {
    path <- acquireFile(project, info$path)

    if ("hdf5_delayed_array" %in% names(info)) {
        copy <- chihaya::knownArrays()
        key <- "custom alabaster local array"
        if (!(key %in% names(copy))) {
            copy[[key]] <- function(file, name, contents) {
                location <- h5read(file, file.path(name, "path"))
                info <- acquireMetadata(project, location)
                loadObject(info, project)
            }
            old <- chihaya::knownArrays(copy)
            on.exit(chihaya::knownArrays(old))
        }

        group <- info$hdf5_delayed_array$group
        return(chihaya::loadDelayed(path, group))
    }

    if ("hdf5_dense_array" %in% names(info)) {
        dtype <- if (identical(info$array$type, "boolean")) "logical" else NA
        ds <- info$hdf5_dense_array$dataset
        out <- HDF5ArraySeed(filepath=path, name=ds, type=dtype)

        attrs <- h5readAttributes(path, ds)
        miss.place <- attrs[["missing-value-placeholder"]]
        if (!is.null(miss.place)) {
            out <- DelayedMask(out, placeholder=miss.place)
        }

        name.group <- if (names) info$hdf5_dense_array$dimnames else NULL
        return(.array_namer(out, path, name.group))
    }

    if ("hdf5_sparse_matrix" %in% names(info)) {
        group <- info$hdf5_sparse_matrix$group
        out <- H5SparseMatrixSeed(filepath=path, group=group)

        attrs <- h5readAttributes(path, paste0(group, "/data"))
        miss.place <- attrs[["missing-value-placeholder"]]
        if (!is.null(miss.place)) {
            out <- DelayedMask(out, placeholder=miss.place)
        }

        if (identical(info$array$type, "boolean")) {
            out <- (DelayedArray(out) != 0L)@seed
        }

        name.group <- if (names) info$hdf5_sparse_matrix$dimnames else NULL
        return(.array_namer(out, path, name.group))
    }

    if ("amalgamated_array" %in% names(info)) {
        return(.load_amalgamated_array(info, project))
    }

    stop("unsupported type '", info$`_extra`$type, "'")
}

#' @importFrom DelayedArray DelayedArray
.array_namer <- function(seed, path, names.group) {
    if (!is.null(names.group)) {
        dnames <- .extractArrayDimnames(path, names.group, length(dim(seed)))
        if (!is.null(dnames)) {
            mat <- DelayedArray(seed)
            dimnames(mat) <- dnames
            seed <- mat@seed
        }
    }
    seed
}

#' @export
#' @rdname createRawArraySeed
#' @importFrom rhdf5 h5read h5ls
.extractArrayDimnames <- function(path, group, ndim) {
    info <- h5ls(path)

    dimnames <- vector("list", ndim)
    all.names <- which(info$group == paste0("/", group)) 
    for (i in all.names) {
        name <- info$name[i]
        dimnames[[as.integer(name) + 1L]] <- as.character(h5read(path, file.path("names", name)))
    }

    if (all(vapply(dimnames, is.null, TRUE))) {
        NULL
    } else {
        dimnames
    }
}

#' Read a sparse matrix from disk
#'
#' Read a sparse matrix from its on-disk representation.
#' This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
#'
#' @param path String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a spars matrix.
#' @param metadata Named list of metadata for this object, see \code{\link{readObject}} for more details.
#' @param ... Further arguments, ignored.
#' 
#' @return A sparse \linkS4class{ReloadedMatrix} object.
#' 
#' @seealso
#' \code{"\link{saveObject,sparseMatrix-method}"}, to create the directory and its contents.
#'
#' @author Aaron Lun
#'
#' @examples
#' mat <- Matrix::rsparsematrix(100, 200, density=0.2)
#' rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
#' dir <- tempfile()
#' saveObject(mat, dir)
#' readObject(dir)
#' 
#' @export
#' @importFrom BiocGenerics type<-
#' @importFrom HDF5Array H5SparseMatrixSeed
readSparseMatrix <- function(path, metadata, ...) {
    fpath <- file.path(path, "matrix.h5")
    name <- "compressed_sparse_matrix"

    details <- local({
        fhandle <- H5Fopen(fpath, flags="H5F_ACC_RDONLY")
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, name)
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

        type <- h5_read_attribute(ghandle, "type")
        layout <- h5_read_attribute(ghandle, "layout")
        shape <- h5_read_vector(ghandle, "shape")

        dhandle <- H5Dopen(ghandle, "data")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        placeholder <- h5_read_attribute(dhandle, "missing-value-placeholder", check=TRUE, default=NULL)

        names <- load_names(ghandle, 2L)
        list(type=type, shape=shape, names=names, placeholder=placeholder, layout=layout)
    })

    out <- DelayedArray(H5SparseMatrixSeed(filepath=fpath, group=name, dim=details$shape, sparse.layout=tolower(details$layout)))
    if (type(out) == "raw") { # ... so that placeholders are correctly substituted.
        type(out) <- "integer"
    }

    if (!is.null(details$names)) {
        dimnames(out) <- details$names
    }
    if (!is.null(details$placeholder)) {
        out <- DelayedMask(out, placeholder=details$placeholder)
        out <- DelayedArray(out)
    }
    intended.type <- from_array_type(details$type)
    if (type(out) != intended.type) {
        type(out) <- intended.type
    }

    ReloadedArray(path, seed=out)
}

# This tests the sparse readers/writers.
# library(testthat); library(alabaster.matrix); source("setup.R"); source("test-SparseMatrix.R")

library(Matrix)
library(SparseArray)
library(DelayedArray)

test_that("reading a sparse matrix works with different output types", {
    x <- rsparsematrix(109, 297, 0.5)

    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(normalizePath(BiocGenerics::path(roundtrip)), normalizePath(tmp))
    expect_s4_class(roundtrip, "ReloadedArray")
    expect_true(is_sparse(roundtrip))
    expect_identical(as(roundtrip, "dgCMatrix"), x)

    # Trying with a logical matrix.
    x <- rsparsematrix(109, 297, 0.5) > 0
    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip2 <- readObject(tmp)
    expect_identical(as.matrix(roundtrip2), as.matrix(x)) # use dense matrix as 'x' still contains zeros in its structural non-zeros.

    # Trying with an integer matrix.
    x <- rsparsematrix(109, 297, 0.5) * 10
    x <- as(x, "SVT_SparseMatrix")
    type(x) <- "integer"
    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip3 <- readObject(tmp)
    expect_identical(as(roundtrip3, "dgCMatrix"), as(x, "dgCMatrix"))
})

test_that("writing to a sparse matrix works as expected for numeric data", {
    for (i in 1:4) {
        for (miss in c(TRUE, FALSE)) {
            x <- rsparsematrix(100, 20, 0.5)
            if (miss) {
                x[1,1] <- NA
            }

            if (i == 2) {
                x <- as(x, "RsparseMatrix")
            } else if (i == 3) {
                x <- as(x, "SVT_SparseMatrix")
            } else if (i == 4) {
                x <- DelayedArray(x) * 1 # force use of the block method.
            }

            tmp <- tempfile(fileext=".h5")
            saveObject(x, tmp)
            roundtrip <- readObject(tmp)
            expect_identical_without_names(as.matrix(roundtrip), as.matrix(x))
        }
    }
})

test_that("writing to a sparse matrix works as expected for logical data", {
    for (i in 1:4) {
        for (miss in c(TRUE, FALSE)) {
            x <- rsparsematrix(100, 20, 0.5) > 0
            if (miss) {
                x[1,1] <- NA
            }

            if (i == 2) {
                x <- as(x, "RsparseMatrix")
            } else if (i == 3) {
                x <- as(x, "SVT_SparseMatrix")
            } else if (i == 4) {
                x <- DelayedArray(x) & TRUE # force use of the block method.
            }

            tmp <- tempfile(fileext=".h5")
            saveObject(x, tmp)
            roundtrip <- readObject(tmp)
            expect_identical_without_names(as.matrix(roundtrip), as.matrix(x))
        }
    }
})

test_that("writing to a sparse matrix works as expected for integer data", {
    for (i in 1:2) {
        for (miss in c(TRUE, FALSE)) {
            x <- round(rsparsematrix(100, 20, 0.5) * 10)
            if (miss) {
                x[1,1] <- NA
            }

            if (i == 1) {
                x <- as(x, "SVT_SparseMatrix")
            } else if (i == 2) {
                x <- DelayedArray(x) * 1L # force use of the block method.
            }
            type(x) <- "integer"

            tmp <- tempfile(fileext=".h5")
            saveObject(x, tmp)
            roundtrip <- readObject(tmp)
            expect_identical_without_names(as.matrix(roundtrip), as.matrix(x))
        }
    }
})

test_that("depositing a large sparseMatrix vector works correctly", {
    # We generate 1 million non-zero elements, which should exceed
    # the size of the chunks used inside the HDF5 file.
    x <- rsparsematrix(10000, 500, 0.2)
    expect_true(length(x@x) > h5_guess_vector_chunks(length(x@x)))

    tmp <- tempfile(fileext=".h5")
    saveObject(x, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as(roundtrip, 'dgCMatrix'), x)

    # Now injecting an NA, which should force it to use chunk-wise processing.
    x@x[1] <- NA

    tmp <- tempfile(fileext=".h5")
    saveObject(x, tmp)
    roundtrip <- readSparseMatrix(tmp)
#    expect_identical(as(roundtrip, 'dgCMatrix'), x) # TODO: bug in HDF5Array
})

test_that("depositing small chunks works correctly", {
    x <- rsparsematrix(1000, 500, 0.2)

    # For DelayedArrays:
    y <- DelayedArray(x) * 1 # force block processing.

    tmp <- tempfile()
    local({
        old <- getAutoBlockSize()
        setAutoBlockSize(2000)
        on.exit(setAutoBlockSize(old))
        saveObject(y, tmp)
    })

    roundtrip <- readObject(tmp)
    expect_identical(as(roundtrip, 'dgCMatrix'), x)

    # For SVT_SparseMatrix objects.
    y <- as(x, "SVT_SparseMatrix")

    tmp <- tempfile()
    local({
        old <- getAutoBlockSize()
        setAutoBlockSize(2000)
        on.exit(setAutoBlockSize(old))
        saveObject(y, tmp)
    })

    roundtrip <- readObject(tmp)
    expect_identical(as(roundtrip, 'dgCMatrix'), x)
})

test_that("fallback to large integer types for indices works correctly", {
    for (i in 1:4) {
        x <- rsparsematrix(100000, 20, 0.001)
        x[100000,20] <- 99 # making sure there's a value at the bottom-right so that we check the index correctly.

        if (i == 2) {
            x <- Matrix::t(x)
            x <- as(x, "RsparseMatrix")
        } else if (i == 3) {
            x <- as(x, "SVT_SparseMatrix")
        } else if (i == 4) {
            x <- DelayedArray(x) * 1 # force use of the block method.
        }

        tmp <- tempfile(fileext=".h5")
        saveObject(x, tmp)
        roundtrip <- readObject(tmp)
        expect_identical_without_names(as.matrix(roundtrip), as.matrix(x))
    }
})

test_that("writing to a sparse matrix works with different integer types", {
    for (i in 1:2) { # 1:3) { # bug in H5SparseMatrixSeed for INT types greater than int32.
        if (i == 1) {
            bits <- 8
        } else if (i == 2) {
            bits <- 16
        } else {
            bits <- 32
        }

        core <- matrix(0, 100, 200)
        chosen <- sample(20000, 1000)
        core[chosen] <- floor(runif(1000, 0, 2^bits))

        tmp <- tempfile()
        y <- as(core, "dgCMatrix")
        saveObject(y, tmp)
        expect_identical(as.matrix(readObject(tmp)), core)

        core[chosen] <- core[chosen] - 2^(bits-1)
        y <- as(core, "dgCMatrix")
        tmp <- tempfile()
        saveObject(y, tmp)
        expect_identical(as.matrix(readObject(tmp)), core)
    }
})

test_that("writing a DelayedMatrixworks with row-major storage", {
    xc <- rsparsematrix(100, 20, 0.5)
    core <- as.matrix(xc)

    # Regular:
    tmp <- tempfile()
    saveObject(xc, tmp)
    expect_identical(rhdf5::h5readAttributes(file.path(tmp, "matrix.h5"), "compressed_sparse_matrix")$layout, "CSC")
    expect_identical(unname(as.matrix(readObject(tmp))), core)

    # Saving as CSC based on chunkdims:
    setClass("TestColMatrix", contains="dgCMatrix")
    setMethod("chunkdim", "TestColMatrix", function(x) c(nrow(x), 1L))
    y <- as(xc, "TestColMatrix")
    y <- DelayedArray(y) * 1 # force the use of block processing.

    tmp <- tempfile()
    saveObject(y, tmp)
    expect_identical(rhdf5::h5readAttributes(file.path(tmp, "matrix.h5"), "compressed_sparse_matrix")$layout, "CSC")
    expect_identical(as.matrix(readObject(tmp)), core)

    # Saving as CSR:
    xr <- as(xc, "RsparseMatrix")

    setClass("TestRowMatrix", contains="dgRMatrix")
    setMethod("chunkdim", "TestRowMatrix", function(x) c(1L, ncol(x)))
    y <- as(xr, "TestRowMatrix")
    y <- DelayedArray(y) * 1 # force the use of block processing.

    tmp <- tempfile()
    saveObject(y, tmp)
    expect_identical(rhdf5::h5readAttributes(file.path(tmp, "matrix.h5"), "compressed_sparse_matrix")$layout, "CSR")
    expect_identical(as.matrix(readObject(tmp)), core)
})

test_that("writing an SVT matrix works with empty or near-empty storage", {
    for (it in 1:3) {
        if (it == 1L) {
            # Near-empty, with small integers.
            x <- rsparsematrix(100, 20, 0.001) * 10
        } else if (it == 2L) {
            # Near-empty, with larger integers. This checks that the type
            # optimization doesn't get screwed up by Infs after performing
            # range() on an empty vector.
            x <- rsparsematrix(100, 20, 0.001) * 1e6
        } else {
            # Actually empty.
            x <- rsparsematrix(100, 20, 0)
        }
        y <- as(x, "SVT_SparseMatrix")
        core <- as.matrix(x)

        tmp <- tempfile()
        saveObject(y, tmp)
        expect_identical(as.matrix(readObject(tmp)), core)

        # Now integers:
        y2 <- y
        type(y2) <- "integer"
        core2 <- core
        storage.mode(core2) <- "integer"

        tmp <- tempfile()
        saveObject(y2, tmp)
        expect_identical(as.matrix(readObject(tmp)), core2)

        # Now booleans:
        y2 <- y
        type(y2) <- "logical"
        core2 <- core
        storage.mode(core2) <- "logical"

        tmp <- tempfile()
        saveObject(y2, tmp)
        expect_identical(as.matrix(readObject(tmp)), core2)
    }
})

test_that("writing a sparse matrix works when the entire matrix consists of NAs", {
    x <- rsparsematrix(100, 20, 0)
    x@x[] <- NA # all NA's.
    y <- as(x, "SVT_SparseMatrix")
    core <- as.matrix(x)

    tmp <- tempfile()
    saveObject(y, tmp)
    expect_identical(as.matrix(readObject(tmp)), core)

    # Now integers:
    y2 <- y
    type(y2) <- "integer"
    core2 <- core
    storage.mode(core2) <- "integer"

    tmp <- tempfile()
    saveObject(y2, tmp)
    expect_identical(as.matrix(readObject(tmp)), core2)

    # Now booleans:
    y2 <- y
    type(y2) <- "logical"
    core2 <- core
    storage.mode(core2) <- "logical"

    tmp <- tempfile()
    saveObject(y2, tmp)
    expect_identical(as.matrix(readObject(tmp)), core2)
})

test_that("saving sparse matrices works with dimnames", {
    x <- rsparsematrix(100, 20, 0.5)
    rownames(x) <- paste0("GENE_", seq_len(nrow(x)))
    colnames(x) <- head(LETTERS, 20)

    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as.matrix(roundtrip), as.matrix(x))
})

test_that("saveObject diverts correctly with pristine sparse DelayedArrays", {
    x <- DelayedArray(Matrix::rsparsematrix(100, 20, 0.2))
    expect_true(isPristine(x))

    tmp <- tempfile()
    saveObject(x, tmp)
    expect_identical(as(readObject(tmp), "dgCMatrix"), x@seed)

    # Also works if we ignore the pristine-ness.
    tmp <- tempfile()
    saveObject(x, tmp, DelayedArray.dispatch.pristine=FALSE)
    expect_identical(as(readObject(tmp), "dgCMatrix"), x@seed)
})

test_that("reading sparse arrays work with non-default NA placeholders", {
    x <- rsparsematrix(100, 20, 0.5)
    first <- x@x[1]

    dir <- tempfile()
    saveObject(x, dir)

    library(rhdf5)
    local({ 
        fhandle <- H5Fopen(file.path(dir, "matrix.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, "compressed_sparse_matrix")
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
        dhandle <- H5Dopen(ghandle, "data")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        alabaster.base::h5_write_attribute(dhandle, "missing-value-placeholder", first, scalar=TRUE)
    })

    arr2 <- readObject(dir)
    ref <- as.matrix(x)
    ref[ref==first] <- NA
    expect_identical(ref, as.matrix(arr2))
})

test_that("saveObject deduplicates sparse arrays correctly", {
    skip_on_os("windows")

    mat <- rsparsematrix(100, 20, 0.5)
    tmp <- tempfile()
    session <- createDedupSession()

    saveObject(mat, tmp, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp)), as.matrix(mat))

    tmp2 <- tempfile()
    saveObject(mat, tmp2, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp2)), as.matrix(mat))
    expect_true(Sys.readlink(file.path(tmp2, "OBJECT")) != "") # Check that a symlink is actually formed.

    # Check that we get the same result from a pristine DelayedArray.
    tmp3 <- tempfile()
    saveObject(DelayedArray(mat), tmp3, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp3)), as.matrix(mat))
    expect_true(Sys.readlink(file.path(tmp3, "OBJECT")) != "")

    # Deduplication works for DelayedArrays wrapping dense matrices.
    tmp4a <- tempfile()
    delayed <- DelayedArray(mat) * 2
    saveObject(delayed, tmp4a, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp4a)), unname(as.matrix(delayed)))
    expect_true(Sys.readlink(file.path(tmp4a, "OBJECT")) == "")

    tmp4b <- tempfile()
    saveObject(delayed, tmp4b, DelayedArray.dispatch.pristine=FALSE, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp4b)), unname(as.matrix(delayed)))
    expect_true(Sys.readlink(file.path(tmp4b, "OBJECT")) != "")
})

test_that("saveObject works correctly with empty matrices", {
    x <- Matrix::rsparsematrix(0, 20, 0.1)
    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as(roundtrip, "CsparseMatrix"), x)

    x2 <- DelayedArray(x) * 1
    tmp <- tempfile()
    saveObject(x2, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as(roundtrip, "CsparseMatrix"), as(x2, "CsparseMatrix"))
})

#' Save a multi-dimensional array to disk
#'
#' Save a high-dimensional array to its on-disk representations.
#'
#' @param x An integer, numeric, logical or character array.
#' Alternatively, any of the \linkS4class{denseMatrix} subclasses from the \pkg{Matrix} package.
#' @param path String containing the path to a directory in which to save \code{x}.
#' @param array.dedup.session A session object created by \code{\link[alabaster.base]{createDedupSession}},
#' specifying which objects should be deduplicated if the same \code{x} is encountered multiple times.
#' @param array.dedup.action String specifying how deduplication should occur, see options for the \code{action=} argument in \code{\link[alabaster.base]{cloneDirectory}}.
#' @param array.character.vls Logical scalar indicating whether to save character arrays in the custom variable length string (VLS) array format.
#' If \code{NULL}, this is determined based on a comparison of the expected storage against a fixed length array.
#' @param ... Further arguments, currently ignored.
#' 
#' @details
#' When \code{saveObject} is called multiple times on the same \code{x}, this method can avoid re-saving the array if a user supplies a deduplication session in \code{array.dedup.session=}.
#' Instead, the method will link or copy (depending on the choice of \code{array.dedup.action=}) the files produced by the first \code{saveObject} call to the new \code{path=}.
#' This saves time and reduces disk usage, and is particularly useful when saving complex data structures like a SummarizedExperiment.
#' For example, if we have a SummarizedExperiment that contains multiple copies of the same array, we can pass a \code{array.dedup.session=} to the \code{saveObject} call on the SummarizedExperiment.
#' This will instruct the internal \code{saveObject} calls to only write the array to disk once and subsequently make copies or links for all duplicates of that array.
#' The same approach can be applied to deduplicate seeds in a DelayedArray, see \code{?"\link{storeDelayedObject}"} for details.
#'
#' @return
#' \code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
#'
#' @seealso
#' \code{\link{readArray}}, to read the directory contents back into the R session.
#'
#' @author Aaron Lun
#' @examples
#' mat <- array(rpois(10000, 10), c(50, 20, 10))
#' dimnames(mat) <- list(
#'    paste0("GENE_", seq_len(nrow(mat))),
#'    letters[1:20],
#'    NULL
#' )
#'
#' dir <- tempfile()
#' saveObject(mat, dir)
#' list.files(dir)
#'
#' @name saveArray
#' @aliases 
#' stageObject,array-method
#' stageObject,Matrix-method
NULL

#' @import alabaster.base rhdf5
.save_array <- function(x, path, extract.native=NULL, array.dedup.session=NULL, array.dedup.action="link", array.character.vls=FALSE, ...) {
    if (!is.null(array.dedup.session)) {
        dedup.path <- checkObjectInDedupSession(x, array.dedup.session)
        if (!is.null(dedup.path)) {
            cloneDirectory(dedup.path, path, array.dedup.action)
            return(invisible(NULL))
        }
        addObjectToDedupSession(x, array.dedup.session, path)
    }

    dir.create(path)
    fpath <- file.path(path, "array.h5")
    name <- "dense_array"

    fhandle <- H5Fcreate(fpath, "H5F_ACC_TRUNC")
    on.exit(.H5Fclose_null(fhandle), add=TRUE, after=FALSE)
    ghandle <- H5Gcreate(fhandle, name)
    on.exit(.H5Gclose_null(ghandle), add=TRUE, after=FALSE)

    optimized <- optimize_storage(x)

    # Only enabling VLS for character matrices, to keep things simple.
    saved.vls <- FALSE
    if (is.character(x) && !isFALSE(array.character.vls)) {
        if (is.null(array.character.vls)) {
            N <- length(x)
            array.character.vls <- (optimized$total_len + 16 * N < optimized$max_len * N)
        }
        if (array.character.vls) {
            ghandle <- .H5Gclose_null(ghandle)
            fhandle <- .H5Fclose_null(fhandle)
            saved.vls <- TRUE
            h5_write_vls_array(
                fpath,
                name,
                "pointers",
                "heap",
                x,
                compress=getHDF5DumpCompressionLevel(),  
                chunks=getHDF5DumpChunkDim(dim(x))
            )
            fhandle <- H5Fopen(fpath)
            ghandle <- H5Gopen(fhandle, name)
            h5_write_attribute(ghandle, "type", "vls", scalar=TRUE)
        }
    }

    if (saved.vls) {
        dhandle <- H5Dopen(ghandle, "pointers")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    } else {
        h5_write_array(
            ghandle, 
            name="data", 
            x=x, 
            type=optimized$type, 
            placeholder=optimized$placeholder, 
            extract.native=extract.native
        )
        h5_write_attribute(ghandle, "type", array_type(x), scalar=TRUE)
        dhandle <- H5Dopen(ghandle, "data")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    }

    if (!is.null(optimized$placeholder)) {
        h5_write_attribute(dhandle, missingPlaceholderName, optimized$placeholder, type=optimized$type, scalar=TRUE)
    }

    h5_write_attribute(ghandle, "transposed", 1L, scalar=TRUE)
    save_names(ghandle, x, transpose=TRUE)
    saveObjectFile(path, name, list(dense_array=list(version="1.1")))
    invisible(NULL)
}

#' @export
#' @rdname saveArray
setMethod("saveObject", "array", function(x, path, array.dedup.session=NULL, array.dedup.action="link", array.character.vls=FALSE, ...) {
    .save_array(x, path, extract.native=identity, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, array.character.vls=array.character.vls, ...)
})

#' @export
#' @rdname saveArray
setMethod("saveObject", "denseMatrix", function(x, path, array.dedup.session=NULL, array.dedup.action="link", ...) {
    extract.native <- NULL
    if (is(x, "dMatrix") || is(x, "lMatrix")) {
        extract.native <- function(y) y@x
    }
    .save_array(x, path, extract.native=extract.native, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, ...)
})

##############################
######### INTERNALS ##########
##############################

#' @importFrom DelayedArray RegularArrayGrid
create_output_grid <- function(dims, chunks, max.size) {
    spacings <- chunks
    for (d in seq_along(spacings)) {
        leftover <- max.size / prod(spacings)
        num.chunks <- dims[d] / spacings[d]
        if (leftover >= num.chunks) {
            spacings[d] <- dims[d]
        } else if (leftover >= 2) {
            spacings[d] <- as.integer(leftover) * spacings[d]
        } else {
            break
        } 
    }
    RegularArrayGrid(dims, spacings=spacings)
}

#' @importFrom HDF5Array getHDF5DumpCompressionLevel getHDF5DumpChunkDim
#' @importFrom DelayedArray currentViewport blockApply getAutoBlockLength
#' @importFrom BiocGenerics start
h5_write_array <- function(handle, name, x, type, placeholder, extract.native=NULL, compress=getHDF5DumpCompressionLevel(), chunks=NULL) {
    shandle <- H5Screate_simple(dim(x))
    on.exit(H5Sclose(shandle), add=TRUE, after=FALSE)

    phandle <- H5Pcreate("H5P_DATASET_CREATE")
    on.exit(H5Pclose(phandle), add=TRUE, after=FALSE)
    H5Pset_fill_time(phandle, "H5D_FILL_TIME_NEVER")
    H5Pset_obj_track_times(phandle, FALSE)

    if (!is.null(compress) && compress > 0 && length(x)) {
        H5Pset_shuffle(phandle) # improve compression by interlacing data elements.
        H5Pset_deflate(phandle, level=compress)
        if (is.null(chunks)) {
            chunks <- getHDF5DumpChunkDim(dim(x))
        }
        H5Pset_chunk(phandle, chunks)
    } else {
        # We'll treat each stretch of the first dimension as a 'chunk' to
        # encourage the block processing to behave contiguously.
        chunks <- rep(1L, length(dim(x)))
        chunks[1] <- dim(x)[1]
    }

    dhandle <- H5Dcreate(handle, name, dtype_id=type, h5space=shandle, dcpl=phandle)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)

    if (is.null(placeholder) && !is.null(extract.native)) {
        # Writing is done from fastest dimension in R to fastest dimension in HDF5,
        # so the transposition is implicit.
        H5Dwrite(dhandle, extract.native(x))

    } else {
        # We want to iterate using the OUTPUT chunk dimensions, not the input
        # block size. This is because we KNOW that the output is a HDF5 handle
        # where it is expensive to cross chunk boundaries, whereas the input
        # may or may not have such issues.
        grid <- create_output_grid(dim(x), chunks, getAutoBlockLength(type(x)))

        blockApply(x, function(y) {
            if (!is.null(placeholder) && anyMissing(y)) {
                y[is.missing(y)] <- placeholder
            }
            mem_shandle <- H5Screate_simple(dim(y))
            on.exit(H5Sclose(mem_shandle), add=TRUE, after=FALSE)
            view <- currentViewport()
            H5Sselect_hyperslab(shandle, "H5S_SELECT_SET", start=start(view), count=dim(view))
            H5Dwrite(dhandle, y, h5spaceMem=mem_shandle, h5spaceFile=shandle)
        }, grid=grid)
    }
}

##############################
######### OLD STUFF ##########
##############################

#' @importFrom S4Arrays is_sparse
#' @importFrom rhdf5 h5createFile 
#' @importFrom HDF5Array writeHDF5Array
#' @importFrom alabaster.base transformVectorForHdf5 addMissingPlaceholderAttributeForHdf5
.stage_array <- function(x, dir, path, child=FALSE, .version=2) {
    dir.create(file.path(dir, path), showWarnings=FALSE)
    xpath <- paste0(path, "/array.h5")
    ofile <- file.path(dir, xpath)

    transformed <- transformVectorForHdf5(x)

    h5createFile(ofile)
    writeHDF5Array(transformed$transformed, filepath=ofile, name="data")
    nm <- .name_saver(x, ofile)

    if (!is.null(transformed$placeholder)) {
        addMissingPlaceholderAttributeForHdf5(ofile, "data", transformed$placeholder)
    }

    list(
        `$schema` = "hdf5_dense_array/v1.json",
        path = xpath,
        is_child = child,
        `array` = .grab_array_type(x),
        hdf5_dense_array = list(
            dataset = "data",
            dimnames = nm,
            version = 2
        )
    )
}

.grab_array_type <- function(x) {
    list(
        dimensions=I(dim(x)),
        type=array_type(x)
    )
}

#' @importFrom rhdf5 h5createGroup h5write
.name_saver <- function(x, path, group = "names") {
    # Chucking in some names. 
    if (!is.null(dimnames(x))) {
        h5createGroup(path, group)
        d <- dimnames(x)
        for (i in seq_along(d)) {
            current <- d[[i]]
            if (!is.null(current)) {
                h5write(current, file=path, name=paste0(group, "/", i - 1L)) # 0-based
            }
        }
        return(group)
    } else {
        return(NULL)
    }
}

#' @export
setMethod("stageObject", "array", function(x, dir, path, child=FALSE) .stage_array(x, dir, path, child=child))

#' @importFrom alabaster.base .stageObject
.stage_delayed <- function(x, dir, path, child, fallback) {
    out <- .check_for_hdf5(x, dir, path, child = child)
    if (!is.null(out)) {
        return(out)
    }

    if (!preserveDelayedOperations()) {
        return(fallback(x, dir, path, child=child))
    }

    dir.create(file.path(dir, path), showWarnings=FALSE)
    xpath <- paste0(path, "/delayed.h5")
    chihaya::saveDelayed(x, file.path(dir, xpath), "data")

    list(
        `$schema`="hdf5_delayed_array/v1.json",
        path=xpath,
        is_child=child,
        `array` = .grab_array_type(x),
        hdf5_delayed_array= list(
            group="data"
        )
    )
}

#' @export
setMethod("stageObject", "DelayedArray", function(x, dir, path, child=FALSE) .stage_delayed(x, dir, path, child = child, fallback = .stage_array))

#' @importFrom rhdf5 h5createFile h5createGroup
#' @importFrom HDF5Array writeHDF5Array
.stage_sparse_matrix <- function(x, dir, path, child=FALSE) {
    dir.create(file.path(dir, path), showWarnings=FALSE)
    xpath <- paste0(path, "/matrix.h5")
    ofile <- file.path(dir, xpath)

    h5createFile(ofile)
    writeSparseMatrix(x, ofile, name="sparse", column=TRUE, tenx=TRUE)
    nm <- .name_saver(x, ofile)

    list(
        `$schema`="hdf5_sparse_matrix/v1.json",
        path=xpath,
        is_child=child,
        `array` = .grab_array_type(x),
        hdf5_sparse_matrix = list(
            group = "sparse",
            format = "tenx_matrix",
            dimnames = nm,
            version = 2 
        )
    )
}

#' @importFrom S4Arrays is_sparse
.stage_any_matrix <- function(x, dir, path, child=FALSE) {
    if (is_sparse(x)) {
        .stage_sparse_matrix(x, dir, path, child=child)
    } else {
        .stage_array(x, dir, path, child=child)
    }
}

#' @export
#' @importClassesFrom Matrix Matrix
setMethod("stageObject", "Matrix", function(x, dir, path, child=FALSE) .stage_any_matrix(x, dir, path, child=child))

#' @export
setMethod("stageObject", "DelayedMatrix", function(x, dir, path, child=FALSE) .stage_delayed(x, dir, path, child = child, fallback = .stage_any_matrix))

.link_or_copy <- function(from, to) {
    if (!file.link(from, to)) {
        if (!file.copy(from, to)) {
            stop("failed to link or copy '", from, "' to '", to, "'")
        }
    }
}

#' @importFrom BiocGenerics path
#' @importFrom rhdf5 h5read
.check_for_hdf5 <- function(x, dir, path, child) {
    if (!recycleHdf5Files()) {
        return(NULL)
    }

    if (is(x, "DelayedArray")) {
        x <- x@seed
    }
    if (is(x, "DelayedSetDimnames")) {
        x <- x@seed # Ignoring the names.
    }

    dense <- FALSE
    if (is(x, "HDF5ArraySeed")) {
        dense <- TRUE
    } else if (is(x, "H5SparseMatrixSeed")) {
        ;
    } else {
        return(NULL)
    }

    src <- path(x)

    if (!dense) {
        # Checking that it's in the 10X format, otherwise we bail.
        if (!is(x, "CSC_H5SparseMatrixSeed")) {
            return(NULL)
        }
        if (!is.null(x@subdata)) {
            return(NULL)
        }

        attempt <- try(h5read(src, paste0(x@group, "/shape")), silent=TRUE)
        if (is(attempt, "try-error") || !identical(as.vector(attempt), dim(x))) {
            return(NULL)
        }

        dir.create(file.path(dir, path), showWarnings=FALSE)
        dest <- paste0(path, "/matrix.h5")
        .link_or_copy(src, file.path(dir, dest))

        list(
            `$schema`="hdf5_sparse_matrix/v1.json",
            path=dest,
            is_child=child,
            `array` = .grab_array_type(x),
            hdf5_sparse_matrix = list(
                group = x@group,
                format = "tenx_matrix",
                version = 2
            )
        )

    } else {
        dir.create(file.path(dir, path), showWarnings=FALSE)
        dest <- paste0(path, "/array.h5")
        .link_or_copy(src, file.path(dir, dest))

        list(
            `$schema`="hdf5_dense_array/v1.json",
            path=dest,
            is_child=child,
            `array` = .grab_array_type(x),
            hdf5_dense_array= list(
                dataset=x@name,
                version = 2
            )
        )
    }
}

# This tests the saveObject generic for arrays.
# library(testthat); library(alabaster.matrix); source("setup.R"); source("test-array.R")

library(DelayedArray)
experiment <- "rnaseq"
assay <- "counts"

arr <- array(rpois(10000, 10), c(50, 20, 10))
dimnames(arr) <- list(
   paste0("GENE_", seq_len(nrow(arr))),
   letters[1:20],
   NULL
)

test_that("saveObject works as expected", {
    tmp <- tempfile()
    saveObject(arr, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(normalizePath(BiocGenerics::path(roundtrip)), normalizePath(tmp))
    expect_identical(as.array(roundtrip), arr)

    # Works when it's officially integer.
    copy <- arr
    storage.mode(copy) <- "integer"
    tmp <- tempfile()
    saveObject(copy, tmp)
    expect_identical(as.array(readObject(tmp)), copy)

    # Works without dimnames.
    copy <- arr
    dimnames(copy) <- NULL
    tmp <- tempfile()
    saveObject(copy, tmp)
    expect_identical(as.array(readObject(tmp)), copy)
})

test_that("saveObject type optimization works as expected for integers", {
    # Small unsigned integers
    mat <- matrix(sample(255, 1000, replace=TRUE), 40, 25)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    storage.mode(mat) <- "integer"
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    # Small signed integers
    mat <- matrix(sample(255, 1000, replace=TRUE) - 128, 40, 25)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    storage.mode(mat) <- "integer"
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    # Large integers
    mat <- trunc(matrix(runif(1000, -1e6, 1e6), 40, 25))
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    storage.mode(mat) <- "integer"
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)
})

test_that("saveObject works as expected for floats", {
    mat <- matrix(rnorm(1000, -1e6, 1e6), 40, 25)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[101] <- NaN
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[102] <- Inf
    mat[103] <- -Inf
    tmp <- tempfile()
    saveObject(arr, tmp)
    expect_identical(as.array(readObject(tmp)), arr)
})

test_that("saveObject works as expected for logicals", {
    mat <- matrix(rbinom(1000, 1, 0.5) == 1, 40, 25)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)
})

test_that("saveObject works as expected for strings", {
    mat <- matrix(sample(LETTERS, 1000, replace=TRUE), 40, 25)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    mat[100] <- NA
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)
})

test_that("saveObject works as expected for all-NA arrays", {
    mat <- matrix(NA, 50, 100)
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    storage.mode(mat) <- "integer"
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)

    storage.mode(mat) <- "double"
    tmp <- tempfile()
    saveObject(mat, tmp)
    expect_identical(as.matrix(readObject(tmp)), mat)
})

test_that("saveObject diverts correctly with pristine dense DelayedArrays", {
    x <- DelayedArray(arr)
    expect_true(isPristine(x))

    tmp <- tempfile()
    saveObject(x, tmp)
    expect_identical(as.array(readObject(tmp)), x@seed)

    # Also works if we ignore the pristine-ness.
    tmp <- tempfile()
    saveObject(x, tmp, DelayedArray.dispatch.pristine=FALSE)
    expect_identical(as.array(readObject(tmp)), x@seed)

    # Works with a custom override.
    old <- altSaveObjectFunction(function(...) {
        print("YAY")
        saveObject(...)
    })
    on.exit(altSaveObjectFunction(old), add=TRUE, after=FALSE)
    tmp <- tempfile()
    expect_output(saveObject(x, tmp), "YAY")
    expect_identical(as.array(readObject(tmp)), x@seed)

    altSaveObjectFunction(function(...) {
        stop("foobar")
    })
    tmp <- tempfile()
    expect_warning(saveObject(x, tmp), "foobar")
    expect_identical(as.array(readObject(tmp)), x@seed)

    setGeneric("stupidSave", function(x, path, ...) standardGeneric("stupidSave"))
    altSaveObjectFunction(stupidSave)
    tmp <- tempfile()
    expect_warning(saveObject(x, tmp), NA)
    expect_identical(as.array(readObject(tmp)), x@seed)
})

test_that("saveObject works correctly for unknown dense DelayedArrays", {
    x <- DelayedArray(new("SuperSeed", dim=c(100L, 50L)))
    tmp <- tempfile()
    saveObject(x, tmp)
    expect_identical(as.array(readObject(tmp)), as.matrix(x))
})

test_that("saveObject deduplicates dense arrays correctly", {
    skip_on_os("windows")

    mat <- matrix(runif(5000), 50, 100)
    tmp <- tempfile()
    session <- createDedupSession()

    saveObject(mat, tmp, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp)), mat)

    tmp2 <- tempfile()
    saveObject(mat, tmp2, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp2)), mat)
    expect_true(Sys.readlink(file.path(tmp2, "OBJECT")) != "") # Check that a symlink is actually formed.

    # Check that we get the same result from a pristine DelayedArray.
    tmp3 <- tempfile()
    saveObject(DelayedArray(mat), tmp3, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp3)), mat)
    expect_true(Sys.readlink(file.path(tmp3, "OBJECT")) != "")

    # Deduplication works for DelayedArrays wrapping dense matrices.
    tmp4a <- tempfile()
    delayed <- DelayedArray(mat) * 2
    saveObject(delayed, tmp4a, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp4a)), as.matrix(delayed))
    expect_true(Sys.readlink(file.path(tmp4a, "OBJECT")) == "")

    tmp4b <- tempfile()
    saveObject(delayed, tmp4b, DelayedArray.dispatch.pristine=FALSE, array.dedup.session=session, array.dedup.action="symlink")
    expect_identical(as.matrix(readObject(tmp4b)), as.matrix(delayed))
    expect_true(Sys.readlink(file.path(tmp4b, "OBJECT")) != "")
})

test_that("saveObject works correctly with dense block processing", {
    x <- DelayedArray(arr) * 1L
    expect_false(isPristine(x))

    for (bs in c(100, 200, 1000, 2000)) {
        tmp <- tempfile(fileext=".h5")
        local({
            oldh <- HDF5Array::getHDF5DumpChunkLength()
            olds <- getAutoBlockSize()
            HDF5Array::setHDF5DumpChunkLength(bs)
            setAutoBlockSize(bs * 4L)
            on.exit(HDF5Array::setHDF5DumpChunkLength(oldh))
            on.exit(setAutoBlockSize(olds), add=TRUE)
            saveObject(x, tmp)
        })

        roundtrip <- readObject(tmp)
        expect_identical(as.array(roundtrip), arr)
    }
})

test_that("reading dense arrays work with non-default NA placeholders", {
    tmp <- tempfile()
    saveObject(arr, tmp)
    first <- arr[1]

    library(rhdf5)
    local({ 
        fhandle <- H5Fopen(file.path(tmp, "array.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, "dense_array")
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
        dhandle <- H5Dopen(ghandle, "data")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        alabaster.base::h5_write_attribute(dhandle, "missing-value-placeholder", first, scalar=TRUE)
    })

    ref <- arr
    ref[ref == first] <- NA
    expect_identical(ref, as.array(readObject(tmp)))
})

test_that("output grid creation works as expected", {
    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(5L, 4L), 20)
    expect_identical(grid@refdim, c(100L, 20L))
    expect_identical(grid@spacings, c(5L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(5L, 4L), 30)
    expect_identical(grid@spacings, c(5L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(5L, 4L), 60)
    expect_identical(grid@spacings, c(15L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(5L, 4L), 400)
    expect_identical(grid@spacings, c(100L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(5L, 4L), 800)
    expect_identical(grid@spacings, c(100L, 8L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(7L, 4L), 200)
    expect_identical(grid@spacings, c(49L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(7L, 4L), 400)
    expect_identical(grid@spacings, c(100L, 4L))

    grid <- alabaster.matrix:::create_output_grid(c(100L, 20L), c(7L, 3L), 30000)
    expect_identical(grid@spacings, c(100L, 20L))
})

test_that("saveObject works correctly with empty arrays", {
    x <- matrix(1, 100, 0)
    tmp <- tempfile()
    saveObject(x, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as.array(roundtrip), x)

    x2 <- DelayedArray(x) * 1
    tmp <- tempfile()
    saveObject(x2, tmp)
    roundtrip <- readObject(tmp)
    expect_identical(as.array(roundtrip), as.matrix(x2))
})

test_that("saveObject works correctly with VLS arrays", {
    tmp <- tempfile()
    dir.create(tmp)

    NR <- 10
    NC <- 12
    x <- matrix(sample(c("A", "BC", "DEFG", "HIJKL", "MNOP", "QRS", "TU", "V"), NR * NC, replace=TRUE), NR, NC)

    saveObject(x, file.path(tmp, "basic"), array.character.vls=TRUE)
    reloaded <- readObject(file.path(tmp, "basic")) 
    expect_identical(x, as.matrix(reloaded))

    y <- x
    y[2] <- NA
    saveObject(y, file.path(tmp, "with_missing"), array.character.vls=TRUE)
    reloaded <- readObject(file.path(tmp, "with_missing")) 
    expect_identical(y, as.matrix(reloaded))

    y <- x
    dimnames(y) <- list(letters[1:NR], seq_len(NC))
    saveObject(y, file.path(tmp, "named"), array.character.vls=TRUE)
    reloaded <- readObject(file.path(tmp, "named")) 
    expect_identical(y, as.matrix(reloaded))

    y <- x
    y[4] <- strrep("HIJKL", 100)
    saveObject(y, file.path(tmp, "auto"), array.character.vls=NULL)
    reloaded <- readObject(file.path(tmp, "auto")) 
    expect_identical(y, as.matrix(reloaded))
})

#' Reloaded \pkg{alabaster} array 
#'
#' An array that was reloaded from disk by the \code{\link{readObject}} function.
#' This allows methods to refer to the existing on-disk representation by inspecting the \code{path}.
#' For example, \code{\link{saveObject}} can just copy/link to the files instead of repeating the saving process.
#'
#' @param path String containing a path to the directory with the on-disk array representation.
#' Alternatively an existing ReloadedArraySeed, which is returned without modification.
#' @param seed Contents of the loaded array, e.g., as an ordinary R array, a \linkS4class{DelayedArray} or a sparse matrix.
#' If \code{NULL}, this is obtained by calling \code{\link{readObject}}.
#' @param ... Further arguments to pass to \code{\link{readObject}} when \code{seed=NULL}.
#'
#' @return
#' For the constructors, an instance of the \linkS4class{ReloadedArraySeed} or \linkS4class{ReloadedArray}.
#'
#' @details
#' The ReloadedArraySeed is a \linkS4class{DelayedUnaryIsoOp} subclass that will just forward all operations to the underlying \code{seed}.
#' Its main purpose is to track the \code{path} that was originally used to generate \code{seed}, which enables optimizations for methods that need to operate on the files.
#'
#' One obvious optimization is the specialization of \code{\link{saveObject}} on ReloadedArray instances.
#' Instead of loading the array data back into the R session and saving it again, the \code{saveObject} method can just link or copy the existing files.
#' This behavior is controlled by the \code{ReloadedArray.reuse.files=} option in the \code{saveObject} method,
#' which can be any of the choices for \code{action=} in \code{\link[alabaster.base]{cloneDirectory}}.
#' It may also be \code{"none"} to ignore existing files and just save the contents by calling \code{"\link{saveObject,DelayedArray-method}"}.
#' 
#' @examples
#' arr <- array(rpois(10000, 10), c(50, 20, 10))
#' dir <- tempfile()
#' saveObject(arr, dir)
#' obj <- readArray(dir)
#' obj
#' DelayedArray::showtree(obj)
#' 
#' @name ReloadedArraySeed
#' @aliases
#' ReloadedArraySeed-class
#' ReloadedArray-class
#' ReloadedMatrix-class
#' DelayedArray,ReloadedArraySeed-method
#' matrixClass,ReloadedArray-method
#' coerce,ReloadedArray,ReloadedMatrix-method
#' coerce,ReloadedMatrix,ReloadedArray-method
#' path,ReloadedArraySeed-method
#' saveObject,ReloadedArray-method
#'
#' @export
ReloadedArraySeed <- function(path, seed=NULL, ...) {
    if (is(path, "ReloadedArraySeed")) {
        return(path)
    }
    if (is.null(seed)) {
        seed <- readObject(path, ...)
    }
    while (is(seed, "DelayedArray")) {
        seed <- seed@seed
    }

    # Need to obtain an absolute path in order for this to be safe in
    # saveObject(), possibly after changes to the working directory has.
    new("ReloadedArraySeed", path=normalizePath(path, mustWork=TRUE), seed=seed)
}

#' @export
#' @rdname ReloadedArraySeed
ReloadedArray <- function(path, seed=NULL, ...) {
    DelayedArray(ReloadedArraySeed(path, seed=seed, ...))
}

#' @export
#' @importFrom DelayedArray DelayedArray new_DelayedArray
setMethod("DelayedArray", "ReloadedArraySeed", function(seed) new_DelayedArray(seed, Class="ReloadedArray"))

#' @export
#' @importFrom DelayedArray matrixClass
setMethod("matrixClass", "ReloadedArray", function(x) "ReloadedMatrix")

# Overrides copied from DelayedArray::ConstantArray.
#' @importFrom S4Vectors new2
setAs("ReloadedArray", "ReloadedMatrix", function(from) new2("ReloadedMatrix", from))
setAs("ReloadedMatrix", "ReloadedArray", function(from) from)

#' @export
setMethod("path", "ReloadedArraySeed", function(object, ...) object@path)

#' @export
setMethod("saveObject", "ReloadedArray", function(x, path, ReloadedArray.reuse.files="link", ...) {
    s <- x@seed
    if (ReloadedArray.reuse.files == "none") {
        x <- DelayedArray(s@seed)
        return(saveObject(x, path, ReloadedArray.reuse.files=ReloadedArray.reuse.files, ...))
    } 
    cloneDirectory(s@path, path, action=ReloadedArray.reuse.files)
    invisible(NULL)
})

#' DelayedArray wrapper seed
#'
#' Deprecated, use \link[DelayedArray]{DelayedUnaryIsoOp} instead.
#'
#' @aliases
#' WrapperArraySeed-class
#' WrapperArray-class
#' WrapperArraySeed
#' WrapperArray
#' loadWrapperArray
#'
#' @name WrapperArraySeed
NULL

#' @importFrom BiocGenerics type
to_array_type <- function(x) {
    switch(type(x),
        integer="integer",
        double="number",
        logical="boolean",
        character="string",
        "other"
    )
}

array_type <- to_array_type

from_array_type <- function(x) {
    switch(x, 
        integer="integer",
        number="double",
        boolean="logical",
        string="character"
    )
}

save_names <- function(handle, x, group = "names", transpose=FALSE) {
    d <- dimnames(x)
    if (is.null(d) || all(vapply(d, is.null, TRUE))) {
        return(NULL)
    }

    if (transpose) { # for the HDF5 array transposition.
        d <- rev(d)
    }

    ghandle <- H5Gcreate(handle, group)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    for (i in seq_along(d)) {
        current <- d[[i]]
        if (!is.null(current)) {
            h5_write_vector(ghandle, as.character(i - 1L), current)
        }
    }
}

load_names <- function(handle, ndim, group = "names") {
    if (!h5_object_exists(handle, group)) {
        return(NULL)
    }

    ghandle <- H5Gopen(handle, group)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    all.named <- h5ls(ghandle, datasetinfo=FALSE, recursive=FALSE)

    names <- vector("list", ndim)
    for (y in all.named$name) {
        names[[as.integer(y) + 1L]] <- h5_read_vector(ghandle, y)
    }

    names
}

#' @importFrom alabaster.base saveObject
try_altSaveObject <- function(x, ...) {
    if (is.null(altSaveObjectFunction())) {
        fun <- selectMethod(saveObject, signature=class(x), optional=TRUE)
        if (!is.null(fun)) {
            fun(x, ...)
            return(TRUE)
        }
        return(FALSE)
    }

    # We can't use selectMethod() here as we don't know what generic system
    # is being used by altSaveObject, so we just have to try it out.
    fail <- try(altSaveObject(x, ...), silent=TRUE)
    if (!is(fail, "try-error")) {
        return(TRUE)
    }

    # If the failure wasn't due to S4 dispatch, we emit a warning.
    msg <- attr(fail, "condition")
    if (!startsWith(msg$message, "unable to find an inherited method")) {
        warning(msg$message)
    }
    return(FALSE)
}

#' @import rhdf5
.H5Fclose_null <- function(handle) {
    if (!is.null(handle)) {
        H5Fclose(handle)
    }
    NULL
}

#' @import rhdf5
.H5Gclose_null <- function(handle) {
    if (!is.null(handle)) {
        H5Gclose(handle)
    }
    NULL
}

#' Store/reload a DelayedArray
#'
#' Store or reload the delayed operations or array-like seeds of a \linkS4class{DelayedArray} in an existing HDF5 file.
#'
#' @param x Any of the delayed operation/array classes from \pkg{DelayedArray}.
#' @param handle An \pkg{rhdf5} handle of a HDF5 file to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).
#' @param name String containing the name of the group in \code{file} to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).
#' @param ... For \code{storeDelayedObject} and \code{reloadDelayedObject}, additional arguments to be passed to specific methods.
#'
#' For \code{altStoreDelayedObject} and \code{altReloadDelayedObject}, arguments to be passed to the alternative functions.
#' @param version Package version of the \pkg{chihaya} format to use when storing or reloading delayed objects.
#' When reloading, the version should be retrieved from the attributes of the outermost group, typically by \code{readDelayedArray}.
#' @param reload Function to reload delayed objects from file.
#' This should accept the same arguments as \code{reloadDelayedObject}.
#' @param store Function (typically a generic) to store delayed objects to file.
#' This should accept the same arguments as \code{storeDelayedObject}.
#' @param type String specifying the type of delayed object, i.e., operation or array.
#' This corresponds to \code{delayed_type} type in the \pkg{chihaya} attributes.
#' @param subtype String specifying the subtype of the delayed object, 
#' This corresponds to \code{delayed_array} or \code{delayed_operation} type (depending on \code{type}) in the \pkg{chihaya} attributes.
#' @param fun Function to reload a delayed object.
#' This should accept the same arguments as \code{reloadDelayedObject} and should return a delayed array (if \code{type="array"}) or operation (otherwise).
#' It may also be \code{NULL} to delete an existing entry in the registry.
#' @param existing Logical scalar indicating the action to take if a function has already been registered for \code{type} and \code{subtype} -
#' keep the old or new function, or throw an error.
#' @param save.external.array Logical scalar indicating whether to save an array-like seed as an external seed,
#' even if a dedicated \code{storeDelayedObject} method is available.
#' @param external.save.args Named list of further arguments to pass to \code{\link{altSaveObject}} when saving an external seed.
#' @param external.dedup.session Deprecated, set \code{external.save.args$array.dedup.session} instead.
#' @param external.dedup.action Deprecated, set \code{external.save.args$array.dedup.action} instead.
#'
#' @section Customization:
#' Developers can easily extend \pkg{alabaster.matrix} to new delayed objects by writing new methods for \code{storeDelayedObject}.
#' Methods should save the contents of the delayed object to the HDF5 file in the \pkg{chihaya} format.
#' Each new store method typically requires a corresponding reloading function to be registered via \code{registerReloadDelayedObjectFunction},
#' so that \code{reloadDelayedObject} knows how to reconstitute the object from file.
#' 
#' Application developers can customize the process of storing/reloading delayed objects
#' by specifying alternative functions in \code{altReloadDelayedObjectFunction} and \code{altStoreDelayedObjectFunction}.
#' For example, if we want to preserve all delayed objects except for \linkS4class{DelayedSubset},
#' we could replace \code{storeDelayedObject} with an \code{altStoreDelayedObject} that realizes any DelayedSubset instance into an ordinary matrix.
#' This is analogous to the overrides for \code{\link{altReadObject}} and \code{\link{altSaveObject}}.
#'
#' Extension developers (i.e., those who write new methods for \code{storeDelayedObject} or new functions for \code{reloadDelayedObject})
#' should generally use \code{altStoreDelayedObject} and \code{altReloadDelayedObject} in their method/funcion bodies.
#' This ensures that any custom overrides specified by application developers are still respected in the extensions to \pkg{alabaster.matrix}.
#'
#' @section External seeds:
#' Whenever \code{\link{storeDelayedObject}} encounters a delayed operation or array-like seed for which it has no methods,
#' the ANY method will save the delayed object as an \dQuote{external seed}.
#' The array is saved via \code{\link{altSaveObject}} into a \code{seeds} directory next to the file associated with \code{handle}.
#' A reference to this external location is then stored in the \code{name} group inside \code{handle}.
#'
#' Users can force this behavior for all array-like seeds by specifying \code{save.external.array=TRUE}.
#' This instructs \code{storeDelayedObject} to save everything as external seeds, including those arrays for which it has methods.
#' Doing so can be beneficial to enable deduplication, e.g., when two delayed arrays perform different operations on the same underlying seed.
#' By saving the seeds externally, file management systems can identify the redundancy to save storage space.
#'
#' Advanced users can explicitly deduplicate external seeds by setting both \code{save.external.array=TRUE} and supplying \code{array.dedup.session=} in \code{external.save.args=}.
#' The \code{array.dedup.session} object is filled up with unique seeds as \code{storeDelayedObject} is called on various DelayedArrays (see \code{?"\link{saveObject,array-method}"} for details).
#' Whenever a duplicate seed is encountered, it is not saved again, but is instead linked or copied from the file path associated with the identical external seed.
#' For example, a new session can be created when saving a SummarizedExperiment to deduplicate seeds across its assays.
#'
#' When external seeds are encountered by \code{reloadDelayedObject}, they are loaded as \linkS4class{ReloadedArray}s (or some variant thereof) by \code{\link{altReadObject}}.
#' Users can forcibly realize the reloaded seed into memory by passing \code{custom.takane.reload=TRUE} in \code{...} for the \code{reloadDelayedObject} call.
#' This is occasionally helpful for providing a more faithful roundtrip from file back into memory. 
#'
#' @return 
#' For \code{storeDelayedObject} and \code{altStoreDelayedObject}, the contents of \code{x} are saved to \code{file}, and \code{NULL} is invisibly returned.
#'
#' For \code{reloadDelayedObject} and \code{altReloadDelayedObject}, a delayed operation or \linkS4class{DelayedArray} is returned.
#'
#' For \code{altStoreDelayedObjectFunction}, the current store function is returned if \code{store} is missing.
#' Otherwise, \code{store} is set as the current store function and the previous store function is returned.
#'
#' For \code{altReloadDelayedObjectFunction}, the current reload function is returned if \code{reload} is missing.
#' Otherwise, \code{reload} is set as the current reload function and the previous reload function is returned.
#' 
#' @author Aaron Lun
#' @examples
#' library(DelayedArray)
#' X <- DelayedArray(matrix(runif(100), ncol=20))
#' Y <- cbind(X, DelayedArray::ConstantArray(value=50, c(5, 10)))
#'
#' library(rhdf5)
#' temp <- tempfile()
#' dir.create(temp)
#'
#' fpath <- file.path(temp, "foo.h5")
#' fhandle <- H5Fcreate(fpath)
#' storeDelayedObject(Y@seed, fhandle, "YAY")
#' rhdf5::h5ls(fhandle)
#' H5Fclose(fhandle)
#'
#' fhandle <- H5Fopen(fpath, "H5F_ACC_RDONLY")
#' reloadDelayedObject(fhandle, "YAY")
#' H5Fclose(fhandle)
#'
#' @seealso
#' \code{\link{saveObject,DelayedArray-method}} and \code{\link{readDelayedArray}}, where these methods are used.
#'
#' \url{https://artifactdb.github.io/chihaya/}, for the file format specification of delayed objects.
#'
#' @aliases
#' storeDelayedObject,ConstantArraySeed-method
#' storeDelayedObject,DelayedAbind-method
#' storeDelayedObject,ANY-method
#' storeDelayedObject,DelayedAperm-method
#' storeDelayedObject,DelayedNaryIsoOp-method
#' storeDelayedObject,DelayedSetDimnames-method
#' storeDelayedObject,DelayedSubassign-method
#' storeDelayedObject,DelayedSubset-method
#' storeDelayedObject,DelayedUnaryIsoOpStack-method
#' storeDelayedObject,DelayedUnaryIsoOpWithArgs-method
#' storeDelayedObject,SVT_SparseMatrix-method
#' storeDelayedObject,array-method
#' storeDelayedObject,denseMatrix-method
#' storeDelayedObject,sparseMatrix-method
#' createExternalSeedDedupSession
#'
#' @name storeDelayedObject
NULL

r2value_type_mapping <- c(logical="BOOLEAN", integer="INTEGER", double="FLOAT", character="STRING")
to_value_type <- function(type) {
    if (!(type %in% names(r2value_type_mapping))) {
        stop("cannot map type '", type, "' to a value type")
    }
    r2value_type_mapping[[type]]
}

value2alabaster_type_mapping <- c(BOOLEAN="boolean", INTEGER="integer", FLOAT="number", STRING="string")
to_alabaster_type <- function(vtype) {
    if (!(vtype %in% names(value2alabaster_type_mapping))) {
        stop("cannot map type '", type, "' to an alabaster type")
    }
    value2alabaster_type_mapping[[vtype]]
}

#' @import alabaster.base rhdf5
load_vector_for_chihaya <- function(handle, name, version) {
    dhandle <- H5Dopen(handle, name)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    contents <- H5Dread(dhandle, drop=TRUE) 
    if (is.raw(contents)) {
        storage.mode(contents) <- "integer"
    }

    type <- h5_read_attribute(dhandle, "type")
    type <- to_alabaster_type(type)
    missing.placeholder <- h5_read_attribute(dhandle, "missing_placeholder", check=TRUE, default=NULL)
    h5_cast(contents, expected.type=type, missing.placeholder=missing.placeholder)
}

#' @import alabaster.base rhdf5
save_vector_for_chihaya <- function(handle, name, x, version, scalar) {
    info <- transformVectorForHdf5(x)

    dtype <- NULL
    rtype <- typeof(x)
    if (rtype == "double") {
        dtype <- "H5T_NATIVE_DOUBLE"
    } else if (rtype == "integer") {
        dtype <- "H5T_NATIVE_INT32"
    } else if (rtype == "logical") {
        dtype <- "H5T_NATIVE_INT8"
    }

    dhandle <- h5_write_vector(handle, name, info$transformed, type=dtype, scalar=scalar, emit=TRUE)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)

    if (!is.null(info$placeholder)) {
        h5_write_attribute(dhandle, "missing_placeholder", info$placeholder, type=dtype, scalar=TRUE)
    }
    h5_write_attribute(dhandle, "type", to_value_type(rtype), scalar=TRUE)

    invisible(NULL)
}

#######################################################
#######################################################

chihaya.registry <- new.env()
chihaya.registry$array <- list()
chihaya.registry$operation <- list()
chihaya.registry$type.hint <- list()

#' @export
#' @rdname storeDelayedObject
reloadDelayedObject <- function(handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gopen(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    output <- NULL
    if (h5_object_exists(ghandle, "_r_type_hint")) {
        hint <- h5_read_vector(ghandle, "_r_type_hint")
        if (hint %in% names(chihaya.registry$type.hint)) {
            FUN <- chihaya.registry$type.hint[[hint]]
            output <- tryCatch(
                FUN(ghandle, version=version, ...),
                error=function(e) NULL
            )
        }
    }

    if (is.null(output)) {
        objtype <- h5_read_attribute(ghandle, "delayed_type")
        if (objtype == "array") {
            arrtype <- h5_read_attribute(ghandle, "delayed_array")
            FUN <- chihaya.registry$array[[arrtype]]
        } else {
            optype <- h5_read_attribute(ghandle, "delayed_operation")
            FUN <- chihaya.registry$operation[[optype]]
        }

        if (is.null(FUN)) {
            stop(objtype, " type '", optype, "' is not yet supported")
        }
        output <- FUN(ghandle, version=version, ...)
    }

    if (!is(output, "DelayedArray")) {
        output <- DelayedArray(output)
    }
    output
}

#' @export
#' @rdname storeDelayedObject
reloadDelayedObjectFunctionRegistry <- function(type=c("operation", "array")) {
    type <- match.arg(type)
    chihaya.registry[[type]]
}

#' @export
#' @rdname storeDelayedObject
registerReloadDelayedObjectFunction <- function(type=c("operation", "array"), subtype, fun, existing=c("old", "new", "error")) {
    type <- match.arg(type)
    if (!is.null(fun)) {
        if (!is.null(chihaya.registry[[type]][[subtype]])) {
            existing <- match.arg(existing)
            if (existing == "old") {
                return(invisible(NULL))
            } else if (existing == "error") {
                stop("function has already been registered for object type '", type, "'")
            }
        }
    }
    chihaya.registry[[type]][[subtype]] <- fun
    invisible(NULL)
}

#######################################################
#######################################################

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "ConstantArraySeed", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "array", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_array", "constant array", scalar=TRUE)
    h5_write_vector(ghandle, "dimensions", dim(x), compress=0, type="H5T_NATIVE_UINT32")
    save_vector_for_chihaya(ghandle, "value", x@value, version=version, scalar=TRUE)
})

#' @import rhdf5 DelayedArray
chihaya.registry$array[["constant array"]] <- function(handle, version, ...) {
    dim <- h5_read_vector(handle, "dimensions")
    val <- load_vector_for_chihaya(handle, "value", version=version)
    ConstantArray(dim, value=val)
}

#######################################################
#######################################################

#' @import rhdf5 alabaster.base
save_dense_array_for_chihaya <- function(x, handle, name, extract.native, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "array", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_array", "dense array", scalar=TRUE)

    optimized <- optimize_storage(x)
    h5_write_array(
        ghandle, 
        name="data", 
        x=x, 
        type=optimized$type, 
        placeholder=optimized$placeholder, 
        extract.native=extract.native
    )

    dhandle <- H5Dopen(ghandle, "data")
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    h5_write_attribute(dhandle, "type", to_value_type(type(x)), scalar=TRUE)
    if (!is.null(optimized$placeholder)) {
        h5_write_attribute(dhandle, "missing_placeholder", optimized$placeholder, type=optimized$type, scalar=TRUE)
    }

    h5_write_vector(ghandle, "native", 0L, type="H5T_NATIVE_INT8", scalar=TRUE)

    dn <- dimnames(x)
    if (!is.null(dn)) {
        save_dimnames_for_chihaya(ghandle, rev(dn))
    }

    invisible(NULL)
}

#' @export
#' @rdname storeDelayedObject
setMethod("storeDelayedObject", "array", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
    if (save.external.array) {
        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
    }
    save_dense_array_for_chihaya(x, handle, name, extract.native=identity, version=version, ...)
})

#' @export
setMethod("storeDelayedObject", "denseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
    if (save.external.array) {
        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
    } 

    extract.native <- NULL
    if (is(x, "dMatrix") || is(x, "lMatrix")) {
        extract.native <- function(y) y@x
    }
    save_dense_array_for_chihaya(x, handle, name, extract.native=extract.native, version=version, ...)
})

#' @import rhdf5 DelayedArray
chihaya.registry$array[["dense array"]] <- function(handle, version, ...) {
    dhandle <- H5Dopen(handle, "data")
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    data <- H5Dread(dhandle)
    if (is.raw(data)) {
        storage.mode(data) <- "integer"
    }

    type <- h5_read_attribute(dhandle, "type")
    type <- to_alabaster_type(type)
    missing.placeholder <- h5_read_attribute(dhandle, "missing_placeholder", check=TRUE, default=NULL)
    data <- h5_cast(data, expected.type=type, missing.placeholder=missing.placeholder)

    if (h5_object_exists(handle, "dimnames")) {
        dn <- load_dimnames_for_chihaya(handle)
        dimnames(data) <- rev(dn)
    }

    if (h5_read_vector(handle, "native") == 1L) {
        data <- t(data)
    }

    data
}

#######################################################
#######################################################

#' @import rhdf5
save_sparse_matrix_for_chihaya <- function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "array", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_array", "sparse matrix", scalar=TRUE)

    h5_write_vector(ghandle, "shape", dim(x), type="H5T_NATIVE_UINT32")

    optimized <- optimize_storage(x)
    h5_write_sparse_matrix(x, handle=ghandle, details=optimized)

    dhandle <- H5Dopen(ghandle, "data")
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
    h5_write_attribute(dhandle, "type", to_value_type(type(x)), scalar=TRUE)
    if (!is.null(optimized$placeholder)) {
        h5_write_attribute(dhandle, missingPlaceholderName, optimized$placeholder, type=optimized$type, scalar=TRUE)
    }

    # Better if we didn't save the layout at all, but whatever.
    col <- h5_read_attribute(ghandle, "layout") == "CSC"
    h5_write_vector(ghandle, "by_column", as.integer(col), type="H5T_NATIVE_INT8", scalar=TRUE)
    H5Adelete(ghandle, "layout") 

    dn <- dimnames(x)
    if (!is.null(dn)) {
        save_dimnames_for_chihaya(ghandle, dn)
    }

    invisible(NULL)
}

#' @export
setMethod("storeDelayedObject", "sparseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
    if (save.external.array) {
        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
    }
    save_sparse_matrix_for_chihaya(x, handle, name, version=version, ...)
})

#' @export
setMethod("storeDelayedObject", "SVT_SparseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
    if (save.external.array) {
        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
    }
    save_sparse_matrix_for_chihaya(x, handle, name, version=version, ...)
})

#' @import rhdf5 DelayedArray
chihaya.registry$array[["sparse matrix"]] <- function(handle, version, ...) {
    indices <- h5_read_vector(handle, "indices")
    indptr <- h5_read_vector(handle, "indptr")
    data <- load_vector_for_chihaya(handle, "data", version=version)

    svt <- vector("list", length(indptr) - 1L)
    for (i in seq_along(svt)) {
        idx <- indptr[i] + seq_len(indptr[i+1] - indptr[i])
        if (length(idx)) {
            # As of version >= 1, data is first and indices are second.
            svt[[i]] <- list(data[idx], indices[idx])
        }
    }

    dim <- h5_read_vector(handle, "shape")
    dn <- list(NULL, NULL)
    if (h5_object_exists(handle, "dimnames")) {
        dn <- load_dimnames_for_chihaya(handle)
    }

    transposed <- (h5_read_vector(handle, "by_column") == 0L)
    if (transposed) {
        dn <- rev(dn)
        dim <- rev(dim)
    }

    output <- new("SVT_SparseMatrix", SVT=svt, dim=dim, type=typeof(data), dimnames=dn)
    if (transposed) {
        output <- t(output)
    }

    output
}

#######################################################
#######################################################

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedAbind", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_operation", "combine", scalar=TRUE)
    h5_write_vector(ghandle, "along", x@along - 1L, type="H5T_NATIVE_UINT32", scalar=TRUE)

    shandle <- H5Gcreate(ghandle, "seeds")
    on.exit(H5Gclose(shandle), add=TRUE, after=FALSE)
    h5_write_attribute(shandle, "length", length(x@seeds), type="H5T_NATIVE_UINT32", scalar=TRUE)

    for (i in seq_along(x@seeds)) {
        altStoreDelayedObject(x@seeds[[i]], shandle, as.character(i - 1L), version=version, ...)
    }

    invisible(NULL)
})

#' @import alabaster.base rhdf5 DelayedArray
chihaya.registry$operation[["combine"]] <- function(handle, version, ...) {
    shandle <- H5Gopen(handle, "seeds")
    on.exit(H5Gclose(shandle), add=TRUE, after=FALSE)

    len <- h5_read_attribute(shandle, "length")
    along <- h5_read_vector(handle, "along")

    seeds <- vector("list", len)
    for (i in seq_len(len)) {
        seeds[[i]] <- altReloadDelayedObject(shandle, as.character(i - 1L), version=version, ...)
    }

    if (along == 0L) {
        do.call(arbind, seeds)
    } else {
        do.call(acbind, seeds)
    }
}

#######################################################
#######################################################

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedAperm", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_operation", "transpose", scalar=TRUE)

    h5_write_vector(ghandle, "permutation", x@perm - 1L, type="H5T_NATIVE_UINT32")
    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
    invisible(NULL)
})

#' @import DelayedArray rhdf5
chihaya.registry$operation[["transpose"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    perm <- h5_read_vector(handle, "permutation")
    aperm(x, perm + 1L)
}

#######################################################
#######################################################

chihaya.supported.Arith <- c("+", "-", "*", "^", "/", "%%", "%/%")
chihaya.supported.Compare <- c("==", ">", "<", "!=", "<=", ">=")
chihaya.supported.Logic <- c("&", "|")
chihaya.supported.Ops <- c(chihaya.supported.Arith, chihaya.supported.Compare, chihaya.supported.Logic)

translate_logic_Ops_for_chihaya <- function(method) {
    if (method == "&") {
        "&&"
    } else if (method == "|") {
        "||"
    } else {
        method
    }
}

translate_logic_Ops_from_chihaya <- function(method) {
    if (method == "&&") {
        "&"
    } else if (method == "||") {
        "|"
    } else {
        method
    }
}

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedNaryIsoOp", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)

    # Figuring out the identity of the operation.
    chosen <- NULL
    for (p in chihaya.supported.Ops) {
        if (identical(x@OP, get(p, envir=baseenv()))) {
            chosen <- p
            break
        }
    }
    if (is.null(chosen)) {
        stop("unknown operation in ", class(x))
    }

    if (chosen %in% chihaya.supported.Arith) {
        op <- "binary arithmetic"
    } else if (chosen %in% chihaya.supported.Compare) {
        op <- "binary comparison"
    } else if (chosen %in% chihaya.supported.Logic) {
        op <- "binary logic"
        chosen <- translate_logic_Ops_for_chihaya(chosen)
    }
    h5_write_attribute(ghandle, "delayed_operation", op, scalar=TRUE)
    h5_write_vector(ghandle, "method", chosen, scalar=TRUE)

    if (length(x@seeds) != 2) {
        stop("expected exactly two seeds for 'DelayedNaryIsoOp'")
    }
    if (length(x@Rargs)) {
        stop("expected no additional right arguments for 'DelayedNaryIsoOp'")
    }

    altStoreDelayedObject(x@seeds[[1]], ghandle, "left")
    altStoreDelayedObject(x@seeds[[2]], ghandle, "right")
    invisible(NULL)
})

#' @import DelayedArray
chihaya_load_binary_op <- function(handle, version, logic, ...) {
    left <- altReloadDelayedObject(handle, "left", version=version, ...)
    right <- altReloadDelayedObject(handle, "right", version=version, ...)
    op <- h5_read_vector(handle, "method")
    if (logic) {
        op <- translate_logic_Ops_from_chihaya(op)
    }
    get(op, envir=baseenv())(left, right)
}

chihaya.registry$operation[["binary arithmetic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
chihaya.registry$operation[["binary comparison"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
chihaya.registry$operation[["binary logic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle=handle, version=version, logic=TRUE, ...)

#######################################################
#######################################################

#' @import rhdf5 alabaster.base
save_dimnames_for_chihaya <- function(handle, dimnames) {
    dhandle <- H5Gcreate(handle, "dimnames")
    on.exit(H5Gclose(dhandle), add=TRUE, after=FALSE)
    h5_write_attribute(dhandle, "length", length(dimnames), type="H5T_NATIVE_UINT32", scalar=TRUE)

    for (i in seq_along(dimnames)) {
        dn <- dimnames[[i]]
        if (is.character(dn)) { # avoid NULLs, -1's.
            h5_write_vector(dhandle, as.character(i - 1L), dn)
        }
    }
}

#' @import rhdf5 alabaster.base
load_dimnames_for_chihaya <- function(handle) {
    dhandle <- H5Gopen(handle, "dimnames")
    on.exit(H5Gclose(dhandle), add=TRUE, after=FALSE)

    dlen <- h5_read_attribute(dhandle, "length")
    dnames <- vector("list", dlen)
    for (i in seq_along(dnames)) {
        n <- as.character(i - 1L)
        if (h5_object_exists(dhandle, n)) {
            dnames[[i]] <- h5_read_vector(dhandle, n)
        }
    }

    dnames
}

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedSetDimnames", function(x, handle, name, version=package_version('1.1'), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_operation", "dimnames", scalar=TRUE)
    save_dimnames_for_chihaya(ghandle, x@dimnames)

    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
    invisible(NULL)
})

#' @import rhdf5 DelayedArray
chihaya.registry$operation[["dimnames"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    dimnames(x) <- load_dimnames_for_chihaya(handle)
    x
}

#######################################################
#######################################################

#' @import rhdf5 alabaster.base
save_chihaya_indices <- function(handle, name, indices) { 
    ihandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ihandle), add=TRUE, after=FALSE)
    h5_write_attribute(ihandle, "length", length(indices), type="H5T_NATIVE_UINT32", scalar=TRUE)

    for (i in seq_along(indices)) {
        ii <- indices[[i]]
        if (!is.null(ii)) {
            h5_write_vector(ihandle, as.character(i - 1L), ii - 1L, type="H5T_NATIVE_UINT32")
        }
    }
}

#' @import rhdf5 alabaster.base
load_chihaya_indices <- function(handle, name) {
    ihandle <- H5Gopen(handle, name)
    on.exit(H5Gclose(ihandle), add=TRUE, after=FALSE)
    ilen <- h5_read_attribute(ihandle, "length")

    indices <- vector("list", ilen)
    for (i in seq_len(ilen)) {
        n <- as.character(i - 1L)
        if (h5_object_exists(ihandle, n)) {
            indices[[i]] <- h5_read_vector(ihandle, n) + 1L
        } else {
            indices[[i]] <- substitute()
        }
    }

    indices
}

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedSubassign", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_operation", "subset assignment", scalar=TRUE)

    save_chihaya_indices(ghandle, "index", x@Lindex) 
    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
    altStoreDelayedObject(x@Rvalue, ghandle, "value", version=version, ...)
    invisible(NULL)
})

#' @import rhdf5 DelayedArray
chihaya.registry$operation[["subset assignment"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    value <- altReloadDelayedObject(handle, "value", version=version, ...)
    indices <- load_chihaya_indices(handle, "index")
    do.call(`[<-`, c(list(x=x), indices, list(value=value)))
} 

#######################################################
#######################################################

#' @export
#' @import rhdf5
setMethod("storeDelayedObject", "DelayedSubset", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
    h5_write_attribute(ghandle, "delayed_operation", "subset", scalar=TRUE)

    save_chihaya_indices(ghandle, "index", x@index) 
    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
    invisible(NULL)
})

#' @import rhdf5 DelayedArray
chihaya.registry$operation[["subset"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    indices <- load_chihaya_indices(handle, "index")
    do.call(`[`, c(list(x), indices, list(drop=FALSE)))
} 

#######################################################
#######################################################

chihaya.dump.unary.Math <- function(handle, OP) {
    envir <- environment(OP)
    generic <- envir$`.Generic`

    if (!is.null(generic)) {
        direct.support <- c("abs", "sign", "sqrt", "ceiling", "floor", "trunc",
                            "exp", "expm1", "log1p", "cos", "cosh", "sin",
                            "sinh", "tan", "tanh", "acos", "acosh", "asin",
                            "asinh", "atan", "atanh", "gamma", "lgamma",
                            "digamma", "trigamma")

        if (generic %in% direct.support) {
            h5_write_attribute(handle, "delayed_operation", 'unary math', scalar=TRUE)
            h5_write_vector(handle, "method", generic, scalar=TRUE)
            return(TRUE)
        }

        log.base.support <- c(log2=2, log10=10)
        if (generic %in% names(log.base.support)) {
            h5_write_attribute(handle, "delayed_operation", 'unary math', scalar=TRUE)
            h5_write_vector(handle, "method", "log", scalar=TRUE)
            h5_write_vector(handle, "base", log.base.support[[generic]], type="H5T_NATIVE_DOUBLE", scalar=TRUE)
            return(TRUE)
        }
    }

    # Special case for the general case log.
    if (isTRUE(all.equal(as.character(body(OP)), c("log", "a", "base")))) {
        h5_write_attribute(handle, "delayed_operation", 'unary math', scalar=TRUE)
        h5_write_vector(handle, "method", "log", scalar=TRUE)

        base <- envir$base
        if (base != exp(1)) {
            h5_write_vector(handle, "base", base, type="H5T_NATIVE_DOUBLE", scalar=TRUE)
        }
        return(TRUE)
    }

    FALSE
}

chihaya.dump.unary.Math2 <- function(handle, OP) {
    envir <- environment(OP)
    generic <- envir$`.Generic`

    if (generic %in% getGroupMembers("Math2")) {
        h5_write_attribute(handle, "delayed_operation", 'unary math', scalar=TRUE)
        h5_write_vector(handle, "method", generic, scalar=TRUE)
        h5_write_vector(handle, "digits", envir$digits, type="H5T_NATIVE_INT32", scalar=TRUE)
        return(TRUE)
    }

    FALSE
}

chihaya.dump.unary.Ops <- function(handle, OP) {
    envir <- environment(OP)
    generic <- envir$`.Generic`

    if (!(generic %in% chihaya.supported.Ops)) {
        return(FALSE)
    }

    delayed_op <- NULL
    if (generic %in% chihaya.supported.Arith) {
        delayed_op <- "unary arithmetic"
    } else if (generic %in% chihaya.supported.Compare) {
        delayed_op <- "unary comparison"
    } else if (generic %in% chihaya.supported.Logic) {
        delayed_op <- "unary logic"
        generic <- translate_logic_Ops_for_chihaya(generic)
    }
    h5_write_attribute(handle, "delayed_operation", delayed_op, scalar=TRUE)
    h5_write_vector(handle, "method", generic, scalar=TRUE)

    e1 <- envir$e1
    e2 <- envir$e2

    if (missing(e2)) {
        if (!generic %in% c("+", "-")) {
            stop("second argument can only be missing for unary '+' or '-'")
        }
        h5_write_vector(handle, "side", "none", scalar=TRUE)

    } else {
        right <- is(e1, "DelayedArray") # i.e., is the operation applied to the left of the seed?
        left <- is(e2, "DelayedArray") # i.e., is the operation applied to the left of the seed?

        h5_write_vector(handle, "side", if (left) "left" else "right", scalar=TRUE)
        val <- if (left) e1 else e2
        if (length(val) == 1) {
            save_vector_for_chihaya(handle, "value", val, version=version, scalar=TRUE)
        } else {
            # Don't think this ever gets called, because otherwise
            # we'd be dealing with a DelayedIsoOpWithArgs.
            # Nonetheless, we'll throw in the necessary code.
            h5_write_vector(handle, "along", 0, type="H5T_NATIVE_UINT32", scalar=TRUE)
            save_vector_for_chihaya(handle, "value", val, version=version)
        }
    }

    TRUE
}

chihaya.unary.logic.Ops <- c(is.infinite="is_infinite", is.nan="is_nan", is.finite="is_finite")

chihaya.dump.unary.other <- function(handle, OP) {
    envir <- environment(OP)
    generic <- envir$`.Generic`

    if (generic == "!") {
        h5_write_attribute(handle, "delayed_operation", "unary logic", scalar=TRUE)
        h5_write_vector(handle, "method", "!", scalar=TRUE)
        return(TRUE)

    } else if (generic %in% names(chihaya.unary.logic.Ops)) {
        h5_write_attribute(handle, "delayed_operation", "unary special check", scalar=TRUE)
        h5_write_vector(handle, "method", chihaya.unary.logic.Ops[[generic]], scalar=TRUE)
        return(TRUE)
    }

    FALSE
}

save_iso_op_stack <- function(handle, name, OPS, i, seed, version, ...) {
    if (i == 0L) {
        altStoreDelayedObject(seed, handle, name, version=version, ...)
        return(NULL)
    }

    OP <- OPS[[i]]
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    h5_write_attribute(ghandle, "delayed_type", 'operation', scalar=TRUE)

    if (chihaya.dump.unary.Math(ghandle, OP)) {
        return(save_iso_op_stack(ghandle, "seed", OPS, i - 1L, seed=seed, version=version, ...))
    }

    if (chihaya.dump.unary.Math2(ghandle, OP)) {
        return(save_iso_op_stack(ghandle, "seed", OPS, i - 1L, seed=seed, version=version, ...))
    }

    if (chihaya.dump.unary.Ops(ghandle, OP)) {
        return(save_iso_op_stack(ghandle, "seed", OPS, i - 1L, seed=seed, version=version, ...))
    }

    if (chihaya.dump.unary.other(ghandle, OP)) {
        return(save_iso_op_stack(ghandle, "seed", OPS, i - 1L, seed=seed, version=version, ...))
    }

    stop("unknown OPS[[", i, "]] function when saving a 'DelayedUnaryIsoOpStack'")
}

#' @export
setMethod("storeDelayedObject", "DelayedUnaryIsoOpStack", function(x, handle, name, version=package_version("1.1"), ...) {
    # This saves in reverse order, as first operation is first applied (and thus needs to be closer to the leaf of the delayed tree).
    save_iso_op_stack(handle, name, OPS=x@OPS, i=length(x@OPS), seed=x@seed, version=version, ...)
    invisible(NULL)
})

#' @export
setMethod("storeDelayedObject", "DelayedUnaryIsoOpWithArgs", function(x, handle, name, version=package_version("1.1"), ...) {
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    h5_write_attribute(ghandle, "delayed_type", 'operation', scalar=TRUE)

    # Figuring out the identity of the operation.
    chosen <- NULL
    for (p in chihaya.supported.Ops) {
        if (identical(x@OP, get(p, envir=baseenv()))) {
            chosen <- p
            break
        }
    }
    if (is.null(chosen)) {
        stop("unknown operation in ", class(x))
    }

    op <- NULL
    if (chosen %in% chihaya.supported.Logic) {
        op <- "unary logic"
        chosen <- translate_logic_Ops_for_chihaya(chosen)
    } else if (chosen %in% chihaya.supported.Compare) {
        op <- "unary comparison"
    } else if (chosen %in% chihaya.supported.Arith) {
        op <- "unary arithmetic"
    }
    h5_write_attribute(ghandle, "delayed_operation", op, scalar=TRUE)
    h5_write_vector(ghandle, "method", chosen, scalar=TRUE)

    # Saving the left and right args. There should only be one or the other.
    # as the presence of both is not commutative.
    if (length(x@Rargs) + length(x@Largs) !=1) {
        stop("'DelayedUnaryIsoApWithArgs' should operate on exactly one argument")
    }

    left <- length(x@Largs) > 0
    if (left) {
        args <- x@Largs[[1]]
        along <- x@Lalong[1]
    } else {
        args <- x@Rargs[[1]]
        along <- x@Ralong[1]
    }

    h5_write_vector(ghandle, "side", if (left) "left" else "right", scalar=TRUE)
    h5_write_vector(ghandle, "along", along - 1L, type="H5T_NATIVE_UINT32", scalar=TRUE)

    if (length(args) == 1L) {
        save_vector_for_chihaya(ghandle, "value", args, version=version, scalar=TRUE)
    } else if (is.null(dim(args)) || length(dim(args)) == 1L) {
        save_vector_for_chihaya(ghandle, "value", args, version=version)
    } else {
        stop("multi-dimensional 'value' not supported in 'DelayedUnaryIsoOpWithArgs'")
    }

    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...) 
    invisible(NULL)
})

chihaya.registry$operation[["unary math"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    method <- h5_read_vector(handle, "method")
    output <- NULL

    if (method == "log") {
        if (h5_object_exists(handle, "base")){
            base <- h5_read_vector(handle, "base")
            output <- log(x, base=base)
        } else {
            output <- log(x)
        }

    } else if (method %in% getGroupMembers("Math2")) {
        digits <- h5_read_vector(handle, "digits")
        FUN <- get(method, envir=baseenv())
        output <- FUN(x, digits=digits)

    } else {
        FUN <- get(method, envir=baseenv())
        output <- FUN(x)
    }

    output
}

apply_unary_op_with_value <- function(x, op, side, handle, version) {
    value <- load_vector_for_chihaya(handle, "value", version=version)
    FUN <- get(op, envir=baseenv())

    simple <- FALSE
    if (length(value) == 1L) {
        simple <- TRUE
    } else {
        along <- h5_read_vector(handle, "along")
        simple <- along == 0
    }

    output <- NULL
    if (simple) {
        if (side == "left") {
            output <- FUN(value, x)
        } else if (side == "right") {
            output <- FUN(x, value)
        } else {
            stop("unknown side '", side, "'")
        }

    } else {
        # Stolen from base::sweep.
        along <- along + 1L
        perm <- c(along, seq_along(dim(x))[-along])
        tmp <- aperm(x, perm)

        if (side == "left") {
            tmp <- FUN(value, tmp)
        } else if (side == "right") {
            tmp <- FUN(tmp, value)
        } else {
            stop("unknown side '", side, "'")
        }

        output <- aperm(tmp, order(perm))
    }

    output
}

chihaya.registry$operation[["unary logic"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    method <- h5_read_vector(handle, "method")

    output <- NULL
    if (method == "!") {
        output <- !x
    } else {
        method <- translate_logic_Ops_from_chihaya(method)
        side <- h5_read_vector(handle, "side")
        output <- apply_unary_op_with_value(x, op=method, side=side, handle=handle, version=version)
    }

    output
}

chihaya.registry$operation[["unary comparison"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    method <- h5_read_vector(handle, "method")
    side <- h5_read_vector(handle, "side")
    apply_unary_op_with_value(x, op=method, side=side, handle=handle, version=version)
}

chihaya.registry$operation[["unary special check"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    method <- h5_read_vector(handle, "method")
    method <- sub("_", ".", method)
    get(method, envir=baseenv())(x)
}

chihaya.registry$operation[["unary arithmetic"]] <- function(handle, version, ...) {
    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
    method <- h5_read_vector(handle, "method")
    side <- h5_read_vector(handle, "side")

    output <- NULL
    if (side == "none") {
        if (method == "+") {
            output <- x
        } else if (method == "-") {
            output <- -x
        } else {
            stop("unsupported method '", method, "' with side 'none'")
        }
    } else {
        output <- apply_unary_op_with_value(x, op=method, side=side, handle=handle, version=version)
    }

    output
}

#######################################################
#######################################################

#' @export
#' @import rhdf5
#' @rdname storeDelayedObject
setMethod("storeDelayedObject", "ANY", function(
    x,
    handle,
    name,
    version=package_version("1.1"),
    external.save.args=list(),
    external.dedup.session=NULL,
    external.dedup.action=NULL,
    ...) 
{
    ghandle <- H5Gcreate(handle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    if (is(x, "LowRankMatrixSeed")) { # From BiocSingular.
        h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
        h5_write_attribute(ghandle, "delayed_operation", "matrix product", scalar=TRUE)

        altStoreDelayedObject(x@rotation, ghandle, "left_seed", version=version, ...)
        h5_write_vector(ghandle, "left_orientation", "N", scalar=TRUE)
        altStoreDelayedObject(x@components, ghandle, "right_seed", version=version, ...)
        h5_write_vector(ghandle, "right_orientation", "T", scalar=TRUE)

    } else if (is(x, "ResidualMatrixSeed")) {
        h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
        h5_write_vector(ghandle, "_r_type_hint", "residual matrix", scalar=TRUE)

        # Mimic a transposition operation.
        xhandle <- ghandle
        if (x@transposed) {
            h5_write_attribute(ghandle, "delayed_operation", "transpose", scalar=TRUE)
            h5_write_vector(ghandle, "permutation", c(1L, 0L), type="H5T_NATIVE_UINT32") 
            xhandle <- H5Gcreate(ghandle, "seed")
            on.exit(H5Gclose(xhandle), add=TRUE, after=FALSE)
            h5_write_attribute(xhandle, "delayed_type", "operation", scalar=TRUE)
        }

        # Mimic a binary subtraction.
        h5_write_attribute(xhandle, "delayed_operation", "binary arithmetic", scalar=TRUE)
        h5_write_vector(xhandle, "method", "-", scalar=TRUE)
        altStoreDelayedObject(x@.matrix, xhandle, "left", version=version, ...)

        # Mimic a matrix product.
        rhandle <- H5Gcreate(xhandle, "right")
        on.exit(H5Gclose(rhandle), add=TRUE, after=FALSE)
        h5_write_attribute(rhandle, "delayed_type", "operation", scalar=TRUE)
        h5_write_attribute(rhandle, "delayed_operation", "matrix product", scalar=TRUE)
        altStoreDelayedObject(x@Q, rhandle, "left_seed", version=version, ...)
        h5_write_vector(rhandle, "left_orientation", "N", scalar=TRUE)
        altStoreDelayedObject(x@Qty, rhandle, "right_seed", version=version, ...)
        h5_write_vector(rhandle, "right_orientation", "N", scalar=TRUE)

    } else {
        h5_write_attribute(ghandle, "delayed_type", "array", scalar=TRUE)
        h5_write_attribute(ghandle, "delayed_array", "custom takane seed array", scalar=TRUE)

        exdir <- file.path(dirname(H5Fget_name(handle)), "seeds")
        dir.create(exdir, showWarnings=FALSE)
        n <- length(list.files(exdir))
        output <- file.path(exdir, n)

        # For back-compatibility.
        if (!is.null(external.dedup.session)) {
            external.save.args$array.dedup.session <- external.dedup.session
        }
        if (!is.null(external.dedup.action)) {
            external.save.args$array.dedup.action <- external.dedup.action
        }

        # We try twice, once for the seed itself, and then again after
        # wrapping it in a DelayedArray, just in case the seed has an Array
        # class with an altSaveObject method. If not, the DelayedArray call
        # will just fall back to the default dense/sparse savers.
        exargs <- c(list(x=x, path=output), external.save.args)
        status <- do.call(try_altSaveObject, exargs)
        if (!status) {
            exargs$x <- DelayedArray(exargs$x)
            exargs$DelayedArray.dispatch.pristine <- FALSE # we only got here if we failed to save the seed itself, so no point trying again.
            exargs$DelayedArray.preserve.ops <- FALSE # avoid infinite recursion between storeDelayedObject and saveObject.
            do.call(altSaveObject, exargs)
        }

        h5_write_vector(ghandle, "dimensions", dim(x), type="H5T_NATIVE_UINT32", compress=0)
        h5_write_vector(ghandle, "type", to_value_type(type(x)), scalar=TRUE)
        h5_write_vector(ghandle, "index", n, type="H5T_NATIVE_UINT32", scalar=TRUE)
    }
})

#' @import alabaster.base rhdf5 DelayedArray
chihaya.registry$array[["custom takane seed array"]] <- function(handle, version, custom.takane.realize=FALSE, ...) {
    index <- h5_read_vector(handle, "index")
    out <- altReadObject(file.path(dirname(H5Fget_name(handle)), "seeds", index), ...)

    if (custom.takane.realize) {
        if (is_sparse(out)) {
            out <- as(out, "SVT_SparseArray")
        } else {
            out <- as.array(out)
        }
    }

    out
}

#' @importFrom Matrix t
chihaya.registry$operation[["matrix product"]] <- function(handle, version, ...) {
    L <- as.matrix(altReloadDelayedObject(handle, "left_seed"))
    Lori <- h5_read_vector(handle, "left_orientation")
    if (length(Lori) == 1 && as.character(Lori) == "T") {
        L <- t(L)
    } 

    R <- as.matrix(altReloadDelayedObject(handle, "right_seed"))
    Rori <- h5_read_vector(handle, "right_orientation")
    if (length(Rori) == 1 && as.character(Rori) == "N") {
        R <- t(R)
    } 

    BiocSingular::LowRankMatrix(L, R)
}

chihaya.registry$type.hint[["residual matrix"]] <- function(handle, version, ...) {
    if (!isNamespaceLoaded("ResidualMatrix")) {
        loadNamespace("ResidualMatrix")
    }

    stopifnot(identical(h5_read_attribute(handle, "delayed_type"), "operation"))
    optype <- h5_read_attribute(handle, "delayed_operation")

    transposed <- FALSE
    xhandle <- handle
    if (optype == "transpose") {
        transposed <- TRUE
        xhandle <- H5Gopen(handle, "seed")
        on.exit(H5Gclose(xhandle), add=TRUE, after=FALSE)
        stopifnot(identical(h5_read_attribute(xhandle, "delayed_type"), "operation"))
        optype <- h5_read_attribute(xhandle, "delayed_operation")
    }

    stopifnot(identical(optype, "binary arithmetic"))
    stopifnot(identical(h5_read_vector(xhandle, "method"), "-"))
    .matrix <- altReloadDelayedObject(xhandle, "left")
    .matrix <- .matrix@seed # can't be a DelayedArray inside ResidualMatrix, for various reasons...

    rhandle <- H5Gopen(xhandle, "right")
    on.exit(H5Gclose(rhandle), add=TRUE, after=FALSE)
    stopifnot(identical(h5_read_attribute(rhandle, "delayed_type"), "operation"))
    stopifnot(identical(h5_read_attribute(rhandle, "delayed_operation"), "matrix product"))
    stopifnot(identical(h5_read_vector(rhandle, "left_orientation"), "N"))
    stopifnot(identical(h5_read_vector(rhandle, "right_orientation"), "N"))

    Q <- as.matrix(altReloadDelayedObject(rhandle, "left_seed", version=version, ...))
    Qty <- as.matrix(altReloadDelayedObject(rhandle, "right_seed", version=version, ...))
    seed <- new("ResidualMatrixSeed", .matrix = .matrix, Q = Q, Qty = Qty, transposed = transposed)
    DelayedArray(seed)
}

#######################################################
#######################################################

#' @export
#' @rdname storeDelayedObject
altStoreDelayedObjectFunction <- (function() {
    cur.env <- new.env()
    cur.env$store <- NULL
    function(store) {
        prev <- cur.env$store
        if (missing(store)) {
            prev
        } else {
            cur.env$store <- store
            invisible(prev)
        }
    }
})()

#' @export
#' @rdname storeDelayedObject
altStoreDelayedObject <- function(...) {
    FUN <- altStoreDelayedObjectFunction()
    if (is.null(FUN)) {
        FUN <- storeDelayedObject
    }
    FUN(...)
}

#' @export
#' @rdname storeDelayedObject
altReloadDelayedObjectFunction <- (function() {
    cur.env <- new.env()
    cur.env$reload <- NULL
    function(reload) {
        prev <- cur.env$reload
        if (missing(reload)) {
            prev
        } else {
            cur.env$reload <- reload
            invisible(prev)
        }
    }
})()

#' @export
#' @rdname storeDelayedObject
altReloadDelayedObject <- function(...) {
    FUN <- altReloadDelayedObjectFunction()
    if (is.null(FUN)) {
        FUN <- reloadDelayedObject
    }
    FUN(...)
}

#' @export
createExternalSeedDedupSession <- function() {
    createDedupSession()
}

# This tests the behavior of the various HDF4-based seeds.
# library(testthat); library(alabaster.matrix); source("setup.R"); source("test-storeDelayedObject.R")

library(DelayedArray)
library(rhdf5)

saveDelayed <- function(x, ...) {
    tmp <- tempfile()
    saveObject(x, tmp, DelayedArray.dispatch.pristine=FALSE, DelayedArray.preserve.ops=TRUE, ...)
    stopifnot(readObjectFile(tmp)$type == "delayed_array")
    tmp
}

loadDelayed <- function(path, ...) {
    raw <- readObject(path, delayed_array.reload.args=list(...))
    expect_s4_class(raw, "ReloadedArray")
    DelayedArray(raw@seed@seed) # removing the ReloadedArray for convenience.
}

#######################################################
#######################################################

test_that("ConstantArrays are saved correctly", {
    # Behaves correctly for logical values.
    X <- ConstantArray(dim=c(12, 7))
    temp <- saveDelayed(X)

    out <- loadDelayed(temp)
    expect_identical(X, out)

    # Behaves correctly for string values.
    X <- ConstantArray(dim=c(12, 7), value="foo")
    temp <- saveDelayed(X)

    out <- loadDelayed(temp)
    expect_identical(X, out)

    # Behaves correctly for numeric values.
    X <- ConstantArray(dim=c(12, 7), value=2L)
    temp <- saveDelayed(X)

    out <- loadDelayed(temp)
    expect_identical(X, out)
})

test_that("ConstantArrays are still saved correctly after some deep nesting", {
    X <- ConstantArray(dim=c(12, 7), value=23)
    Y <- DelayedArray(matrix(runif(60), nrow=12))
    Z <- cbind(X, Y)

    temp <- saveDelayed(Z)
    out <- loadDelayed(temp)
    expect_identical(Z, out)
})

test_that("ConstantArrays behave correctly with NAs", {
    X <- ConstantArray(dim=c(12, 7), value=NA)
    temp <- saveDelayed(X)
    out <- loadDelayed(temp)
    expect_identical(X, out)

    # Trying a non-Default NA.
    X <- ConstantArray(dim=c(12, 7), value=1.2)
    temp <- saveDelayed(X)

    library(rhdf5)
    (function() {
        fhandle <- H5Fopen(file.path(temp, "array.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE)
        ghandle <- H5Gopen(fhandle, "delayed_array")
        on.exit(H5Gclose(ghandle), add=TRUE)
        dhandle <- H5Dopen(ghandle, "value")
        on.exit(H5Dclose(dhandle), add=TRUE)
        h5_write_attribute(dhandle, "missing_placeholder", 1.2, type="H5T_NATIVE_DOUBLE", scalar=TRUE)
    })()

    out <- loadDelayed(temp)
    expect_identical(ConstantArray(c(12, 7), value=NA_real_), out)
})

#######################################################
#######################################################

test_that("dense arrays are saved correctly", {
    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
    temp <- saveDelayed(X)
    roundtrip <- loadDelayed(temp)
    expect_identical(X, roundtrip)

    X <- DelayedArray(Matrix::Matrix(runif(60), ncol=20))
    temp <- saveDelayed(X)
    roundtrip <- loadDelayed(temp)
    expect_identical(as.array(X), as.array(roundtrip@seed))

    X <- DelayedArray(Matrix::Matrix(runif(60) > 0.5, ncol=20, sparse=FALSE))
    temp <- saveDelayed(X)
    roundtrip <- loadDelayed(temp)
    expect_identical(as.array(X), as.array(roundtrip@seed))

    # Check that this works with other dense types.
    X <- DelayedArray(Matrix::triu(Matrix::Matrix(runif(100), ncol=10)))
    expect_s4_class(X@seed, "dtrMatrix")
    temp <- saveDelayed(X)
    roundtrip <- loadDelayed(temp)
    expect_identical(as.array(X), roundtrip@seed)
})

test_that("dense array saving handles dimnames", {
    X <- matrix(rpois(60, 5), ncol=20)
    rownames(X) <- LETTERS[1:3]
    colnames(X) <- letters[1:20]

    temp <- saveDelayed(DelayedArray(X))
    roundtrip <- loadDelayed(temp)
    expect_identical(X, roundtrip@seed)
})

test_that("dense array saving handles a bit of transposition", {
    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
    temp <- saveDelayed(X)

    (function() {
        fhandle <- H5Fopen(file.path(temp, "array.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, "delayed_array")
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
        H5Ldelete(ghandle, "native")
        h5_write_vector(ghandle, "native", 1, type="H5T_NATIVE_UINT8", scalar=TRUE)
    })()

    roundtrip <- loadDelayed(temp)
    expect_identical(t(X@seed), roundtrip@seed)
})

test_that("dense arrays are saved to external arrays if requested", {
    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
    temp <- saveDelayed(X, DelayedArray.force.external=TRUE)
    expect_true(file.exists(file.path(temp, "seeds", 0)))
    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
    expect_identical(X, roundtrip)

    X <- DelayedArray(Matrix::Matrix(runif(60), ncol=20))
    temp <- saveDelayed(X, DelayedArray.force.external=TRUE)
    expect_true(file.exists(file.path(temp, "seeds", 0)))
    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
    expect_identical(unname(as.array(X)), as.array(roundtrip@seed))
})

#######################################################
#######################################################

test_that("sparse matrices are saved correctly", {
    X <- DelayedArray(Matrix::rsparsematrix(20, 30, 0.2))
    temp <- saveDelayed(X)
    roundtrip <- loadDelayed(temp)
    expect_s4_class(roundtrip@seed, "SVT_SparseMatrix")
    expect_identical(unname(as.matrix(X)), as.matrix(roundtrip))

    # Checking that it works in CSR mode. 
    X <- Matrix::rsparsematrix(20, 30, 0.2)
    X <- as(X > 0, "RsparseMatrix")
    Z <- DelayedArray(X)
    temp <- saveDelayed(Z)
    roundtrip <- loadDelayed(temp)
    expect_s4_class(roundtrip@seed, "SVT_SparseMatrix")
    expect_identical(unname(as.matrix(X)), as.matrix(roundtrip))

    # Works with SVT sparse matrix inputs.
    X <- matrix(rpois(100, 0.5) * 10L, ncol=25)
    X <- as(X, "SVT_SparseMatrix")
    Z <- DelayedArray(X)
    temp <- saveDelayed(Z)
    roundtrip <- loadDelayed(temp)
    expect_s4_class(roundtrip@seed, "SVT_SparseMatrix")
    expect_identical(unname(as.matrix(X)), as.matrix(roundtrip))
})

test_that("sparse matrix saving handles dimnames", {
    X <- Matrix::rsparsematrix(20, 30, 0.2)
    rownames(X) <- LETTERS[1:20]
    colnames(X) <- 1:30

    temp <- saveDelayed(DelayedArray(X))
    roundtrip <- loadDelayed(temp)
    expect_identical(type(X), type(roundtrip@seed))
    expect_identical(X, as(roundtrip@seed, "dgCMatrix"))
})

test_that("sparse matrices are saved to external arrays if requested", {
    X <- matrix(rpois(200, 1) - 1L, ncol=20)
    X <- as(X, "SVT_SparseMatrix")
    temp <- saveDelayed(DelayedArray(X), DelayedArray.force.external=TRUE)
    expect_true(file.exists(file.path(temp, "seeds", 0)))

    roundtrip <- loadDelayed(temp)
    expect_true(is_sparse(roundtrip))
    expect_identical(as.matrix(X), as.matrix(roundtrip))
})

#######################################################
#######################################################

test_that("DelayedAbind works along rows", {
    X <- DelayedArray(matrix(runif(60), ncol=20))
    Y <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- rbind(X, Y)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedAbind")
})

test_that("DelayedAbind works along columns", {
    A <- DelayedArray(matrix(runif(60), nrow=20))
    B <- DelayedArray(matrix(runif(100), nrow=20))
    C <- DelayedArray(matrix(runif(200), nrow=20)) # throwing in another dataset for some variety.
    Z <- BiocGenerics::cbind(A, B, C)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedAbind")
})

test_that("DelayedAbind works for 3D arrays", {
    A <- DelayedArray(array(runif(100), c(10, 5, 4)))
    B <- DelayedArray(array(runif(100), c(10, 5, 4)))
    Z <- arbind(A, B)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedAbind")
})

#######################################################
#######################################################

test_that("DelayedAperm works along rows", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- t(X)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedAperm")
})

test_that("DelayedAperm works for 3D arrays", {
    A <- DelayedArray(array(runif(100), c(10, 5, 4)))
    perm <- c(3L, 1L, 2L)
    Z <- aperm(A, perm)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedAperm")
})

#######################################################
#######################################################

test_that("DelayedNaryIsoOp works as expected (arithmetic)", {
    X <- DelayedArray(matrix(runif(100), ncol=5))
    Y <- DelayedArray(matrix(runif(100), ncol=5))
    Z <- X * Y
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedNaryIsoOp")
})

test_that("DelayedNaryIsoOp works as expected (comparison)", {
    X <- DelayedArray(matrix(rpois(100, 5), ncol=5))
    Y <- DelayedArray(matrix(rpois(100, 2), ncol=5))
    Z <- X > Y
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedNaryIsoOp")
})

test_that("DelayedNaryIsoOp works as expected (logic)", {
    X <- DelayedArray(matrix(rbinom(100, 1, 0.5) != 0, ncol=5))
    Y <- DelayedArray(matrix(rbinom(100, 1, 0.5) != 0, ncol=5))
    Z <- X | Y
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedNaryIsoOp")
})

test_that("DelayedNaryIsoOp works for 3D arrays", {
    A <- DelayedArray(array(runif(100), c(10, 5, 4)))
    B <- DelayedArray(array(runif(100), c(10, 5, 4)))
    Z <- A <= B
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedNaryIsoOp")
})

test_that("DelayedNaryIsoOp works with multiple seeds", {
    X <- DelayedArray(matrix(rpois(100, 5), ncol=5))
    Y <- DelayedArray(matrix(rpois(100, 2), ncol=5))
    Z <- DelayedArray(matrix(rbinom(100, 1, 0.5) == 0, ncol=5))
    AA <- X - Y + Z
    temp <- saveDelayed(AA)

    roundtrip <- loadDelayed(temp)
    expect_identical(AA, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedNaryIsoOp")
})

#######################################################
#######################################################

test_that("DelayedSetDimnames works as expected (colnames only)", {
    Z <- DelayedArray(matrix(runif(100), ncol=20))
    colnames(Z) <- LETTERS[1:20]
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSetDimnames")
})

test_that("DelayedSetDimnames works as expected (rownames only)", {
    Z <- DelayedArray(matrix(runif(100) < 0.1, ncol=20))
    rownames(Z) <- letters[1:5]
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSetDimnames")
})

test_that("DelayedSetDimnames works as expected (both sets of names)", {
    Z <- DelayedArray(matrix(rpois(100, 5), ncol=20))
    dimnames(Z) <- list(letters[1:5], LETTERS[1:20])
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(Z, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSetDimnames")
})

#######################################################
#######################################################

test_that("DelayedSubassign works when all indices are supplied", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    X[1:2,c(1, 10, 20)] <- matrix(-runif(6), ncol=3)
    temp <- saveDelayed(X)

    roundtrip <- loadDelayed(temp)
    expect_identical(X, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSubassign")
})

test_that("DelayedSubassign works when only one index is supplied", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    X[c(1, 5), ] <- matrix(-rpois(2*ncol(X), 12), ncol=ncol(X))
    temp <- saveDelayed(X)

    roundtrip <- loadDelayed(temp)
    expect_identical(X, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSubassign")
})

test_that("DelayedSubassign works when the replacement is a DelayedArray", {
    X <- DelayedArray(matrix(rbinom(100, 1, 0.5) == 0, ncol=20))
    X[1:2,3:5] <- DelayedArray(matrix(-runif(6), ncol=3)) + 1
    temp <- saveDelayed(X)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(X), as.matrix(roundtrip)) # see comments below for the DelayedUnaryIsoOpStack tests.
    expect_s4_class(roundtrip@seed, "DelayedSubassign")
})

#######################################################
#######################################################

test_that("DelayedSubset works when all indices are supplied", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Y <- X[1:2,3:5]
    temp <- saveDelayed(Y)

    roundtrip <- loadDelayed(temp)
    expect_identical(Y, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSubset")

    # Trying with out-of-order indices.
    Y <- X[c(1,3,5),c(2,8,6,4)]
    temp <- saveDelayed(Y)
    roundtrip <- loadDelayed(temp)
    expect_identical(Y, roundtrip)
})

test_that("DelayedSubset works when only one index is supplied", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Y <- X[,c(20, 1, 3, 15, 9)] 
    temp <- saveDelayed(Y)

    roundtrip <- loadDelayed(temp)
    expect_identical(Y, roundtrip)
    expect_s4_class(roundtrip@seed, "DelayedSubset")
})

#######################################################
#######################################################

# These use as.matrix() in comparisons to avoid problems with environment
# comparisons in the functions stored in OPS.

test_that("DelayedUnaryIsoOpStack works for multiple operations", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- abs(X - 0.5)
    expect_s4_class(Z@seed, "DelayedUnaryIsoOpStack")
    expect_type(Z@seed@seed, "double")

    temp <- saveDelayed(Z)
    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works on both sides", {
    X <- DelayedArray(matrix(rpois(100, 5) + 1L, ncol=20))
    Z <- 5 / X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    # Trying on the other side with another non-commutative op.
    Z <- X - 10
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works for log", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- log(X)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    # Works with non-default base.
    Z <- log(X, base=3)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works (comparisons)", {
    X <- DelayedArray(matrix(runif(1000), ncol=20))
    Z <- 0.5 < X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    Z <- X <= 0.2
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works (logic operations)", {
    X <- DelayedArray(matrix(runif(1000), ncol=20) > 0.5)
    Z <- X & TRUE
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    # Same for the ||.
    Z <- X | FALSE
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    # Same for !
    Z <- !X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works for unary arithmetic", {
    X <- DelayedArray(matrix(rnorm(1000), ncol=20))
    Z <- -X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    Z <- +X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_type(roundtrip@seed, "double")
})

test_that("DelayedUnaryIsoOpStack works for other unary operations", {
    suppressWarnings(X <- DelayedArray(matrix(log(rnorm(1000)), ncol=20)))
    Z <- is.nan(X)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    suppressWarnings(expect_identical(as.matrix(Z), as.matrix(roundtrip)))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

test_that("DelayedUnaryIsoOpStack works for Math2", {
    X <- DelayedArray(matrix(rnorm(1000) * 10, ncol=20))
    Z <- round(X)
    temp <- tempfile(fileext=".h5")
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")

    # Throwing in some non-standard digits.
    Z <- signif(X, digits=3)
    temp <- tempfile(fileext=".h5")
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpStack")
})

#######################################################
#######################################################

test_that("DelayedUnaryIsoOpWithArgs works as expected", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- X - runif(5)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

test_that("DelayedUnaryIsoOpWithArgs works along the other dimension", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- X - runif(5)
    temp <- saveDelayed(Z)

    # Manually injecting along=1, because we can't actually seem to stage a
    # DelayedArray directly with along=1; calling DelayedArray::sweep does
    # a double-transpose instead.
    library(rhdf5)
    vec <- runif(20)
    (function() {
        fhandle <- H5Fopen(file.path(temp, "array.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)
        ghandle <- H5Gopen(fhandle, "delayed_array")
        on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
        H5Ldelete(ghandle, "along")
        h5_write_vector(ghandle, "along", 1, type="H5T_NATIVE_UINT8", scalar=TRUE)
        H5Ldelete(ghandle, "value")
        dhandle <- h5_write_vector(ghandle, "value", vec, type="H5T_NATIVE_DOUBLE", emit=TRUE)
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        h5_write_attribute(dhandle, "type", "FLOAT", scalar=TRUE)
    })()

    roundtrip <- loadDelayed(temp)
    expected <- sweep(as.matrix(X), MARGIN=2, STATS=vec, FUN="-")
    expect_identical(as.matrix(expected), as.matrix(roundtrip))
})

test_that("DelayedUnaryIsoOpWithArgs handles logical renaming", {
    X <- DelayedArray(matrix(runif(100) > 0.5, ncol=20))
    Z <- X & runif(5) > 0.5
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

test_that("DelayedUnaryIsoOpWithArgs handles NAs correctly", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    vec <- runif(5)
    vec[1] <- NA
    Z <- X - vec
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")

    # Works with non-default NAs
    vec <- 1:5
    Z <- X / vec
    temp <- saveDelayed(Z)

    library(rhdf5)
    (function() {
        fhandle <- H5Fopen(file.path(temp, "array.h5"))
        on.exit(H5Fclose(fhandle), add=TRUE)
        dhandle <- H5Dopen(fhandle, "delayed_array/value")
        on.exit(H5Dclose(dhandle), add=TRUE)
        h5_write_attribute(dhandle, "missing_placeholder", 3, type="H5T_NATIVE_DOUBLE", scalar=TRUE)
    })()

    roundtrip <- loadDelayed(temp)
    vec[3] <- NA
    expected <- X / vec
    expect_identical(as.matrix(expected), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

test_that("DelayedUnaryIsoOpWithArgs handles side-ness correctly", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- runif(5) / X
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

test_that("DelayedUnaryIsoOpWithArgs handles repeated operations correctly (same side)", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- (X + runif(5)) + runif(5)
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

test_that("DelayedUnaryIsoOpWithArgs handles repeated operations correctly (opposite sides)", {
    X <- DelayedArray(matrix(runif(100), ncol=20))
    Z <- runif(5) + (X + runif(5))
    temp <- saveDelayed(Z)

    roundtrip <- loadDelayed(temp)
    expect_identical(as.matrix(Z), as.matrix(roundtrip))
    expect_s4_class(roundtrip@seed, "DelayedUnaryIsoOpWithArgs")
})

#######################################################
#######################################################

test_that("saving of a LowRankMatrix works correctly", {
    left <- matrix(rnorm(100000), ncol=20)
    right <- matrix(rnorm(50000), ncol=20)

    library(BiocSingular)
    thing <- LowRankMatrix(left, right)
    temp <- saveDelayed(thing)

    out <- loadDelayed(temp)
    expect_identical(thing, out)
})

test_that("saving of a ResidualMatrix works correctly", {
    y <- rsparsematrix(80, 50, 0.5)
    design <- model.matrix(~gl(8, 10))
    thing <- ResidualMatrix::ResidualMatrix(y, design=design)

    # Round-trips properly.
    temp <- saveDelayed(thing)
    out <- loadDelayed(temp)
    out@seed@.matrix <- as(out@seed@.matrix, "dgCMatrix")
    expect_identical(thing, out)
    expect_s4_class(out, "ResidualMatrix")

    # Works when transposed.
    thing2 <- t(thing)
    temp2 <- saveDelayed(thing2)
    out <- loadDelayed(temp2)
    out@seed@.matrix <- as(out@seed@.matrix, "dgCMatrix")
    expect_identical(thing2, out)
    expect_s4_class(out, "ResidualMatrix")

    # Same result if we ignore the type hint.
    (function() {
        fhandle <- H5Fopen(file.path(temp, "array.h5"), "H5F_ACC_RDWR")
        on.exit(H5Fclose(fhandle))
        ghandle <- H5Gopen(fhandle, "delayed_array")
        on.exit(H5Gclose(ghandle))
        H5Ldelete(ghandle, "_r_type_hint")
    })()

    out <- loadDelayed(temp)
    expect_false(is(out, "ResidualMatrix"))
    expect_identical(unname(as.matrix(thing)), unname(as.matrix(out)))

    (function() {
        fhandle <- H5Fopen(file.path(temp2, "array.h5"), "H5F_ACC_RDWR")
        on.exit(H5Fclose(fhandle))
        ghandle <- H5Gopen(fhandle, "delayed_array")
        on.exit(H5Gclose(ghandle))
        H5Ldelete(ghandle, "_r_type_hint")
    })()

    out <- loadDelayed(temp2)
    expect_false(is(out, "ResidualMatrix"))
    expect_identical(unname(as.matrix(thing2)), unname(as.matrix(out)))
})

#######################################################
#######################################################

test_that("readDelayedObject registry works as expected", {
    expect_false(is.null(reloadDelayedObjectFunctionRegistry("array")[["constant array"]])) 

    expect_null(reloadDelayedObjectFunctionRegistry("array")[["foobar"]]) 
    registerReloadDelayedObjectFunction("array", "foobar", function(...) "WHEE")
    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "WHEE")

    expect_error(registerReloadDelayedObjectFunction("array", "foobar", function(...) "WHEE", "error"), "already been registered")
    registerReloadDelayedObjectFunction("array", "foobar", function(...) "FART", "old")
    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "WHEE")
    registerReloadDelayedObjectFunction("array", "foobar", function(...) "FART", "new")
    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "FART")

    registerReloadDelayedObjectFunction("array", "foobar", NULL)
    expect_null(reloadDelayedObjectFunctionRegistry("array")[["foobar"]]) 
})

test_that("alternative delayed object function overrides work", {
    old <- altStoreDelayedObjectFunction()
    altStoreDelayedObjectFunction(function() "whee")
    on.exit(altStoreDelayedObjectFunction(old), after=FALSE, add=TRUE)
    expect_identical(altStoreDelayedObject(), "whee")

    old <- altReloadDelayedObjectFunction()
    altReloadDelayedObjectFunction(function() "whee")
    on.exit(altReloadDelayedObjectFunction(old), after=FALSE, add=TRUE)
    expect_identical(altReloadDelayedObject(), "whee")
})

#######################################################
#######################################################

test_that("external saving handles seeds without saveObject methods", {
    X <- DelayedArray(new("SuperSeed", dim=c(100L, 50L))) + 100

    temp <- saveDelayed(X)
    expect_true(file.exists(file.path(temp, "seeds", 0)))
    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
    expect_identical(as.matrix(X), as.matrix(roundtrip))
})

test_that("external saving avoids over-nesting of ReloadedArraySeed objects", {
    X <- DelayedArray(new("SuperSeed", dim=c(100L, 50L)))

    temp <- saveDelayed(X)
    expect_true(file.exists(file.path(temp, "seeds", 0)))
    roundtrip <- readObject(temp)

    expect_s4_class(roundtrip@seed, "ReloadedArraySeed")
    expect_identical(roundtrip@seed@path, normalizePath(temp))
    expect_false(is(roundtrip@seed@seed, "ReloadedArraySeed"))

    expect_identical(as.matrix(X), as.matrix(roundtrip))
})

test_that("external deduplication is done correctly", {
    dedup.session <- createDedupSession()

    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
    temp <- saveDelayed(X, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session)
    expect_true(file.exists(file.path(temp, "seeds", 0)))
    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
    expect_identical(X, roundtrip)

    Y <- X + 1
    temp2a <- saveDelayed(Y, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session, array.dedup.action="link")
    expect_true(file.exists(file.path(temp2a, "seeds", 0)))
    roundtrip <- loadDelayed(temp2a, custom.takane.realize=TRUE)
    expect_equal(Y, roundtrip)

    temp2b <- saveDelayed(Y, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session, array.dedup.action="copy")
    expect_true(file.exists(file.path(temp2b, "seeds", 0)))
    roundtrip <- loadDelayed(temp2b, custom.takane.realize=TRUE)
    expect_equal(Y, roundtrip)

    Z <- DelayedArray(matrix(rpois(30, 5), ncol=5)) # checking that a different array doesn't trigger the deduplicator.
    temp3 <- saveDelayed(Z, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session)
    expect_true(file.exists(file.path(temp3, "seeds", 0)))
    roundtrip <- loadDelayed(temp3, custom.takane.realize=TRUE)
    expect_identical(Z, roundtrip)

    if (.Platform$OS.type=="unix") { 
        expect_identical(Sys.readlink(file.path(temp, "seeds", "0", "OBJECT")), "")
        expect_identical(Sys.readlink(file.path(temp2a, "seeds", "0", "OBJECT")), "")
        expect_identical(Sys.readlink(file.path(temp2b, "seeds", "0", "OBJECT")), "")
        expect_identical(Sys.readlink(file.path(temp3, "seeds", "0", "OBJECT")), "")

        temp2c <- saveDelayed(Y, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session, array.dedup.action="symlink")
        expect_true(file.exists(file.path(temp2c, "seeds", 0)))
        expect_true(startsWith(Sys.readlink(file.path(temp2c, "seeds", "0", "OBJECT")), "/"))
        roundtrip <- loadDelayed(temp2c, custom.takane.realize=TRUE)
        expect_equal(Y, roundtrip)

        temp2d <- saveDelayed(Y, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session, array.dedup.action="relsymlink")
        expect_true(file.exists(file.path(temp2d, "seeds", 0)))
        expect_true(startsWith(Sys.readlink(file.path(temp2d, "seeds", "0", "OBJECT")), "../"))
        roundtrip <- loadDelayed(temp2d, custom.takane.realize=TRUE)
        expect_equal(Y, roundtrip)

        # Back-compatibility check.
        temp2e <- saveDelayed(Y, DelayedArray.store.args=list(save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="symlink"))
        expect_true(file.exists(file.path(temp2e, "seeds", 0)))
        expect_true(startsWith(Sys.readlink(file.path(temp2e, "seeds", "0", "OBJECT")), "/"))
        roundtrip <- loadDelayed(temp2e, custom.takane.realize=TRUE)
        expect_equal(Y, roundtrip)

        temp3b <- saveDelayed(Z, DelayedArray.force.external=TRUE, array.dedup.session=dedup.session, array.dedup.action="symlink")
        expect_true(file.exists(file.path(temp3b, "seeds", 0)))
        expect_true(startsWith(Sys.readlink(file.path(temp3b, "seeds", "0", "OBJECT")), "/"))
        roundtrip <- loadDelayed(temp3b, custom.takane.realize=TRUE)
        expect_equal(Z, roundtrip)
    }
})

#' @export
#' @import methods
setClass("WrapperArraySeed", contains=c("VIRTUAL", "DelayedUnaryIsoOp"), slots=c(seed="ANY"))

#' @export
#' @importClassesFrom DelayedArray DelayedArray
setClass("WrapperArray", contains=c("VIRTUAL", "DelayedArray"), slots=c(seed="WrapperArraySeed"))

#' @export
#' @importClassesFrom DelayedArray DelayedAbind
setClass("AmalgamatedArraySeed", contains="DelayedAbind", slots=c(samples = "character"))

#' @export
#' @importClassesFrom DelayedArray DelayedArray
setClass("AmalgamatedArray", contains="DelayedArray", slots=c(seed = "AmalgamatedArraySeed"))

#' @export
#' @importClassesFrom DelayedArray DelayedMatrix
setClass("AmalgamatedMatrix", contains=c("AmalgamatedArray", "DelayedMatrix"))

#' @export
setClass("ReloadedArraySeed", contains="DelayedUnaryIsoOp", slots=c(path="character"))

#' @export
setClass("ReloadedArray", contains="DelayedArray", slots=c(seed="ReloadedArraySeed"))

#' @export
#' @importClassesFrom DelayedArray DelayedMatrix
setClass("ReloadedMatrix", contains=c("ReloadedArray", "DelayedMatrix"))

optimize_storage <- function(x) {
    tt <- type(x)
    if (tt == "character") {
        optimize_string_storage(x)
    } else if (tt == "double") {
        optimize_float_storage(x)
    } else if (tt == "integer") {
        optimize_integer_storage(x)
    } else if (tt == "logical") {
        optimize_boolean_storage(x)
    } else {
        stop("unsupported type '", tt, "'")
    }
}

###################################################
###################################################

aggregate_range <- function(collated, name) {
    c(
        min(unlist(lapply(collated, function(y) y[[name]][1]))),
        max(unlist(lapply(collated, function(y) y[[name]][2])))
    )
}

aggregate_any <- function(collated, name) {
    any(vapply(collated, function(y) y[[name]], TRUE))
}

aggregate_max <- function(collated, name) {
    max(unlist(lapply(collated, function(y) y[[name]])), na.rm=TRUE)
}

aggregate_sum <- function(collated, name) {
    sum(vapply(collated, function(y) y[[name]], 0L))
}

collect_from_SVT <- function(x, fun, tt) {
    svt <- x@SVT
    if (is.null(svt)) {
        attrs <- fun(tt(0))
        attrs$non_zero <- 0L
        return(list(attrs))
    }

    # In versions >= 1L, the values are the first element of each node.
    version <- x@.svt_version
    val.index <- if (version == 1L) 1L else 2L
    idx.index <- if (version == 1L) 2L else 1L

    none <- tt(0)
    one <- as(1, type(x))

    output <- vector("list", length(svt))
    for (i in seq_along(svt)) {
        node <- svt[[i]] 
        if (is.null(node)) {
            val <- none
        } else {
            val <- node[[val.index]]
        }

        # Handle lacunar leaf nodes.
        if (is.null(val)) {
            val <- rep(one, length(node[[idx.index]]))
        }

        attrs <- fun(val)
        attrs$non_zero <- length(val)
        output[[i]] <- attrs
    }
    output
}

###################################################
###################################################

setGeneric("collect_integer_attributes", function(x) standardGeneric("collect_integer_attributes"))

.simple_integer_collector <- function(x) {
    list(
        range=suppressWarnings(range(x, na.rm=TRUE)),
        missing=anyNA(x)
    )
}

.combine_integer_attributes <- function(collated) {
    list(
        range=aggregate_range(collated, "range"),
        missing=aggregate_any(collated, "missing")
    )
}

setMethod("collect_integer_attributes", "array", .simple_integer_collector)

setMethod("collect_integer_attributes", "SVT_SparseMatrix", function(x) {
    collated <- collect_from_SVT(x, .simple_integer_collector, integer)
    output <- list(non_zero=aggregate_sum(collated, "non_zero"))
    c(output, .combine_integer_attributes(collated))
})

#' @importFrom S4Arrays is_sparse
#' @importFrom SparseArray nzvals
setMethod("collect_integer_attributes", "ANY", function(x) {
    output <- list()

    if (is_sparse(x)) {
        collated <- blockApply(x, function(y) {
            y_nzvals <- nzvals(y)
            out <- .simple_integer_collector(y_nzvals)
            out$non_zero <- length(y_nzvals)
            out
        }, as.sparse=TRUE)
        output$non_zero <- aggregate_sum(collated, "non_zero")
    } else {
        collated <- blockApply(x, .simple_integer_collector)
    }

    c(output, .combine_integer_attributes(collated))
})

optimize_integer_storage <- function(x) {
    attr <- collect_integer_attributes(x)

    if (attr$missing) {
        lower <- attr$range[1]
        upper <- attr$range[2]

        # If it's infinite, that means that there are only missing values in
        # 'x', otherwise there should have been at least one finite value
        # available. In any case, it means we can just do whatever we want so
        # we'll just use the smallest type.
        if (is.infinite(lower)) {
            return(list(type="H5T_NATIVE_INT8", placeholder=as.integer(-2^7), size=attr$non_zero))
        }

        if (lower < 0L) {
            if (lower > -2^7 && upper < 2^7) {
                return(list(type="H5T_NATIVE_INT8", placeholder=as.integer(-2^7), size=attr$non_zero))
            } else if (lower > -2^15 && upper < 2^15) {
                return(list(type="H5T_NATIVE_INT16", placeholder=as.integer(-2^15), size=attr$non_zero))
            }
        } else {
            if (upper < 2^8 - 1) {
                return(list(type="H5T_NATIVE_UINT8", placeholder=as.integer(2^8-1), size=attr$non_zero))
            } else if (upper < 2^16 - 1) {
                return(list(type="H5T_NATIVE_UINT16", placeholder=as.integer(2^16-1), size=attr$non_zero))
            }
        }

        return(list(type="H5T_NATIVE_INT32", placeholder=NA_integer_, size=attr$non_zero))

    } else {
        lower <- attr$range[1]
        upper <- attr$range[2]

        # If it's infinite, that means that 'x' is of length zero, otherwise
        # there should have been at least one finite value available. Here,
        # the type doesn't matter, so we'll just use the smallest. 
        if (is.infinite(lower)) {
            return(list(type="H5T_NATIVE_INT8", placeholder=NULL, size=attr$non_zero))
        }

        if (lower < 0L) {
            if (lower >= -2^7 && upper < 2^7) {
                return(list(type="H5T_NATIVE_INT8", placeholder=NULL, size=attr$non_zero))
            } else if (lower >= -2^15 && upper < 2^15) {
                return(list(type="H5T_NATIVE_INT16", placeholder=NULL, size=attr$non_zero))
            }
        } else {
            if (upper < 2^8) {
                return(list(type="H5T_NATIVE_UINT8", placeholder=NULL, size=attr$non_zero))
            } else if (upper < 2^16) {
                return(list(type="H5T_NATIVE_UINT16", placeholder=NULL, size=attr$non_zero))
            }
        }

        return(list(type="H5T_NATIVE_INT32", placeholder=NULL, size=attr$non_zero))
    }
}

###################################################
###################################################

setGeneric("collect_float_attributes", function(x) standardGeneric("collect_float_attributes"))

setMethod("collect_float_attributes", "array", collect_double_attributes)

setMethod("collect_float_attributes", "ddenseMatrix", function(x) collect_double_attributes(x@x))

setMethod("collect_float_attributes", "dsparseMatrix", function(x) {
    out <- collect_double_attributes(x@x)
    out$non_zero <- length(x@x)
    out
})

.combine_float_attributes <- function(collated) {
    output <- list(range=aggregate_range(collated, "range"))
    for (n in c("missing", "non_integer", "has_NaN", "has_Inf", "has_nInf", "has_lowest", "has_highest")) {
        output[[n]] <- aggregate_any(collated, n)
    }
    output
}

setMethod("collect_float_attributes", "SVT_SparseMatrix", function(x) {
    collated <- collect_from_SVT(x, collect_double_attributes, double)
    output <- list(non_zero=aggregate_sum(collated, "non_zero"))
    c(output, .combine_float_attributes(collated))
})

#' @importFrom S4Arrays is_sparse
#' @importFrom SparseArray nzvals
setMethod("collect_float_attributes", "ANY", function(x) {
    output <- list()
    if (is_sparse(x)) {
        collated <- blockApply(x, function(y) {
            y_nzvals <- nzvals(y)
            out <- collect_double_attributes(y_nzvals)
            out$non_zero <- length(y_nzvals)
            out
        }, as.sparse=TRUE)
        output$non_zero <- aggregate_sum(collated, "non_zero")
    } else {
        collated <- blockApply(x, collect_double_attributes)
    }
    c(output, .combine_float_attributes(collated))
})

optimize_float_storage <- function(x) {
    attr <- collect_float_attributes(x)

    if (attr$missing) {
        if (!attr$non_integer) {
            lower <- attr$range[1]
            upper <- attr$range[2]
            if (lower < 0L) {
                if (lower > -2^7 && upper < 2^7) {
                    return(list(type="H5T_NATIVE_INT8", placeholder=-2^7, size=attr$non_zero))
                } else if (lower > -2^15 && upper < 2^15) {
                    return(list(type="H5T_NATIVE_INT16", placeholder=-2^15, size=attr$non_zero))
                } else if (lower > -2^31 && upper < 2^31) {
                    return(list(type="H5T_NATIVE_INT32", placeholder=-2^31, size=attr$non_zero))
                }
            } else {
                if (upper < 2^8-1) {
                    return(list(type="H5T_NATIVE_UINT8", placeholder=2^8-1, size=attr$non_zero))
                } else if (upper < 2^16-1) {
                    return(list(type="H5T_NATIVE_UINT16", placeholder=2^16-1, size=attr$non_zero))
                } else if (upper < 2^32-1) {
                    return(list(type="H5T_NATIVE_UINT32", placeholder=2^32-1, size=attr$non_zero))
                }
            }
        }

        placeholder <- NULL
        if (!attr$has_NaN) {
            placeholder <- NaN
        } else if (!attr$has_Inf) {
            placeholder <- Inf
        } else if (!attr$has_nInf) {
            placeholder <- -Inf
        } else if (!attr$has_lowest) {
            placeholder <- lowest_double()
        } else if (!attr$has_highest) {
            placeholder <- highest_double()
        }

        # Fallback that just goes through and pulls out all unique values.
        if (is.null(placeholder)) {
            if (is_sparse(x)) {
                u <- Reduce(union, blockApply(x, function(y) unique(nzvals(y))))
            } else {
                u <- Reduce(union, blockApply(x, function(y) unique(as.vector(y))))
            }
            placeholder <- chooseMissingPlaceholderForHdf5(u)
        }

        return(list(type="H5T_NATIVE_DOUBLE", placeholder=placeholder, size=attr$non_zero))

    } else {
        if (!attr$non_integer) {
            lower <- attr$range[1]
            upper <- attr$range[2]
            if (lower < 0L) {
                if (lower >= -2^7 && upper < 2^7) {
                    return(list(type="H5T_NATIVE_INT8", placeholder=NULL, size=attr$non_zero))
                } else if (lower >= -2^15 && upper < 2^15) {
                    return(list(type="H5T_NATIVE_INT16", placeholder=NULL, size=attr$non_zero))
                } else if (lower >= -2^31 && upper < 2^31) {
                    return(list(type="H5T_NATIVE_INT32", placeholder=NULL, size=attr$non_zero))
                }
            } else {
                if (upper < 2^8) {
                    return(list(type="H5T_NATIVE_UINT8", placeholder=NULL, size=attr$non_zero))
                } else if (upper < 2^16) {
                    return(list(type="H5T_NATIVE_UINT16", placeholder=NULL, size=attr$non_zero))
                } else if (upper < 2^32) {
                    return(list(type="H5T_NATIVE_UINT32", placeholder=NULL, size=attr$non_zero))
                }
            }
        }

        return(list(type="H5T_NATIVE_DOUBLE", placeholder=NULL, size=attr$non_zero))
    }
}

###################################################
###################################################

setGeneric("collect_string_attributes", function(x) standardGeneric("collect_string_attributes"))

setMethod("collect_string_attributes", "ANY", function(x) {
    collected <- blockApply(x, function(y) {
        strlen <- nchar(y, "bytes")
        list(
            has_na1=any(y == "NA", na.rm=TRUE),
            has_na2=any(y == "_NA", na.rm=TRUE),
            max_len=suppressWarnings(max(strlen, na.rm=TRUE)),
            total_len=sum(strlen, na.rm=TRUE),
            num_missing=sum(is.na(y)),
            encoding=unique(Encoding(y))
        )
    })

    list(
        has_na1=aggregate_any(collected, "has_na1"),
        has_na2=aggregate_any(collected, "has_na2"),
        max_len=aggregate_max(collected, "max_len"),
        total_len=aggregate_sum(collected, "total_len"),
        num_missing=aggregate_sum(collected, "num_missing"),
        encoding=Reduce(union, lapply(collected, function(y) y$encoding))
    )
})

optimize_string_storage <- function(x) {
    attr <- collect_string_attributes(x)

    placeholder <- NULL
    if (attr$num_missing > 0L) {
        if (!attr$has_na1) {
            placeholder <- "NA"
        } else if (!attr$has_na2) {
            placeholder <- "_NA"
        } else {
            u <- Reduce(union, blockApply(x, function(y) unique(as.vector(y))))
            placeholder <- chooseMissingPlaceholderForHdf5(u)
        }
        plclen <- nchar(placeholder, "bytes")
        attr$max_len <- max(attr$max_len, plclen)
        attr$total_len <- attr$total_len + plclen * attr$num_missing
    }

    tid <- H5Tcopy("H5T_C_S1")
    H5Tset_strpad(tid, strpad = "NULLPAD")
    H5Tset_size(tid, max(1L, attr$max_len))
    if ("UTF-8" %in% attr$encoding) {
        H5Tset_cset(tid, "UTF8")
    } else {
        H5Tset_cset(tid, "ASCII")
    }

    list(type=tid, placeholder=placeholder, max_len=attr$max_len, total_len=attr$total_len)
}

###################################################
###################################################

setGeneric("collect_boolean_attributes", function(x) standardGeneric("collect_boolean_attributes"))

setMethod("collect_boolean_attributes", "ANY", function(x) {
    output <- list()
    if (is_sparse(x)) {
        collated <- blockApply(x, function(x) {
            x_nzvals <- nzvals(x)
            list(missing=anyNA(x_nzvals), non_zero=length(x_nzvals))
        }, as.sparse=TRUE)
        output$non_zero <- aggregate_sum(collated, "non_zero")
    } else {
        collated <- list(list(missing=anyNA(x)))
    }
    output$missing <- aggregate_any(collated, "missing")
    output
})

setMethod("collect_boolean_attributes", "lsparseMatrix", function(x) {
    list(missing=anyNA(x), non_zero=length(x@x))
})

setMethod("collect_boolean_attributes", "SVT_SparseMatrix", function(x) {
    collated <- collect_from_SVT(x, function(vals) { list(missing=anyNA(vals)) }, logical)
    list(
        missing=aggregate_any(collated, "missing"),
        non_zero=aggregate_sum(collated, "non_zero")
    )
})

optimize_boolean_storage <- function(x) {
    attr <- collect_boolean_attributes(x)
    if (attr$missing) {
        list(type="H5T_NATIVE_INT8", placeholder=-1L, size=attr$non_zero)
    } else {
        list(type="H5T_NATIVE_INT8", placeholder=NULL, size=attr$non_zero)
    }
}

#' Read a delayed array from disk
#'
#' Read a delayed high-dimensional array from its on-disk representation.
#' This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
#'
#' @param path String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a delayed array.
#' @param metadata Named list of metadata for this object, see \code{\link{readObject}} for more details.
#' @param delayed_array.reload.args Named list of arguments to be passed to \code{\link{reloadDelayedObject}}.
#' @param ... Further arguments, ignored.
#' 
#' @return A multi-dimensional array-like object.
#'
#' @seealso
#' \code{"\link{saveObject,DelayedArray-method}"}, to create the directory and its contents.
#'
#' \code{\link{reloadDelayedObject}}, for the methods to reload each delayed operation.
#'
#' @author Aaron Lun
#'
#' @examples
#' arr <- array(rpois(10000, 10), c(50, 20, 10))
#' dimnames(arr) <- list(
#'    paste0("GENE_", seq_len(nrow(arr))),
#'    letters[1:20],
#'    NULL
#' )
#'
#' dir <- tempfile()
#' saveObject(arr, dir)
#' readObject(dir)
#' 
#' @export
readDelayedArray <- function(path, metadata, delayed_array.reload.args=list(), ...) {
    fpath <- file.path(path, "array.h5")
    fhandle <- H5Fopen(fpath, "H5F_ACC_RDONLY")
    on.exit(H5Fclose(fhandle))
    out <- do.call(altReloadDelayedObject, c(list(fhandle, "delayed_array"), delayed_array.reload.args))

    # Avoid excessive nesting of ReloadedArrays from altReloadDelayedObject,
    # as we only need the outermost ReloadedArray to track the 'path'.
    if (is(out, "ReloadedArray")) {
        out <- out@seed
        while (is(out, "ReloadedArraySeed")) {
            out <- out@seed
        }
    }
    ReloadedArray(path=path, seed=out)
}

#' Amalgamated array class
#'
#' Implements an amalgamated array, equivalent to a delayed combination of DelayedArray objects.
#' It allows \code{\link{stageObject}} to save a combination of multiple matrices without actually aggregating their data into a single file.
#'
#' @section Constructors:
#' \code{AmalgamatedArraySeed(..., along=1)} accepts any number of named array-like objects and returns a AmalgamatedArraySeed.
#' Each object corresponds to a block and should be named accordingly; names should be unique and non-empty.
#' The \code{along} argument specifies the dimension in which matrices should be combined - for matrices, this is 1 for rows, 2 for columns.
#'
#' \code{AmalgamatedArray(..., along=1)} accepts any number of named array-like objects and returns a AmalgamatedArray.
#' Alternatively, a single AmalgamatedArraySeed may be provided in \code{...}. 
#' 
#' @section Functions:
#' \code{componentNames(x)} will return a character vector of names of component arrays in a AmalgamatedArray(Seed) object \code{x}.
#'
#' \code{extractComponents(x)} will return a named list of array-like objects,
#' corresponding to the component arrays used to construct the AmalgamatedArray(Seed) object \code{x}.
#'
#' \code{\link{stageObject}(x, dir, path, child = FALSE)} will save the AmalgamatedArray \code{x} and its components into the \code{path} inside \code{dir}.
#' Each component array is staged into its own subdirectory inside \code{path}.
#'
#' @section Comments on usage:
#' The AmalgamatedArraySeed is closely related to (and in fact, is a subclass of) the \linkS4class{DelayedAbind} class.
#' This means that we can leverage many of the \pkg{DelayedArray} methods for handling the delayed bind.
#' In theory, we could just use a DelayedAbind directly and save it with \pkg{chihaya} in \code{\link{stageObject}} (via \code{\link{preserveDelayedOperations}(TRUE)}).
#' However, this provides fewer opportunities for tracking and manipulating the samples.
#' It also saves the per-sample matrices into a single file, which eliminates possibilities for per-file deduplication and linking, e.g., with \code{\link{recycleHdf5Files}(TRUE)}.
#' 
#' @author Aaron Lun
#' @examples
#' first <- Matrix::rsparsematrix(10, 10, 0.1)
#' second <- Matrix::rsparsematrix(10, 20, 0.1)
#' mat <- AmalgamatedArray(list(foo = first, bar = second), along=2)
#' mat
#'
#' componentNames(mat)
#' out <- extractComponents(mat)
#' lapply(out, dim)
#'
#' @aliases
#' AmalgamatedArray
#' AmalgamatedArray-class
#' AmalgamatedMatrix-class
#' AmalgamatedArraySeed
#' AmalgamatedArraySeed-class
#' DelayedArray,AmalgamatedArraySeed-method
#' matrixClass,AmalgamatedArray-method
#' componentNames
#' extractComponents
#' stageObject,AmalgamatedArray-method
#' coerce,AmalgamatedArray,AmalgamatedMatrix-method
#' coerce,AmalgamatedMatrix,AmalgamatedArray-method
#' 
#' @name AmalgamatedArray
NULL

#' @export
#' @importFrom S4Arrays arbind acbind
AmalgamatedArraySeed <- function(components, along = 1) {
    sample.names <- names(components)
    if (anyDuplicated(sample.names) || any(sample.names == "")) {
        stop("sample names must be unique and non-empty in a AmalgamatedArraySeed")
    }

    FUN <- if (along == 1) arbind else acbind
    combined <- do.call(FUN, lapply(components, DelayedArray))
    new("AmalgamatedArraySeed", combined@seed, samples = sample.names)
}

#' @export
#' @importFrom DelayedArray DelayedArray new_DelayedArray
setMethod("DelayedArray", "AmalgamatedArraySeed", function(seed) new_DelayedArray(seed, Class="AmalgamatedArray"))

#' @export
#' @importFrom DelayedArray matrixClass
setMethod("matrixClass", "AmalgamatedArray", function(x) "AmalgamatedMatrix")

# Overrides copied from DelayedArray::ConstantArray.
#' @importFrom S4Vectors new2
setAs("AmalgamatedArray", "AmalgamatedMatrix", function(from) new2("AmalgamatedMatrix", from))
setAs("AmalgamatedMatrix", "AmalgamatedArray", function(from) from)

#' @export
AmalgamatedArray <- function(components, along = 1) {
    if (!is(components, "AmalgamatedArraySeed")) {
        seed <- AmalgamatedArraySeed(components, along = along)
    } else {
        seed <- components
    }
    DelayedArray(seed)
}

#' @export
componentNames <- function(x) {
    if (is(x, "DelayedMatrix")) {
        x <- x@seed
    }
    x@samples
}

#' @export
extractComponents <- function(x) {
    if (is(x, "DelayedMatrix")) {
        x <- x@seed
    }
    out <- x@seeds
    names(out) <- x@samples
    out
}

#' @export
#' @importFrom alabaster.base .stageObject .writeMetadata
#' @importFrom DelayedArray DelayedArray
setMethod("stageObject", "AmalgamatedArray", function(x, dir, path, child = FALSE) {
    dir.create(file.path(dir, path), showWarnings=FALSE)

    seed <- x@seed
    seeds <- seed@seeds
    components <- vector("list", length(seeds))

    for (i in seq_along(seeds)) {
        current.seed <- seeds[[i]]
        out <- try(current.seed[0,0], silent=TRUE)
        if (is(out, "try-error")) {
            current.seed <- DelayedArray(current.seed)
        }
        meta <- .stageObject(current.seed, dir, paste0(path, "/component", i), child = TRUE)
        components[[i]] <- list(name = seed@samples[i], resource = .writeMetadata(meta, dir))
    }

    list(
        `$schema` = "amalgamated_array/v1.json",
        path = paste0(path, "/array.json"),
        is_child = child,
        `array` = .grab_array_type(x),
        amalgamated_array = list(
            along = x@seed@along - 1L,
            components = components
        )
    )
})

#' @importFrom alabaster.base acquireMetadata .loadObject
.load_amalgamated_array <- function(info, project) {
    comp <- info$amalgamated_array$components
    all.names <- character(length(comp))

    for (i in seq_along(comp)) {
        current <- comp[[i]]
        imeta <- acquireMetadata(project, current$resource$path)
        all.names[i] <- current$name
        comp[[i]] <- .loadObject(imeta, project)
    }

    names(comp) <- all.names
    AmalgamatedArraySeed(comp, along = info$amalgamated_array$along + 1L)
}

# Tests the AmalgamatedArray class.
# library(testthat); library(alabaster.matrix); source("test-AmalgamatedArray.R")

test_that("Amalgamated basics work as expected", {
    first <- Matrix::rsparsematrix(10, 10, 0.1)
    second <- Matrix::rsparsematrix(10, 20, 0.1)

    mat <- AmalgamatedArray(list(foo = first, bar = second), along=2)
    expect_s4_class(mat, "AmalgamatedArray")
    expect_s4_class(mat, "AmalgamatedMatrix")
    expect_identical(as.array(mat), as.array(BiocGenerics::cbind(first, second)))

    expect_identical(componentNames(mat), c("foo", "bar"))
    out <- extractComponents(mat)
    expect_identical(out$foo, first)
    expect_identical(out$bar, second)

    # Same for the other dimension.
    tsecond <- Matrix::t(second)
    mat <- AmalgamatedArray(list(stuff = first, whee = tsecond), along=1)
    expect_identical(as.array(mat), as.array(BiocGenerics::rbind(first, tsecond)))
})

library(alabaster.base)
test_that("Amalgamated staging and loading works as expected", {
    first <- Matrix::rsparsematrix(10, 10, 0.1)
    second <- Matrix::rsparsematrix(10, 20, 0.1)
    mat <- AmalgamatedArray(list(foo = first, bar = second), along=2)

    temp <- tempfile()
    dir.create(temp)
    meta <- stageObject(mat, temp, "amalgam")
    expect_identical(meta[["$schema"]], "amalgamated_array/v1.json")
    expect_identical(meta$amalgamated_array$along, 1L)
    info <- .writeMetadata(meta, temp)

    remeta <- acquireMetadata(temp, info$path)
    roundtrip <- loadObject(remeta, temp)
    expect_s4_class(roundtrip, "AmalgamatedArray")
    expect_identical(as.array(roundtrip), as.array(mat))
})

library(alabaster.base)
test_that("Amalgamated staging and loading works when the inner arrays are delayed wrappers", {
    first <- DelayedArray(Matrix::rsparsematrix(10, 10, 0.1)) * 5
    second <- DelayedArray(Matrix::rsparsematrix(10, 20, 0.1))
    colnames(second) <- letters[1:20]
    mat <- AmalgamatedArray(list(foo = first, bar = second), along=2)

    temp <- tempfile()
    dir.create(temp)
    meta <- stageObject(mat, temp, "amalgam")
    expect_identical(meta[["$schema"]], "amalgamated_array/v1.json")
    expect_identical(meta$amalgamated_array$along, 1L)
    info <- .writeMetadata(meta, temp)

    remeta <- acquireMetadata(temp, info$path)
    roundtrip <- loadObject(remeta, temp)
    expect_s4_class(roundtrip, "AmalgamatedArray")
    expect_identical(as.array(roundtrip), as.array(mat))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

lowest_double <- function() {
    .Call(`_alabaster_matrix_lowest_double`)
}

highest_double <- function() {
    .Call(`_alabaster_matrix_highest_double`)
}

collect_double_attributes <- function(x) {
    .Call(`_alabaster_matrix_collect_double_attributes`, x)
}

# This checks that the storage optimization works as expected.
# library(testthat); library(alabaster.matrix); source("test-optimize_storage.R", encoding="UTF-8")

library(DelayedArray)
library(rhdf5)

test_that("storage optimization works for integers, no missing values", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        # < 8-bit.
        mat <- matrix(seq(0, 255, length.out=1000), 50, 20)
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT8")
        expect_null(out$placeholder)
        out <- alabaster.matrix:::optimize_storage(fun(mat - 128L))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        out <- alabaster.matrix:::optimize_storage(fun(mat + 1L))
        expect_equal(out$type, "H5T_NATIVE_UINT16")

        # < 16-bit.
        mat <- matrix(seq(0, 65535, length.out=1000), 50, 20)
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_null(out$placeholder)
        out <- alabaster.matrix:::optimize_storage(fun(mat - 32768L))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        out <- alabaster.matrix:::optimize_storage(fun(mat + 1L))
        expect_equal(out$type, "H5T_NATIVE_INT32")

        # < 32-bit.
        mat <- matrix(seq(0, .Machine$integer.max, length.out=1000), 50, 20)
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_null(out$placeholder)
    }
})

test_that("storage optimization works for integers, plus missing values", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        # < 8-bit, unsigned.
        mat <- matrix(seq(0, 255, length.out=1000), 50, 20)
        mat[1000] <- NA
        storage.mode(mat) <- "integer" # Note that cast to int does a truncation, so there's only ever one value at a non-zero extreme.

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT8")
        expect_equal(out$placeholder, 255L)

        mat[999] <- 255L
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_equal(out$placeholder, 65535L)

        # 8-bit, signed.
        mat <- matrix(seq(-128, 127, length.out=1000), 50, 20)
        mat[1] <- NA
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        expect_equal(out$placeholder, -128L)

        mat[2] <- -128L
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        expect_equal(out$placeholder, -32768L)

        # < 16-bit, unsigned.
        mat <- matrix(seq(0, 65535, length.out=1000), 50, 20)
        mat[1000] <- NA
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_equal(out$placeholder, 65535)

        mat[999] <- 65535L 
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_equal(out$placeholder, NA_integer_)

        # 16-bit, signed.
        mat <- matrix(seq(-32768, 32767, length.out=1000), 50, 20)
        mat[1] <- NA
        storage.mode(mat) <- "integer"

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        expect_equal(out$placeholder, -32768L)

        mat[2] <- -32768L
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_equal(out$placeholder, NA_integer_)

        # Everything else.
        mat <- matrix(seq(0, .Machine$integer.max, length.out=1000), 50, 20)
        storage.mode(mat) <- "integer"
        mat[1] <- NA
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_equal(out$placeholder, NA_integer_)
    }
})

test_that("storage optimization works for integer-like doubles, no missing values", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        # < 8-bit.
        mat <- round(matrix(seq(0, 255, length.out=1000), 50, 20))

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT8")
        expect_null(out$placeholder)
        out <- alabaster.matrix:::optimize_storage(fun(mat - 128L))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        out <- alabaster.matrix:::optimize_storage(fun(mat + 1L))
        expect_equal(out$type, "H5T_NATIVE_UINT16")

        # < 16-bit.
        mat <- round(matrix(seq(0, 65535, length.out=1000), 50, 20))

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_null(out$placeholder)
        out <- alabaster.matrix:::optimize_storage(fun(mat - 32768L))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        out <- alabaster.matrix:::optimize_storage(fun(mat + 1L))
        expect_equal(out$type, "H5T_NATIVE_UINT32")

        # < 32-bit.
        mat <- round(matrix(seq(0, 2^32-1, length.out=1000), 50, 20))

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT32")
        expect_null(out$placeholder)
        out <- alabaster.matrix:::optimize_storage(fun(mat - 2^31))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        out <- alabaster.matrix:::optimize_storage(fun(mat + 1L))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
    }
})

test_that("storage optimization works for integer-like doubles, plus missing values", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        # < 8-bit, unsigned
        mat <- round(matrix(seq(0, 255, length.out=1000), 50, 20))
        mat[mat == 255] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT8")
        expect_equal(out$placeholder, 255)

        mat[999] <- 255
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_equal(out$placeholder, 65535)

        # < 8-bit, signed
        mat <- round(matrix(seq(-128, 127, length.out=1000), 50, 20))
        mat[mat==-128] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        expect_equal(out$placeholder, -128)

        mat[2] <- -128 
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        expect_equal(out$placeholder, -32768)

        # < 16-bit, unsigned
        mat <- round(matrix(seq(0, 65535, length.out=1000), 50, 20))
        mat[mat==65535] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT16")
        expect_equal(out$placeholder, 65535)

        mat[999] <- 65535
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT32")
        expect_equal(out$placeholder, 2^32-1)

        # < 16-bit, signed 
        mat <- round(matrix(seq(-2^15, 2^15-1, length.out=1000), 50, 20))
        mat[mat==-32768] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT16")
        expect_equal(out$placeholder, -32768)

        mat[2] <- -32768
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_equal(out$placeholder, -2^31)

        # < 32-bit, unsigned
        mat <- round(matrix(seq(0, 2^32-1, length.out=1000), 50, 20))
        mat[mat==2^32-1] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_UINT32")
        expect_equal(out$placeholder, 2^32-1)

        mat[999] <- 2^32-1
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, NaN)

        # < 32-bit, signed
        mat <- round(matrix(seq(-2^31, 2^31-1, length.out=1000), 50, 20))
        mat[mat == -2^31] <- NA

        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT32")
        expect_equal(out$placeholder, -2^31)

        mat[999] <- -2^31
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, NaN)
    }
})

test_that("storage optimization works non-integer doubles", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        mat <- matrix(rnorm(1000), 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_null(out$placeholder)

        # Running through the gamut of missing value placeholders.
        mat <- matrix(rnorm(1000), 50, 20)
        mat[1] <- NA
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, NaN)

        mat[2] <- NaN
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, Inf)

        mat[3] <- Inf
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, -Inf)

        mat[4] <- -Inf
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, alabaster.matrix:::lowest_double())

        mat[5] <- alabaster.matrix:::lowest_double()
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_equal(out$placeholder, alabaster.matrix:::highest_double())

        mat[6] <- alabaster.matrix:::highest_double()
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_DOUBLE")
        expect_true(!any(mat==out$placeholder, na.rm=TRUE))
    }
})

test_that("storage optimization works for strings", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        mat <- matrix(sample(LETTERS, 1000, replace=TRUE), 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 1)
        expect_equal(H5Tget_cset(out$type), 0) # aka ASCII
        expect_equal(out$placeholder, NULL)

        mat[1] <- NA
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 2)
        expect_equal(out$placeholder, "NA")

        mat[2] <- "NA"
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 3)
        expect_equal(out$placeholder, "_NA")

        mat[3] <- "_NA"
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 4)
        expect_equal(out$placeholder, "__NA")

        # Correct size determination.
        mat <- matrix(sample(LETTERS, 1000, replace=TRUE), 50, 20)
        mat[1] <- "Aaron"
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 5) 

        mat <- matrix(sample(LETTERS, 1000, replace=TRUE), 50, 20)
        mat[1000] <- "Aaron"
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 5) 

        mat <- matrix("", 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 1) 

        # Checking for correct behavior in all-NA cases.
        mat <- matrix(NA_character_, 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 2) 
        expect_identical(out$placeholder, "NA")

        # Handles UTF-8. Assumes that this file was sourced as UTF-8.
        mat <- matrix("α", 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(H5Tget_size(out$type), 2) 
        if (Encoding(mat[1]) == "UTF-8") {
            expect_equal(H5Tget_cset(out$type), 1) # aka UTF-8
        }
    }
})

test_that("storage optimization works for booleans", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:2) {
        fun <- identity
        if (i == 1L) {
            fun <- DelayedArray
        }

        mat <- matrix(c(TRUE, FALSE), 50, 20)
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        expect_null(out$placeholder)

        mat[1] <- NA
        out <- alabaster.matrix:::optimize_storage(fun(mat))
        expect_equal(out$type, "H5T_NATIVE_INT8")
        expect_equal(out$placeholder, -1L)
    }
})

test_that("storage optimization works for sparse objects", {
    old <- getAutoBlockSize()
    setAutoBlockSize(400)
    on.exit(setAutoBlockSize(old))

    for (i in 1:3) {
        if (i == 1L) {
            fun <- function(x) as(x, "sparseMatrix")
        } else if (i == 2L) {
            fun <- function(x) as(x, "SVT_SparseMatrix")
        } else {
            fun <- function(x) DelayedArray(as(x, "COO_SparseArray"))
        }

        # Integer.
        {
             y <- matrix(0L, nrow=100, ncol=10)
             y[1+sample(999, 100)] <- 10000L
             y[1] <- NA

             out <- alabaster.matrix:::optimize_storage(fun(y))
             expect_identical(out$type, "H5T_NATIVE_UINT16")
             expect_equal(out$placeholder, 2^16-1)
             expect_identical(out$size, 101L)
        }

        # Boolean
        {
             y <- matrix(FALSE, nrow=100, ncol=10)
             y[1+sample(999, 50)] <- TRUE 
             y[1] <- NA

             out <- alabaster.matrix:::optimize_storage(fun(y))
             expect_identical(out$type, "H5T_NATIVE_INT8")
             expect_equal(out$placeholder, -1L)
             expect_identical(out$size, 51L)
        }

        # Number
        {
             y <- matrix(0, nrow=100, ncol=10)
             y[1+sample(999, 200)] <- runif(200)
             y[1] <- NA

             out <- alabaster.matrix:::optimize_storage(fun(y))
             expect_identical(out$type, "H5T_NATIVE_DOUBLE")
             expect_equal(out$placeholder, NaN)
             expect_identical(out$size, 201L)
        }
    }
})

#' Save DelayedArrays to disk
#'
#' Save \link{DelayedArray} objects to their on-disk representation.
#'
#' @param x A \linkS4class{DelayedArray} object.
#' @param path String containing a path to a directory in which to save \code{x}.
#' @param DelayedArray.dispatch.pristine Logical scalar indicating whether to call the \code{\link{saveObject}} methods of seeds of pristine arrays. 
#' @param DelayedArray.preserve.ops Logical scalar indicating whether delayed operations should be preserved on-disk.
#' @param DelayedArray.force.external Logical scalar indicating to save the seeds of \code{x} as external arrays.
#' This is passed directly to \code{storeDelayedObject} as the \code{save.external.array=} argument, see \code{?\link{storeDelayedObject}} for details.
#' @param DelayedArray.store.args More named arguments to pass to \code{\link{storeDelayedObject}}.
#' @param array.dedup.session,array.dedup.action Arguments controlling deduplication of \code{x}, see \code{?"\link{saveObject,array-method}"} for details.
#' If \code{x} is not a duplicate of an existing object, these arguments will be passed to further methods as described for \code{...}.
#' @param ... Further arguments passed to \code{storeDelayedObject} as \code{external.save.args}, if the delayed operations are to be preserved;
#' otherwise, they are passed to \code{\link{saveObject,array-method}} or \code{\link{saveObject,sparseMatrix-method}}.
#'
#' @details
#' Supplying \code{array.dedup.session=} by itself is only guaranteed to deduplicate \code{x} itself and may not deduplicate its seeds.
#' Users should combine this with \code{DelayedArray.force.external=TRUE} to force seeds to be saved via \code{saveObject},
#' which exposes the seeds to the deduplication machinery in their respective \code{saveObject} methods.
#' Check out \code{?"\link{storeDelayedObject}"} for more details.
#'
#' @return
#' \code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
#'
#' @author Aaron Lun
#' @examples
#' mat <- Matrix::rsparsematrix(100, 200, density=0.2)
#' rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
#' dmat <- DelayedArray::DelayedArray(mat) * 1
#'
#' dir <- tempfile()
#' saveObject(dmat, dir, delayed.preserve.ops=TRUE)
#' list.files(dir)
#'
#' @seealso
#' \code{\link{storeDelayedObject}}, for the methods to save each delayed operation.
#'
#' @name saveDelayedArray
#' @aliases 
#' stageObject,DelayedArray-method
#' stageObject,DelayedMatrix-method
NULL

#' @export
#' @rdname saveDelayedArray
#' @importFrom DelayedArray isPristine seed
setMethod("saveObject", "DelayedArray", function(
    x,
    path,
    DelayedArray.dispatch.pristine=TRUE,
    DelayedArray.preserve.ops=FALSE,
    DelayedArray.force.external=FALSE,
    DelayedArray.store.args=list(),
    array.dedup.session=NULL,
    array.dedup.action=NULL,
    ...)
{
    if (DelayedArray.dispatch.pristine && isPristine(x)) {
        s <- seed(x)
        out <- try_altSaveObject(
            s,
            path=path,
            DelayedArray.dispatch.pristine=DelayedArray.dispatch.pristine,
            DelayedArray.preserve.ops=DelayedArray.preserve.ops,
            DelayedArray.store.args=DelayedArray.store.args,
            array.dedup.session=array.dedup.session,
            array.dedup.action=array.dedup.action,
            ...
        )
        if (out) {
            return(invisible(NULL))
        }
    }

    if (!DelayedArray.preserve.ops) {
        if (is_sparse(x)) {
            .save_compressed_sparse_matrix(x, path, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, ...)
        } else {
            .save_array(x, path, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, ...)
        }
        return(invisible(NULL))
    }

    # This needs to be after the .save_array, .save_compressed_sparse_matrix
    # calls, as we don't want to add the object to the dedup session and then
    # have it get picked up by checkObjectInDedupSession within those calls.
    if (!is.null(array.dedup.session)) {
        dedup.path <- checkObjectInDedupSession(x, array.dedup.session)
        if (!is.null(dedup.path)) {
            cloneDirectory(dedup.path, path, array.dedup.action)
            return(invisible(NULL))
        }
        addObjectToDedupSession(x, array.dedup.session, path)
    }

    dir.create(path)
    saveObjectFile(path, "delayed_array", list(delayed_array=list(version="1.0")))

    fhandle <- H5Fcreate(file.path(path, "array.h5"))
    on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)

    if (!("external.save.args" %in% names(DelayedArray.store.args))) {
        # Technically, it seems that we should pass along the various
        # DelayedArray.* arguments for any call to altSaveObject to save
        # external seeds. However, there's no point, because any call to
        # saveObject,DelayedArray-method from storeDelayedObject will not
        # preserve delayed ops, otherwise we'd get an infinite recursion.
        DelayedArray.store.args$external.save.args <- list(..., array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action)
    }
    if (!("save.external.array" %in% names(DelayedArray.store.args))) {
        DelayedArray.store.args$save.external.array <- DelayedArray.force.external
    }
    do.call(altStoreDelayedObject, c(list(x@seed, handle=fhandle, name="delayed_array"), DelayedArray.store.args))

    ghandle <- H5Gopen(fhandle, "delayed_array")
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
    h5_write_attribute(ghandle, "delayed_version", "1.1")

    invisible(NULL)
})

#' Delayed masking
#'
#' Delayed masking of missing values, based on replacement of placeholder values with \code{NA}.
#' This allows missingness to be encoded in frameworks without the same concept of \code{NA} as R.
#'
#' @param x An existing \pkg{DelayedArray} seed. 
#' @param placeholder Placeholder value to replace with \code{NA}.
#' This should be of the same type as \code{\link{type}(x)}.
#'
#' @details
#' If \code{\link{is.na}(placeholder)} is true for double-precision \code{x}, masking is performed for all values of \code{x} where \code{is.na} is true.
#' This includes both NaNs and NAs; no attempt is made to distinguish between the NaN payloads.
#'
#' Currently, an error is raised for any integer \code{x} that produces non-missing values of -2^31 without a placeholder of \code{NA_integer_}.
#' This is because R cannot distinguish the integer -2^31 from an integer-type NA.
#'
#' @return A DelayedMask object, to be wrapped in a \code{\link{DelayedArray}}.
#'
#' @author Aaron Lun
#'
#' @aliases
#' DelayedMask-class
#' dim,DelayedMask-method
#' dimnames,DelayedMask-method
#' chunkdim,DelayedMask-method
#' path,DelayedMask-method
#' is_sparse,DelayedMask-method
#' extract_array,DelayedMask-method
#' extract_sparse_array,DelayedMask-method
#'
#' @examples
#' original <- DelayedArray(matrix(rpois(40, lambda=2), ncol=5))
#' original
#' masked <- DelayedMask(original, 0)
#' DelayedArray(masked)
#'
#' @export
#' @import methods
#' @importClassesFrom DelayedArray DelayedArray
DelayedMask <- function(x, placeholder) {
    if (is(x, "DelayedArray")) {
        x <- x@seed
    }
    new("DelayedMask", seed=x, placeholder=placeholder)
}

#' @export
#' @importClassesFrom DelayedArray DelayedUnaryOp
setClass("DelayedMask", contains="DelayedUnaryOp", slots=c(seed="ANY", placeholder="ANY"))

#' @export
setMethod("dim", "DelayedMask", function(x) callGeneric(x@seed))

#' @export
setMethod("dimnames", "DelayedMask", function(x) callGeneric(x@seed))

#' @export
#' @importFrom S4Arrays is_sparse
setMethod("is_sparse", "DelayedMask", function(x) {
    if (is.finite(x@placeholder) && x@placeholder == 0) {
        return(FALSE)
    } else {
        return(callGeneric(x@seed))
    }
})

#' @export
#' @importFrom S4Arrays extract_array
setMethod("extract_array", "DelayedMask", function(x, index) {
    ans <- callGeneric(x@seed, index)
    .replace_missing(ans, x@placeholder)
})

#' @export
#' @importFrom SparseArray extract_sparse_array
setMethod("extract_sparse_array", "DelayedMask", function(x, index) {
    ans <- callGeneric(x@seed, index)
    if (is(ans, "COO_SparseArray")) {
        ans@nzdata <- .replace_missing(ans@nzdata, x@placeholder)
    } else {
        nzidx <- nzwhich(ans)
        ans[nzidx] <- .replace_missing(ans[nzidx], x@placeholder)
    }
    ans
})

.replace_missing <- function(vec, placeholder) {
    if (is.na(placeholder)) {
        if (anyNA(vec)) {
            if (is.double(vec)) {
                vec[is.na(vec)] <- NA # convert NaNs as well.
            } else {
                # no-op, everything is already NA.
            }
        }
    } else {
        if (anyNA(vec)) {
            if (is.integer(vec)) {
                # We don't want to incorrectly propagate -2^31 as NAs, so we
                # just throw an error here. Better to fail than to silently
                # modify the meaning of the data.
                stop("integer arrays containing both -2^31 and NA are not yet supported")
            } else {
                vec[which(vec == placeholder)] <- NA # "which()" avoids problems with existing NaNs.
            }
        } else {
            vec[vec == placeholder] <- NA
        }
    }
    vec
}

# Tests that the delayed masking works as expected.
# library(testthat); library(alabaster.matrix); source("test-DelayedMask.R")

library(DelayedArray)
set.seed(100000)

test_that("basic tests for DelayedMask", {
    original <- matrix(rpois(400, lambda=2), ncol=10)
    storage.mode(original) <- "integer"
    masked <- DelayedArray(DelayedMask(original, 0))
    expect_identical(dim(original), dim(masked))
    expect_identical(type(original), type(masked))
    expect_false(is_sparse(masked))

    ref <- original
    ref[ref == 0] <- NA
    expect_identical(ref, as.matrix(masked))

    # Other methods work as expected.
    y <- original
    rownames(y) <- sprintf("GENE_%i", seq_len(nrow(y)))
    masked <- DelayedArray(DelayedMask(y, 0))
    expect_identical(rownames(masked), rownames(y))
})

test_that("sparse tests for DelayedMask", {
    original <- round(Matrix::rsparsematrix(100, 20, density=0.2) * 10)
    masked <- DelayedArray(DelayedMask(original, 1))
    expect_true(is_sparse(masked))

    ref <- as.matrix(original)
    dimnames(ref) <- list(NULL, NULL)
    ref[ref == 1] <- NA
    expect_identical(ref, as.matrix(masked))
    dimnames(ref) <- NULL

    spmat <- extract_sparse_array(masked, list(NULL, NULL))
    expect_identical(ref, as.matrix(spmat))

    # Behaves correctly if zero is the placeholder.
    masked <- DelayedArray(DelayedMask(original, 0))
    expect_false(is_sparse(masked))

    ref <- as.matrix(original)
    ref[ref == 0] <- NA
    dimnames(ref) <- list(NULL, NULL)
    expect_identical(ref, as.matrix(masked))
})

test_that("DelayedMask works with special values", {
    original <- matrix(rpois(400, lambda=2), ncol=10)
    masked <- DelayedMask(original, NA)
    expect_identical(original, as.matrix(masked))

    # Also works with NaN values, whether the placeholder is NaN or not.
    original[1,1] <- NaN

    masked <- DelayedMask(original, 1)
    ref <- as.matrix(original)
    ref[ref == 1] <- NA
    expect_identical(ref, as.matrix(masked))

    masked <- DelayedMask(original, NaN)
    ref <- as.matrix(original)
    ref[is.nan(ref)] <- NA
    expect_identical(ref, as.matrix(masked))

    # If the placeholder is NA, both NaNs and NAs are considered to be missing,
    # as we don't rely on the payload (i.e., same logic as is.na()).
    masked <- DelayedMask(original, NA)
    ref <- as.matrix(original)
    ref[is.nan(ref)] <- NA
    expect_identical(ref, as.matrix(masked))

    # No problem if the dataset is integer.
    as.int <- original
    storage.mode(as.int) <- "integer"
    masked <- DelayedMask(as.int, NA_integer_)
    expect_identical(as.int, as.matrix(masked))
})

test_that("DelayedMask fails for integers with non-NA placeholders", {
    original <- matrix(NA_integer_, 10, 10)
    masked <- DelayedMask(original, NA)
    expect_identical(original, as.matrix(masked))

    masked <- DelayedMask(original, 10L)
    expect_error(as.matrix(masked), "not yet supported")
})

#' @export
setGeneric("loadWrapperArray", function(meta, project) standardGeneric("loadWrapperArray"), signature="project")

#' @export
#' @rdname storeDelayedObject
setGeneric("storeDelayedObject", function(x, handle, name, ...) standardGeneric("storeDelayedObject"))

#' Save a sparse matrix to disk
#'
#' Save a sparse matrix to its on-disk representations.
#'
#' @param x A sparse matrix of some kind, typically from either the \pkg{Matrix} or \pkg{SparseArray} packages.
#' @param path String containing the path to a directory in which to save \code{x}.
#' @param array.dedup.session,array.dedup.action Arguments controlling deduplication of \code{x}, see \code{?"\link{saveObject,array-method}"} for details.
#' @param ... Further arguments, currently ignored.
#' 
#' @return
#' \code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
#'
#' @seealso
#' \code{\link{readSparseMatrix}}, to read the directory contents back into the R session.
#'
#' @author Aaron Lun
#' @examples
#' mat <- Matrix::rsparsematrix(100, 200, density=0.2)
#' rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
#'
#' dir <- tempfile()
#' saveObject(mat, dir)
#' list.files(dir)
#'
#' @name saveSparseMatrix
NULL

.save_compressed_sparse_matrix <- function(x, path, array.dedup.session=NULL, array.dedup.action="link", ...) {
    if (!is.null(array.dedup.session)) {
        dedup.path <- checkObjectInDedupSession(x, array.dedup.session)
        if (!is.null(dedup.path)) {
            cloneDirectory(dedup.path, path, array.dedup.action)
            return(invisible(NULL))
        }
        addObjectToDedupSession(x, array.dedup.session, path)
    }

    dir.create(path)
    fpath <- file.path(path, "matrix.h5")
    name <- "compressed_sparse_matrix"

    fhandle <- H5Fcreate(fpath, "H5F_ACC_TRUNC")
    on.exit(H5Fclose(fhandle), add=TRUE, after=FALSE)

    ghandle <- H5Gcreate(fhandle, name)
    on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)

    h5_write_attribute(ghandle, "type", to_array_type(x), scalar=TRUE)
    h5_write_vector(ghandle, "shape", dim(x), type="H5T_NATIVE_UINT32")

    optimized <- optimize_storage(x)
    h5_write_sparse_matrix(x, handle=ghandle, details=optimized)

    if (!is.null(optimized$placeholder)) {
        dhandle <- H5Dopen(ghandle, "data")
        on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
        h5_write_attribute(dhandle, missingPlaceholderName, optimized$placeholder, type=optimized$type, scalar=TRUE)
    }

    save_names(ghandle, x)
    saveObjectFile(path, name, list(compressed_sparse_matrix=list(version="1.0")))
    invisible(NULL)
}

#' @export
#' @rdname saveSparseMatrix
setMethod("saveObject", "sparseMatrix", function(x, path, array.dedup.session=NULL, array.dedup.action="link", ...) {
    .save_compressed_sparse_matrix(x, path, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, ...)
})

#' @export
#' @rdname saveSparseMatrix
setMethod("saveObject", "SVT_SparseMatrix", function(x, path, array.dedup.session=NULL, array.dedup.action="link", ...) {
    .save_compressed_sparse_matrix(x, path, array.dedup.session=array.dedup.session, array.dedup.action=array.dedup.action, ...)
})

##############################
######### INTERNALS ##########
##############################

setGeneric("h5_write_sparse_matrix", function(x, ...) standardGeneric("h5_write_sparse_matrix"))

.deposit_sparse_data_with_placeholder <- function(handle, nzd, type, placeholder) {
    N <- length(nzd)
    chunks <- h5_guess_vector_chunks(N)
    dhandle <- h5_create_vector(handle, "data", N, chunks=chunks, type=type)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)

    shandle <- H5Screate_simple(N)
    on.exit(H5Sclose(shandle), add=TRUE, after=FALSE)
    mem_shandle <- H5Screate_simple(chunks)
    on.exit(H5Sclose(mem_shandle), add=TRUE, after=FALSE)

    pos <- 1L
    while (pos <= N) {
        size <- chunks
        end <- pos + size - 1L
        if (end > N) {
            size <- N - pos + 1L
            end <- N
            H5Sset_extent_simple(mem_shandle, size)
        }
        H5Sselect_hyperslab(shandle, "H5S_SELECT_SET", start=pos, count=size)

        current <- nzd[pos:end]
        current[is.missing(current)] <- placeholder
        H5Dwrite(dhandle, current, h5spaceMem=mem_shandle, h5spaceFile=shandle)
        pos <- end + 1L
    }
}

.choose_itype <- function(ilimit) {
    if (ilimit < 2^16) {
        "H5T_NATIVE_UINT16"
    } else {
        "H5T_NATIVE_UINT32"
    }
}

setMethod("h5_write_sparse_matrix", "CsparseMatrix", function(x, handle, details, ...) {
    if (!is.null(details$placeholder)) {
        .deposit_sparse_data_with_placeholder(handle, x@x, type=details$type, placeholder=details$placeholder)
    } else {
        h5_write_vector(handle, "data", x@x, type=details$type)
    }

    h5_write_vector(handle, "indices", x@i, type=.choose_itype(nrow(x)))
    h5_write_vector(handle, "indptr", x@p, type="H5T_NATIVE_UINT64")
    h5_write_attribute(handle, "layout", "CSC", scalar=TRUE)
})

setMethod("h5_write_sparse_matrix", "RsparseMatrix", function(x, handle, details, ...) {
    if (!is.null(details$placeholder)) {
        .deposit_sparse_data_with_placeholder(handle, x@x, type=details$type, placeholder=details$placeholder)
    } else {
        h5_write_vector(handle, "data", x@x, type=details$type)
    }

    h5_write_vector(handle, "indices", x@j, type=.choose_itype(ncol(x)))
    h5_write_vector(handle, "indptr", x@p, type="H5T_NATIVE_UINT64")
    h5_write_attribute(handle, "layout", "CSR", scalar=TRUE)
})

setMethod("h5_write_sparse_matrix", "SVT_SparseMatrix", function(x, handle, details, ...) {
    N <- details$size
    chunks <- h5_guess_vector_chunks(N)
    dhandle <- h5_create_vector(handle, "data", N, chunks=chunks, type=details$type)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)

    ihandle <- h5_create_vector(handle, "indices", N, chunks=chunks, type=.choose_itype(nrow(x)))
    on.exit(H5Dclose(ihandle), add=TRUE, after=FALSE)

    shandle <- H5Screate_simple(N)
    on.exit(H5Sclose(shandle), add=TRUE, after=FALSE)
    mhandle <- H5Screate_simple(1)
    on.exit(H5Sclose(mhandle), add=TRUE, after=FALSE)

    SVT <- x@SVT
    if (is.null(SVT)) {
        column.counts <- integer(ncol(x))

    } else {
        start <- 1 # don't use integers to avoid overflow
        cached <- 0 # don't use integers to avoid overflow
        last.cleared <- 0L

        tstr <- type(x)
        chunksize <- getAutoBlockLength(tstr)
        none <- as(NULL, tstr)
        one <- as(1, tstr)

        # In versions >= 1L, the indices are the second element of each node.
        version <- x@.svt_version
        val.index <- if (version == 1L) 1L else 2L
        idx.index <- if (version == 1L) 2L else 1L

        column.counts <- integer(length(SVT))
        for (i in seq_along(SVT)) {
            y <- SVT[[i]]
            if (!is.null(y)) {
                column.counts[i] <- length(y[[idx.index]])
            }
        }

        for (i in seq_len(ncol(x))) {
            cached <- cached + column.counts[i]

            if (cached >= chunksize || i == ncol(x)) {
                targets <- SVT[(last.cleared + 1L):i]
                all.i <- all.d <- vector("list", length(targets))
                for (j in seq_along(targets)) {
                    y <- targets[[j]]
                    if (is.null(y)) {
                        all.i[[j]] <- integer(0)
                        all.d[[j]] <- none
                    } else {
                        vals <- y[[val.index]]
                        idxs <- y[[idx.index]]
                        if (is.null(vals)) { # account for lacunar nodes.
                            vals <- rep(one, length(idxs))
                        }
                        all.i[[j]] <- idxs
                        all.d[[j]] <- vals
                    }
                }

                all.i <- unlist(all.i)
                all.d <- unlist(all.d)
                stopifnot(cached == length(all.d))

                if (!is.null(details$placeholder)) {
                    all.d[is.missing(all.d)] <- details$placeholder
                }

                H5Sselect_hyperslab(shandle, "H5S_SELECT_SET", start=start, count=cached)
                H5Sset_extent_simple(mhandle, cached)
                H5Dwrite(dhandle, all.d, h5spaceMem=mhandle, h5spaceFile=shandle)
                H5Dwrite(ihandle, all.i, h5spaceMem=mhandle, h5spaceFile=shandle)

                last.cleared <- i
                start <- start + cached
                cached <- 0
            }
        }
    }

    h5_write_vector(handle, "indptr", c(0, cumsum(column.counts)), type="H5T_NATIVE_UINT64")
    h5_write_attribute(handle, "layout", "CSC", scalar=TRUE)
})

setMethod("h5_write_sparse_matrix", "ANY", function(x, handle, details, ...) {
    cd <- chunkdim(x)
    if (is.null(cd) || !isFALSE(ncol(x) / cd[2] >= nrow(x) / cd[1])) { # i.e., number of chunks per row >= number of chunks per column, indicating it's easier to extract by column.
        layout <- "CSC"
        ilimit <- nrow(x)
        grid <- colAutoGrid(x)
    } else {
        layout <- "CSR"
        ilimit <- ncol(x)
        grid <- rowAutoGrid(x)
    }

    out <- vector("list", length(grid))

    N <- details$size
    chunks <- h5_guess_vector_chunks(N)
    dhandle <- h5_create_vector(handle, "data", N, type=details$type, chunks=chunks)
    on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)

    ihandle <- h5_create_vector(handle, "indices", N, type=.choose_itype(ilimit), chunks=chunks)
    on.exit(H5Dclose(ihandle), add=TRUE, after=FALSE)

    shandle <- H5Screate_simple(N)
    on.exit(H5Sclose(shandle), add=TRUE, after=FALSE)
    mhandle <- H5Screate_simple(1)
    on.exit(H5Sclose(mhandle), add=TRUE, after=FALSE)

    start <- 1 # don't use integers to avoid overflow
    pointers <- list(0)
    for (i in seq_along(grid)) {
        block <- read_block_as_sparse(x, grid[[i]])

        nzdex <- nzwhich(block, arr.ind=TRUE)
        if (layout == "CSC") {
            primary <- nzdex[, 2]
            secondary <- nzdex[, 1]
            ndim <- ncol(block)
        } else {
            primary <- nzdex[, 1]
            secondary <- nzdex[, 2]
            ndim <- nrow(block)
        }
        v <- nzvals(block)

        o <- order(primary, secondary)
        primary <- primary[o]
        secondary <- secondary[o]
        v <- v[o]

        if (!is.null(details$placeholder)) {
            v[is.missing(v)] <- details$placeholder
        }

        H5Sselect_hyperslab(shandle, "H5S_SELECT_SET", start=start, count=length(v))
        H5Sset_extent_simple(mhandle, length(v))
        H5Dwrite(dhandle, v, h5spaceMem=mhandle, h5spaceFile=shandle)
        H5Dwrite(ihandle, secondary - 1L, h5spaceMem=mhandle, h5spaceFile=shandle)
        pointers <- c(pointers, list(tabulate(primary, ndim)))
        start <- start + length(v)
    }

    h5_write_vector(handle, "indptr", cumsum(unlist(pointers)), type="H5T_NATIVE_UINT64")
    h5_write_attribute(handle, "layout", layout, scalar=TRUE)
})

# This checks the ReloadedArray, in particular the saveObject method.
# library(testthat); library(alabaster.matrix); source("test-ReloadedArray.R")

arr <- array(rpois(10000, 10), c(50, 20, 10))
dir <- tempfile()
saveObject(arr, dir)
obj <- readObject(dir)

x <- Matrix::rsparsematrix(1000, 200, 0.1)
dir2 <- tempfile()
saveObject(x, dir2)
obj2 <- readObject(dir2)

test_that("ReloadedArrays work correctly", {
    expect_s4_class(obj, "ReloadedArray")
    expect_identical(BiocGenerics::path(obj), normalizePath(dir))
    expect_identical(dim(obj), dim(arr))
    expect_identical(extract_array(obj, vector("list", 3)), arr)
    expect_identical(as.array(obj), arr)
    expect_false(is_sparse(obj))

    expect_s4_class(obj2, "ReloadedMatrix")
    expect_identical(BiocGenerics::path(obj2), normalizePath(dir2))
    expect_identical(dim(obj2), dim(x))
    expect_identical(as(extract_sparse_array(obj2, vector("list", 2)), "dgCMatrix"), x)
    expect_true(is_sparse(obj2))

    # Coercions work as expected.
    expect_identical(as(obj2, "dgCMatrix"), x)
    expect_identical(as(obj2@seed, "dgCMatrix"), x)
})

test_that("ReloadedArrays store the absolute path", {
    cwd <- getwd()
    setwd(dirname(dir))
    on.exit(setwd(cwd))

    obj <- readObject(basename(dir))
    expect_identical(BiocGenerics::path(obj), normalizePath(dir))
    expect_identical(as.array(obj), arr)
})

test_that("ReloadedArrays save correctly", {
    original <- tempfile()
    saveObject(obj, original, ReloadedArray.reuse.files="none")
    expect_identical(as.array(readObject(original)), arr)

    if (.Platform$OS.type=="unix") { 
        # This test just doesn't seem to work on Windows. Either the symlink
        # doesn't form properly, or the symlink path isn't what we expect.
        tmp <- tempfile()
        saveObject(obj, tmp, ReloadedArray.reuse.files="symlink")
        expect_identical(as.array(readObject(tmp)), arr)
        link.dest <- Sys.readlink(file.path(tmp, "array.h5"))
        expect_true(startsWith(link.dest, "/"))
        expect_identical(normalizePath(link.dest), normalizePath(file.path(dir, "array.h5")))
        expect_identical(as.array(readObject(tmp)), arr)

        # Relative symlinks also work as expected.
        tmp <- tempfile()
        saveObject(obj, tmp, ReloadedArray.reuse.files="relsymlink")
        expect_identical(as.array(readObject(tmp)), arr)
        link.dest <- Sys.readlink(file.path(tmp, "array.h5"))
        expect_true(startsWith(link.dest, ".."))
        expect_identical(normalizePath(file.path(tmp, link.dest)), normalizePath(file.path(dir, "array.h5")))
        expect_identical(as.array(readObject(tmp)), arr)

        # Trying in a more deeply nested target directory.
        tmp0 <- tempfile()
        dir.create(tmp0)
        tmp <- tempfile(tmpdir=tmp0)
        saveObject(obj, tmp, ReloadedArray.reuse.files="relsymlink")
        expect_identical(as.array(readObject(tmp)), arr)
        link.dest <- Sys.readlink(file.path(tmp, "array.h5"))
        expect_true(startsWith(link.dest, "../.."))
        expect_identical(normalizePath(file.path(tmp, link.dest)), normalizePath(file.path(dir, "array.h5")))
        expect_identical(as.array(readObject(tmp)), arr)
    }

    # file.info() doesn't report the inode number so we don't have an easy way
    # to distinguish between hard links and a copy. Oh well.
    tmp <- tempfile()
    saveObject(obj, tmp, ReloadedArray.reuse.files="link")
    expect_identical(as.array(readObject(tmp)), arr)

    tmp <- tempfile()
    saveObject(obj, tmp, ReloadedArray.reuse.files="copy")
    expect_identical(as.array(readObject(tmp)), arr)
})

test_that("We correctly save ReloadedArraySeeds inside DelayedArrays", {
    skip_on_os("windows")
    X <- obj * 10

    temp <- tempfile()
    saveObject(X, temp, ReloadedArray.reuse.files="symlink", DelayedArray.preserve.ops=TRUE)
    link.target <- Sys.readlink(file.path(temp, "seeds", "0", "array.h5"))
    expect_true(file.exists(link.target))

    roundtrip <- readObject(temp)
    expect_identical(as.array(X), as.array(roundtrip))
})

expect_identical_without_names <- function(x, y) {
    if (!is.null(dimnames(x))) {
        if (identical(dimnames(x), vector("list", length(dim(x))))) {
            dimnames(x) <- NULL
        }
    }
    if (!is.null(dimnames(y))) {
        if (identical(dimnames(y), vector("list", length(dim(x))))) {
            dimnames(y) <- NULL
        }
    }
    expect_identical(x, y)
}

library(DelayedArray)
setClass("SuperSeed", slots=c(dim="integer"))
setMethod("type", "SuperSeed", function(x) "integer")
setMethod("dim", "SuperSeed", function(x) x@dim)
setMethod("extract_array", "SuperSeed", function(x, index) {
    nr <- if (is.null(index[[1]])) x@dim[1] else length(index[[1]])
    nc <- if (is.null(index[[2]])) x@dim[2] else length(index[[2]])
    matrix(0L, nr, nc)
})

---
title: Saving arrays to artifacts and back again
author:
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
package: alabaster.matrix
date: "Revised: November 28, 2023"
output:
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Saving and loading arrays}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE}
library(BiocStyle)
self <- Githubpkg("ArtifactDB/alabaster.matrix")
knitr::opts_chunk$set(error=FALSE, warning=FALSE, message=FALSE)
```

# Overview 

The `r self` package implements methods to save matrix-like objects to file artifacts and load them back into R.
Check out the `r Githubpkg("ArtifactDB/alabaster.base")` for more details on the motivation and the **alabaster** framework.

# Quick start

Given an array-like object, we can use `saveObject()` to save it inside a staging directory:

```{r}
library(Matrix)
y <- rsparsematrix(1000, 100, density=0.05)

library(alabaster.matrix)
tmp <- tempfile()
saveObject(y, tmp)

list.files(tmp, recursive=TRUE)
```

We then load it back into our R session with `loadObject()`.
This creates a HDF5-backed S4 array that can be easily coerced into the desired format, e.g., a `dgCMatrix`.

```{r}
roundtrip <- readObject(tmp)
class(roundtrip)
```

This process is supported for all base arrays, `r CRANpkg("Matrix")` objects and `r Biocpkg("DelayedArray")` objects.

# Saving delayed operations

For `DelayedArray`s, we may instead choose to save the delayed operations themselves to file. 
This creates a HDF5 file following the [**chihaya**](https://ltla.github.io/chihaya) format, containing the delayed operations rather than the results of their evaluation.

```{r}
library(DelayedArray)
y <- DelayedArray(rsparsematrix(1000, 100, 0.05))
y <- log1p(abs(y) / 1:100) # adding some delayed ops.

tmp <- tempfile()
saveObject(y, tmp, DelayedArray.preserve.ops=TRUE)

# Inspecting the HDF5 file reveals many delayed operations:
rhdf5::h5ls(file.path(tmp, "array.h5"))

# And indeed, we can recover those same operations.
readObject(tmp)
```

This allows users to avoid evaluation of the operations when saving objects,
which may improve efficiency, e.g., by avoiding loss of sparsity or casting to a larger type.

# Session information {-}

```{r}
sessionInfo()
```

---
title: Saving arrays to artifacts and back again
author:
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
package: alabaster.matrix
date: "Revised: November 28, 2023"
output:
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Saving and loading arrays}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE}
library(BiocStyle)
self <- Githubpkg("ArtifactDB/alabaster.matrix")
knitr::opts_chunk$set(error=FALSE, warning=FALSE, message=FALSE)
```

# Overview 

The `r self` package implements methods to save matrix-like objects to file artifacts and load them back into R.
Check out the `r Githubpkg("ArtifactDB/alabaster.base")` for more details on the motivation and the **alabaster** framework.

# Quick start

Given an array-like object, we can use `saveObject()` to save it inside a staging directory:

```{r}
library(Matrix)
y <- rsparsematrix(1000, 100, density=0.05)

library(alabaster.matrix)
tmp <- tempfile()
saveObject(y, tmp)

list.files(tmp, recursive=TRUE)
```

We then load it back into our R session with `loadObject()`.
This creates a HDF5-backed S4 array that can be easily coerced into the desired format, e.g., a `dgCMatrix`.

```{r}
roundtrip <- readObject(tmp)
class(roundtrip)
```

This process is supported for all base arrays, `r CRANpkg("Matrix")` objects and `r Biocpkg("DelayedArray")` objects.

# Saving delayed operations

🚧🚧🚧 Oops. Yet to migrate the delayed operations to the new `saveObject`/`readObject` world - stay tuned. 🚧🚧🚧

<!--
For `DelayedArray`s, we may instead choose to save the delayed operations themselves to file, using the `r Githubpkg("LTLA/chihaya")` package.
This creates a HDF5 file following the [**chihaya**](https://ltla.github.io/chihaya) format, containing the delayed operations rather than the results of their evaluation.

```{r, eval=FALSE}
library(DelayedArray)
y <- DelayedArray(rsparsematrix(1000, 100, 0.05))
y <- log1p(abs(y) / 1:100) # adding some delayed ops.

preserveDelayedOperations(TRUE)
meta <- stageObject(y, tmp, "delayed")
.writeMetadata(meta, tmp)

meta <- acquireMetadata(tmp, "delayed/delayed.h5")
roundtrip <- loadObject(meta, tmp)
class(roundtrip)
```

However, it is probably best to avoid preserving delayed operations for file-backed `DelayedArray`s if you want the artifacts to be re-usable on different filesystems.
For example, `HDF5Array`s will be saved with a reference to an absolute file path, which will not be portable.
-->

# Session information {-}

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AmalgamatedArray.R
\name{AmalgamatedArray}
\alias{AmalgamatedArray}
\alias{AmalgamatedArray-class}
\alias{AmalgamatedMatrix-class}
\alias{AmalgamatedArraySeed}
\alias{AmalgamatedArraySeed-class}
\alias{DelayedArray,AmalgamatedArraySeed-method}
\alias{matrixClass,AmalgamatedArray-method}
\alias{componentNames}
\alias{extractComponents}
\alias{stageObject,AmalgamatedArray-method}
\alias{coerce,AmalgamatedArray,AmalgamatedMatrix-method}
\alias{coerce,AmalgamatedMatrix,AmalgamatedArray-method}
\title{Amalgamated array class}
\description{
Implements an amalgamated array, equivalent to a delayed combination of DelayedArray objects.
It allows \code{\link{stageObject}} to save a combination of multiple matrices without actually aggregating their data into a single file.
}
\section{Constructors}{

\code{AmalgamatedArraySeed(..., along=1)} accepts any number of named array-like objects and returns a AmalgamatedArraySeed.
Each object corresponds to a block and should be named accordingly; names should be unique and non-empty.
The \code{along} argument specifies the dimension in which matrices should be combined - for matrices, this is 1 for rows, 2 for columns.

\code{AmalgamatedArray(..., along=1)} accepts any number of named array-like objects and returns a AmalgamatedArray.
Alternatively, a single AmalgamatedArraySeed may be provided in \code{...}.
}

\section{Functions}{

\code{componentNames(x)} will return a character vector of names of component arrays in a AmalgamatedArray(Seed) object \code{x}.

\code{extractComponents(x)} will return a named list of array-like objects,
corresponding to the component arrays used to construct the AmalgamatedArray(Seed) object \code{x}.

\code{\link{stageObject}(x, dir, path, child = FALSE)} will save the AmalgamatedArray \code{x} and its components into the \code{path} inside \code{dir}.
Each component array is staged into its own subdirectory inside \code{path}.
}

\section{Comments on usage}{

The AmalgamatedArraySeed is closely related to (and in fact, is a subclass of) the \linkS4class{DelayedAbind} class.
This means that we can leverage many of the \pkg{DelayedArray} methods for handling the delayed bind.
In theory, we could just use a DelayedAbind directly and save it with \pkg{chihaya} in \code{\link{stageObject}} (via \code{\link{preserveDelayedOperations}(TRUE)}).
However, this provides fewer opportunities for tracking and manipulating the samples.
It also saves the per-sample matrices into a single file, which eliminates possibilities for per-file deduplication and linking, e.g., with \code{\link{recycleHdf5Files}(TRUE)}.
}

\examples{
first <- Matrix::rsparsematrix(10, 10, 0.1)
second <- Matrix::rsparsematrix(10, 20, 0.1)
mat <- AmalgamatedArray(list(foo = first, bar = second), along=2)
mat

componentNames(mat)
out <- extractComponents(mat)
lapply(out, dim)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WrapperArraySeed.R
\name{WrapperArraySeed}
\alias{WrapperArraySeed}
\alias{WrapperArraySeed-class}
\alias{WrapperArray-class}
\alias{WrapperArray}
\alias{loadWrapperArray}
\title{DelayedArray wrapper seed}
\description{
Deprecated, use \link[DelayedArray]{DelayedUnaryIsoOp} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/saveArray.R
\name{saveArray}
\alias{saveArray}
\alias{stageObject,array-method}
\alias{stageObject,Matrix-method}
\alias{saveObject,array-method}
\alias{saveObject,denseMatrix-method}
\title{Save a multi-dimensional array to disk}
\usage{
\S4method{saveObject}{array}(
  x,
  path,
  array.dedup.session = NULL,
  array.dedup.action = "link",
  array.character.vls = FALSE,
  ...
)

\S4method{saveObject}{denseMatrix}(
  x,
  path,
  array.dedup.session = NULL,
  array.dedup.action = "link",
  ...
)
}
\arguments{
\item{x}{An integer, numeric, logical or character array.
Alternatively, any of the \linkS4class{denseMatrix} subclasses from the \pkg{Matrix} package.}

\item{path}{String containing the path to a directory in which to save \code{x}.}

\item{array.dedup.session}{A session object created by \code{\link[alabaster.base]{createDedupSession}},
specifying which objects should be deduplicated if the same \code{x} is encountered multiple times.}

\item{array.dedup.action}{String specifying how deduplication should occur, see options for the \code{action=} argument in \code{\link[alabaster.base]{cloneDirectory}}.}

\item{array.character.vls}{Logical scalar indicating whether to save character arrays in the custom variable length string (VLS) array format.
If \code{NULL}, this is determined based on a comparison of the expected storage against a fixed length array.}

\item{...}{Further arguments, currently ignored.}
}
\value{
\code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
}
\description{
Save a high-dimensional array to its on-disk representations.
}
\details{
When \code{saveObject} is called multiple times on the same \code{x}, this method can avoid re-saving the array if a user supplies a deduplication session in \code{array.dedup.session=}.
Instead, the method will link or copy (depending on the choice of \code{array.dedup.action=}) the files produced by the first \code{saveObject} call to the new \code{path=}.
This saves time and reduces disk usage, and is particularly useful when saving complex data structures like a SummarizedExperiment.
For example, if we have a SummarizedExperiment that contains multiple copies of the same array, we can pass a \code{array.dedup.session=} to the \code{saveObject} call on the SummarizedExperiment.
This will instruct the internal \code{saveObject} calls to only write the array to disk once and subsequently make copies or links for all duplicates of that array.
The same approach can be applied to deduplicate seeds in a DelayedArray, see \code{?"\link{storeDelayedObject}"} for details.
}
\examples{
mat <- array(rpois(10000, 10), c(50, 20, 10))
dimnames(mat) <- list(
   paste0("GENE_", seq_len(nrow(mat))),
   letters[1:20],
   NULL
)

dir <- tempfile()
saveObject(mat, dir)
list.files(dir)

}
\seealso{
\code{\link{readArray}}, to read the directory contents back into the R session.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/saveDelayedArray.R
\name{saveDelayedArray}
\alias{saveDelayedArray}
\alias{stageObject,DelayedArray-method}
\alias{stageObject,DelayedMatrix-method}
\alias{saveObject,DelayedArray-method}
\title{Save DelayedArrays to disk}
\usage{
\S4method{saveObject}{DelayedArray}(
  x,
  path,
  DelayedArray.dispatch.pristine = TRUE,
  DelayedArray.preserve.ops = FALSE,
  DelayedArray.force.external = FALSE,
  DelayedArray.store.args = list(),
  array.dedup.session = NULL,
  array.dedup.action = NULL,
  ...
)
}
\arguments{
\item{x}{A \linkS4class{DelayedArray} object.}

\item{path}{String containing a path to a directory in which to save \code{x}.}

\item{DelayedArray.dispatch.pristine}{Logical scalar indicating whether to call the \code{\link{saveObject}} methods of seeds of pristine arrays.}

\item{DelayedArray.preserve.ops}{Logical scalar indicating whether delayed operations should be preserved on-disk.}

\item{DelayedArray.force.external}{Logical scalar indicating to save the seeds of \code{x} as external arrays.
This is passed directly to \code{storeDelayedObject} as the \code{save.external.array=} argument, see \code{?\link{storeDelayedObject}} for details.}

\item{DelayedArray.store.args}{More named arguments to pass to \code{\link{storeDelayedObject}}.}

\item{array.dedup.session, array.dedup.action}{Arguments controlling deduplication of \code{x}, see \code{?"\link{saveObject,array-method}"} for details.
If \code{x} is not a duplicate of an existing object, these arguments will be passed to further methods as described for \code{...}.}

\item{...}{Further arguments passed to \code{storeDelayedObject} as \code{external.save.args}, if the delayed operations are to be preserved;
otherwise, they are passed to \code{\link{saveObject,array-method}} or \code{\link{saveObject,sparseMatrix-method}}.}
}
\value{
\code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
}
\description{
Save \link{DelayedArray} objects to their on-disk representation.
}
\details{
Supplying \code{array.dedup.session=} by itself is only guaranteed to deduplicate \code{x} itself and may not deduplicate its seeds.
Users should combine this with \code{DelayedArray.force.external=TRUE} to force seeds to be saved via \code{saveObject},
which exposes the seeds to the deduplication machinery in their respective \code{saveObject} methods.
Check out \code{?"\link{storeDelayedObject}"} for more details.
}
\examples{
mat <- Matrix::rsparsematrix(100, 200, density=0.2)
rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
dmat <- DelayedArray::DelayedArray(mat) * 1

dir <- tempfile()
saveObject(dmat, dir, delayed.preserve.ops=TRUE)
list.files(dir)

}
\seealso{
\code{\link{storeDelayedObject}}, for the methods to save each delayed operation.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/storeDelayedObject.R
\name{storeDelayedObject}
\alias{storeDelayedObject}
\alias{storeDelayedObject,ConstantArraySeed-method}
\alias{storeDelayedObject,DelayedAbind-method}
\alias{storeDelayedObject,ANY-method}
\alias{storeDelayedObject,DelayedAperm-method}
\alias{storeDelayedObject,DelayedNaryIsoOp-method}
\alias{storeDelayedObject,DelayedSetDimnames-method}
\alias{storeDelayedObject,DelayedSubassign-method}
\alias{storeDelayedObject,DelayedSubset-method}
\alias{storeDelayedObject,DelayedUnaryIsoOpStack-method}
\alias{storeDelayedObject,DelayedUnaryIsoOpWithArgs-method}
\alias{storeDelayedObject,SVT_SparseMatrix-method}
\alias{storeDelayedObject,array-method}
\alias{storeDelayedObject,denseMatrix-method}
\alias{storeDelayedObject,sparseMatrix-method}
\alias{createExternalSeedDedupSession}
\alias{reloadDelayedObject}
\alias{reloadDelayedObjectFunctionRegistry}
\alias{registerReloadDelayedObjectFunction}
\alias{altStoreDelayedObjectFunction}
\alias{altStoreDelayedObject}
\alias{altReloadDelayedObjectFunction}
\alias{altReloadDelayedObject}
\title{Store/reload a DelayedArray}
\usage{
storeDelayedObject(x, handle, name, ...)

reloadDelayedObject(handle, name, version = package_version("1.1"), ...)

reloadDelayedObjectFunctionRegistry(type = c("operation", "array"))

registerReloadDelayedObjectFunction(
  type = c("operation", "array"),
  subtype,
  fun,
  existing = c("old", "new", "error")
)

\S4method{storeDelayedObject}{array}(
  x,
  handle,
  name,
  version = package_version("1.1"),
  save.external.array = FALSE,
  ...
)

\S4method{storeDelayedObject}{ANY}(
  x,
  handle,
  name,
  version = package_version("1.1"),
  external.save.args = list(),
  external.dedup.session = NULL,
  external.dedup.action = NULL,
  ...
)

altStoreDelayedObjectFunction(store)

altStoreDelayedObject(...)

altReloadDelayedObjectFunction(reload)

altReloadDelayedObject(...)
}
\arguments{
\item{x}{Any of the delayed operation/array classes from \pkg{DelayedArray}.}

\item{handle}{An \pkg{rhdf5} handle of a HDF5 file to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).}

\item{name}{String containing the name of the group in \code{file} to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).}

\item{...}{For \code{storeDelayedObject} and \code{reloadDelayedObject}, additional arguments to be passed to specific methods.

For \code{altStoreDelayedObject} and \code{altReloadDelayedObject}, arguments to be passed to the alternative functions.}

\item{version}{Package version of the \pkg{chihaya} format to use when storing or reloading delayed objects.
When reloading, the version should be retrieved from the attributes of the outermost group, typically by \code{readDelayedArray}.}

\item{type}{String specifying the type of delayed object, i.e., operation or array.
This corresponds to \code{delayed_type} type in the \pkg{chihaya} attributes.}

\item{subtype}{String specifying the subtype of the delayed object, 
This corresponds to \code{delayed_array} or \code{delayed_operation} type (depending on \code{type}) in the \pkg{chihaya} attributes.}

\item{fun}{Function to reload a delayed object.
This should accept the same arguments as \code{reloadDelayedObject} and should return a delayed array (if \code{type="array"}) or operation (otherwise).
It may also be \code{NULL} to delete an existing entry in the registry.}

\item{existing}{Logical scalar indicating the action to take if a function has already been registered for \code{type} and \code{subtype} -
keep the old or new function, or throw an error.}

\item{save.external.array}{Logical scalar indicating whether to save an array-like seed as an external seed,
even if a dedicated \code{storeDelayedObject} method is available.}

\item{external.save.args}{Named list of further arguments to pass to \code{\link{altSaveObject}} when saving an external seed.}

\item{external.dedup.session}{Deprecated, set \code{external.save.args$array.dedup.session} instead.}

\item{external.dedup.action}{Deprecated, set \code{external.save.args$array.dedup.action} instead.}

\item{store}{Function (typically a generic) to store delayed objects to file.
This should accept the same arguments as \code{storeDelayedObject}.}

\item{reload}{Function to reload delayed objects from file.
This should accept the same arguments as \code{reloadDelayedObject}.}
}
\value{
For \code{storeDelayedObject} and \code{altStoreDelayedObject}, the contents of \code{x} are saved to \code{file}, and \code{NULL} is invisibly returned.

For \code{reloadDelayedObject} and \code{altReloadDelayedObject}, a delayed operation or \linkS4class{DelayedArray} is returned.

For \code{altStoreDelayedObjectFunction}, the current store function is returned if \code{store} is missing.
Otherwise, \code{store} is set as the current store function and the previous store function is returned.

For \code{altReloadDelayedObjectFunction}, the current reload function is returned if \code{reload} is missing.
Otherwise, \code{reload} is set as the current reload function and the previous reload function is returned.
}
\description{
Store or reload the delayed operations or array-like seeds of a \linkS4class{DelayedArray} in an existing HDF5 file.
}
\section{Customization}{

Developers can easily extend \pkg{alabaster.matrix} to new delayed objects by writing new methods for \code{storeDelayedObject}.
Methods should save the contents of the delayed object to the HDF5 file in the \pkg{chihaya} format.
Each new store method typically requires a corresponding reloading function to be registered via \code{registerReloadDelayedObjectFunction},
so that \code{reloadDelayedObject} knows how to reconstitute the object from file.

Application developers can customize the process of storing/reloading delayed objects
by specifying alternative functions in \code{altReloadDelayedObjectFunction} and \code{altStoreDelayedObjectFunction}.
For example, if we want to preserve all delayed objects except for \linkS4class{DelayedSubset},
we could replace \code{storeDelayedObject} with an \code{altStoreDelayedObject} that realizes any DelayedSubset instance into an ordinary matrix.
This is analogous to the overrides for \code{\link{altReadObject}} and \code{\link{altSaveObject}}.

Extension developers (i.e., those who write new methods for \code{storeDelayedObject} or new functions for \code{reloadDelayedObject})
should generally use \code{altStoreDelayedObject} and \code{altReloadDelayedObject} in their method/funcion bodies.
This ensures that any custom overrides specified by application developers are still respected in the extensions to \pkg{alabaster.matrix}.
}

\section{External seeds}{

Whenever \code{\link{storeDelayedObject}} encounters a delayed operation or array-like seed for which it has no methods,
the ANY method will save the delayed object as an \dQuote{external seed}.
The array is saved via \code{\link{altSaveObject}} into a \code{seeds} directory next to the file associated with \code{handle}.
A reference to this external location is then stored in the \code{name} group inside \code{handle}.

Users can force this behavior for all array-like seeds by specifying \code{save.external.array=TRUE}.
This instructs \code{storeDelayedObject} to save everything as external seeds, including those arrays for which it has methods.
Doing so can be beneficial to enable deduplication, e.g., when two delayed arrays perform different operations on the same underlying seed.
By saving the seeds externally, file management systems can identify the redundancy to save storage space.

Advanced users can explicitly deduplicate external seeds by setting both \code{save.external.array=TRUE} and supplying \code{array.dedup.session=} in \code{external.save.args=}.
The \code{array.dedup.session} object is filled up with unique seeds as \code{storeDelayedObject} is called on various DelayedArrays (see \code{?"\link{saveObject,array-method}"} for details).
Whenever a duplicate seed is encountered, it is not saved again, but is instead linked or copied from the file path associated with the identical external seed.
For example, a new session can be created when saving a SummarizedExperiment to deduplicate seeds across its assays.

When external seeds are encountered by \code{reloadDelayedObject}, they are loaded as \linkS4class{ReloadedArray}s (or some variant thereof) by \code{\link{altReadObject}}.
Users can forcibly realize the reloaded seed into memory by passing \code{custom.takane.reload=TRUE} in \code{...} for the \code{reloadDelayedObject} call.
This is occasionally helpful for providing a more faithful roundtrip from file back into memory.
}

\examples{
library(DelayedArray)
X <- DelayedArray(matrix(runif(100), ncol=20))
Y <- cbind(X, DelayedArray::ConstantArray(value=50, c(5, 10)))

library(rhdf5)
temp <- tempfile()
dir.create(temp)

fpath <- file.path(temp, "foo.h5")
fhandle <- H5Fcreate(fpath)
storeDelayedObject(Y@seed, fhandle, "YAY")
rhdf5::h5ls(fhandle)
H5Fclose(fhandle)

fhandle <- H5Fopen(fpath, "H5F_ACC_RDONLY")
reloadDelayedObject(fhandle, "YAY")
H5Fclose(fhandle)

}
\seealso{
\code{\link{saveObject,DelayedArray-method}} and \code{\link{readDelayedArray}}, where these methods are used.

\url{https://artifactdb.github.io/chihaya/}, for the file format specification of delayed objects.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DelayedMask.R
\name{DelayedMask}
\alias{DelayedMask}
\alias{DelayedMask-class}
\alias{dim,DelayedMask-method}
\alias{dimnames,DelayedMask-method}
\alias{chunkdim,DelayedMask-method}
\alias{path,DelayedMask-method}
\alias{is_sparse,DelayedMask-method}
\alias{extract_array,DelayedMask-method}
\alias{extract_sparse_array,DelayedMask-method}
\title{Delayed masking}
\usage{
DelayedMask(x, placeholder)
}
\arguments{
\item{x}{An existing \pkg{DelayedArray} seed.}

\item{placeholder}{Placeholder value to replace with \code{NA}.
This should be of the same type as \code{\link{type}(x)}.}
}
\value{
A DelayedMask object, to be wrapped in a \code{\link{DelayedArray}}.
}
\description{
Delayed masking of missing values, based on replacement of placeholder values with \code{NA}.
This allows missingness to be encoded in frameworks without the same concept of \code{NA} as R.
}
\details{
If \code{\link{is.na}(placeholder)} is true for double-precision \code{x}, masking is performed for all values of \code{x} where \code{is.na} is true.
This includes both NaNs and NAs; no attempt is made to distinguish between the NaN payloads.

Currently, an error is raised for any integer \code{x} that produces non-missing values of -2^31 without a placeholder of \code{NA_integer_}.
This is because R cannot distinguish the integer -2^31 from an integer-type NA.
}
\examples{
original <- DelayedArray(matrix(rpois(40, lambda=2), ncol=5))
original
masked <- DelayedMask(original, 0)
DelayedArray(masked)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeSparseMatrix.R
\name{writeSparseMatrix}
\alias{writeSparseMatrix}
\title{Write a sparse matrix}
\usage{
writeSparseMatrix(
  x,
  file,
  name,
  chunk = 10000,
  column = TRUE,
  tenx = FALSE,
  guess.integer = TRUE
)
}
\arguments{
\item{x}{A sparse matrix of some sort.
This includes sparse \linkS4class{DelayedMatrix} objects.}

\item{file}{String containing a path to the HDF5 file.
The file is created if it is not already present.}

\item{name}{String containing the name of the group to store \code{x}.}

\item{chunk}{Integer scalar specifying the chunk size for the indices and values.}

\item{column}{Logical scalar indicating whether to store as compressed sparse column format.}

\item{tenx}{Logical scalar indicating whether to use the 10X compressed sparse column format.}

\item{guess.integer}{Logical scalar specifying whether to guess an appropriate integer type from \code{x}.}
}
\value{
A \code{NULL} invisibly.
The contents of \code{x} are written to \code{name} in \code{file}.
}
\description{
Writes a sparse matrix to file in a compressed sparse format.
}
\details{
This writes a sparse matrix to file in various formats:
\itemize{
\item \code{column=TRUE} and \code{tenx=FALSE} uses H5AD's \code{csr_matrix} format.
\item \code{column=FALSE} and \code{tenx=FALSE} uses H5AD's \code{csc_matrix} format.
\item \code{tenx=TRUE} uses 10X Genomics' HDF5 matrix format.
}
For the first two formats, the apparent transposition is deliberate, because columns in R are interpreted as rows in H5AD.
This allows us to retain consistency the interpretation of samples (columns in R, rows in H5AD) and features (vice versa).
Constructors for classes like \linkS4class{H5SparseMatrix} will automatically transpose so no extra work is required.

If \code{guess.integer=TRUE}, we attempt to save \code{x}'s values into the smallest type that will accommodate all of its values. 
If \code{x} only contains unsigned integers, we will attempt to save either 8-, 16- or 32-bit unsigned integers.
If \code{x} contains signed integers, we will fall back to 32-bit signed integers.
For all other values, we will fall back to double-precision floating point values.

We attempt to save \code{x}'s indices to unsigned 16-bit integers if the relevant dimension of \code{x} is small enough.
Otherwise we will save it as an unsigned 32-bit integer.
}
\examples{
library(Matrix)
x <- rsparsematrix(100, 20, 0.5)
tmp <- tempfile(fileext=".h5")
writeSparseMatrix(x, tmp, "csc_matrix")
writeSparseMatrix(x, tmp, "csr_matrix", column=FALSE)
writeSparseMatrix(x, tmp, "tenx_matrix", tenx = TRUE)

rhdf5::h5ls(tmp)
library(HDF5Array)
H5SparseMatrix(tmp, "csc_matrix")
H5SparseMatrix(tmp, "csr_matrix")
H5SparseMatrix(tmp, "tenx_matrix")
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createRawArraySeed.R
\name{createRawArraySeed}
\alias{createRawArraySeed}
\alias{.createRawArraySeed}
\alias{.extractArrayDimnames}
\title{Array loading utilities}
\usage{
.createRawArraySeed(info, project, names = TRUE)

.extractArrayDimnames(path, group, ndim)
}
\arguments{
\item{info}{A named list of metadata for this array.}

\item{project}{Any argument accepted by the acquisition functions, see \code{?\link{acquireFile}}. 
By default, this should be a string containing the path to a staging directory.}

\item{names}{Logical scalar indicating whether the seed should be annotated with dimnames (if available).}

\item{path}{String containing the path to the file containing said array.}

\item{group}{String containing the name of the group with the dimnames.}

\item{ndim}{Integer scalar specifying the number of dimensions.}
}
\value{
\code{.createRawArraySeed} returns a seed that can be used in the \code{\link{DelayedArray}} constructor.

\code{.extractArrayDimnames} returns a list of character vectors or \code{NULL}, containing the dimnames.
}
\description{
Utilities for loading an array saved by \code{\link{stageObject}}.
}
\details{
For \code{.createArraySeed}, the array should be one of:
\itemize{
\item \code{hdf5_dense_array}
\item \code{hdf5_sparse_matrix}
\item \code{hdf5_delayed_array}
\item \code{amalgamated_array}
}

For delayed arrays, the file may contain a seed array with the \code{"custom alabaster local array"} type.
This should have a \code{path} dataset containing a relative path to another array in the same \code{project}, which is loaded and used as the seed for this delayed array.
Callers can overwrite this behavior by setting \code{"custom alabaster local array"} in the \code{knownArrays} from \pkg{chihaya} before calling \code{.createRawArraySeed}.

For \code{.extractArrayDimnames}, \code{path} is expected to be a HDF5 file with a group specified by \code{group}.
Each child of this group is a string dataset named after a (0-indexed) dimension, containing the names for that dimension.
}
\examples{
# Staging an array as an example:
dir <- tempfile()
dir.create(dir)
mat <- array(rpois(10000, 10), c(50, 20, 10))
meta <- stageObject(mat, dir, "whee")

# Loading it back as a DelayedArray seed:
.createRawArraySeed(meta, project=dir)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readArray.R
\name{readArray}
\alias{readArray}
\alias{loadArray}
\title{Read a dense array from disk}
\usage{
readArray(path, metadata, ...)
}
\arguments{
\item{path}{String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a dense array.}

\item{metadata}{Named list of metadata for this object, see \code{\link{readObject}} for more details.}

\item{...}{Further arguments, ignored.}
}
\value{
A dense file-backed \linkS4class{ReloadedArray}.
}
\description{
Read a dense high-dimensional array from its on-disk representation.
This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
}
\examples{
arr <- array(rpois(10000, 10), c(50, 20, 10))
dimnames(arr) <- list(
   paste0("GENE_", seq_len(nrow(arr))),
   letters[1:20],
   NULL
)

dir <- tempfile()
saveObject(arr, dir)
readObject(dir)

}
\seealso{
\code{"\link{saveObject,array-method}"}, to create the directory and its contents.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ReloadedArray.R
\name{ReloadedArraySeed}
\alias{ReloadedArraySeed}
\alias{ReloadedArraySeed-class}
\alias{ReloadedArray-class}
\alias{ReloadedMatrix-class}
\alias{DelayedArray,ReloadedArraySeed-method}
\alias{matrixClass,ReloadedArray-method}
\alias{coerce,ReloadedArray,ReloadedMatrix-method}
\alias{coerce,ReloadedMatrix,ReloadedArray-method}
\alias{path,ReloadedArraySeed-method}
\alias{saveObject,ReloadedArray-method}
\alias{ReloadedArray}
\title{Reloaded \pkg{alabaster} array}
\usage{
ReloadedArraySeed(path, seed = NULL, ...)

ReloadedArray(path, seed = NULL, ...)
}
\arguments{
\item{path}{String containing a path to the directory with the on-disk array representation.
Alternatively an existing ReloadedArraySeed, which is returned without modification.}

\item{seed}{Contents of the loaded array, e.g., as an ordinary R array, a \linkS4class{DelayedArray} or a sparse matrix.
If \code{NULL}, this is obtained by calling \code{\link{readObject}}.}

\item{...}{Further arguments to pass to \code{\link{readObject}} when \code{seed=NULL}.}
}
\value{
For the constructors, an instance of the \linkS4class{ReloadedArraySeed} or \linkS4class{ReloadedArray}.
}
\description{
An array that was reloaded from disk by the \code{\link{readObject}} function.
This allows methods to refer to the existing on-disk representation by inspecting the \code{path}.
For example, \code{\link{saveObject}} can just copy/link to the files instead of repeating the saving process.
}
\details{
The ReloadedArraySeed is a \linkS4class{DelayedUnaryIsoOp} subclass that will just forward all operations to the underlying \code{seed}.
Its main purpose is to track the \code{path} that was originally used to generate \code{seed}, which enables optimizations for methods that need to operate on the files.

One obvious optimization is the specialization of \code{\link{saveObject}} on ReloadedArray instances.
Instead of loading the array data back into the R session and saving it again, the \code{saveObject} method can just link or copy the existing files.
This behavior is controlled by the \code{ReloadedArray.reuse.files=} option in the \code{saveObject} method,
which can be any of the choices for \code{action=} in \code{\link[alabaster.base]{cloneDirectory}}.
It may also be \code{"none"} to ignore existing files and just save the contents by calling \code{"\link{saveObject,DelayedArray-method}"}.
}
\examples{
arr <- array(rpois(10000, 10), c(50, 20, 10))
dir <- tempfile()
saveObject(arr, dir)
obj <- readArray(dir)
obj
DelayedArray::showtree(obj)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readSparseMatrix.R
\name{readSparseMatrix}
\alias{readSparseMatrix}
\title{Read a sparse matrix from disk}
\usage{
readSparseMatrix(path, metadata, ...)
}
\arguments{
\item{path}{String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a spars matrix.}

\item{metadata}{Named list of metadata for this object, see \code{\link{readObject}} for more details.}

\item{...}{Further arguments, ignored.}
}
\value{
A sparse \linkS4class{ReloadedMatrix} object.
}
\description{
Read a sparse matrix from its on-disk representation.
This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
}
\examples{
mat <- Matrix::rsparsematrix(100, 200, density=0.2)
rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))
dir <- tempfile()
saveObject(mat, dir)
readObject(dir)

}
\seealso{
\code{"\link{saveObject,sparseMatrix-method}"}, to create the directory and its contents.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{recycleHdf5Files}
\alias{recycleHdf5Files}
\title{Recycle existing HDF5 files}
\usage{
recycleHdf5Files(recycle)
}
\arguments{
\item{recycle}{Whether to recycle existing files for HDF5-backed DelayedArrays.}
}
\value{
Logical scalar indicating whether HDF5 files are to be reused.
If \code{recycle} is supplied, it is used to set this scalar, and the \emph{previous} value of the scalar is invisibly returned.
}
\description{
Re-use existing files in HDF5-backed arrays rather than reserializing them in \code{\link{stageObject}}.
}
\details{
If this options is enabled, \code{stageObject} will attempt to link/copy existing files for any HDF5-backed DelayedArray instances -
most specifically, \linkS4class{HDF5Array} objects and \linkS4class{H5SparseMatrix} objects using the 10X format.
This avoids re-serialization of the data for faster staging.
It also allows advanced users to add their own customizations into the HDF5 file during staging, as long as they do not interfere with \code{\link{loadArray}}.

By default, this option is disabled as the properties of the existing file are not known in the general case.
In particular, the file might contain other groups/datasets that are irrelevant, and use up extra disk space if copied; or confidential, and should not be stored in the staging directory.
Users should only enable this option if they have full control over the generation and contents of the backing HDF5 files.

Also note that any dimnames on \code{x} will be ignored during recycling.
}
\examples{
recycleHdf5Files()
old <- recycleHdf5Files(TRUE)
recycleHdf5Files()
recycleHdf5Files(old)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{preserveDelayedOperations}
\alias{preserveDelayedOperations}
\title{Preserve delayed operations during staging}
\usage{
preserveDelayedOperations(preserve)
}
\arguments{
\item{preserve}{Whether to preserve delayed operations using the \pkg{chihaya} specification.}
}
\value{
Logical scalar indicating whether delayed operations are to be preserved by the DelayedArray method.
If \code{preserve} is supplied, it is used to set this scalar, and the \emph{previous} value of the scalar is invisibly returned.
}
\description{
Preserve delayed operations via \pkg{chihaya} when staging a \linkS4class{DelayedArray} with \code{\link{stageObject}}.
}
\details{
By default, any DelayedArray in \code{\link{stageObject}} will be saved as a new dense array or sparse matrix.
However, if this option is enabled, DelayedArrays will instead be saved in the \pkg{chihaya} specification,
where the delayed operations are themselves stored in the HDF5 file (see \url{https://artifactdb.github.io/chihaya/} for details).

The \pkg{chihaya} specification is more complicated to parse but can be helpful in reducing disk usage.
One simple example is to avoid sparsity-breaking or integer-to-float operations by storing their delayed representations in the file.
If the seed matrix is derived from some immutable reference location, advanced users can even store links to that location instead of duplicating the seed data.
}
\examples{
preserveDelayedOperations()
old <- preserveDelayedOperations(TRUE)
preserveDelayedOperations()
preserveDelayedOperations(old)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readDelayedArray.R
\name{readDelayedArray}
\alias{readDelayedArray}
\title{Read a delayed array from disk}
\usage{
readDelayedArray(path, metadata, delayed_array.reload.args = list(), ...)
}
\arguments{
\item{path}{String containing a path to a directory, itself created by the \code{\link{saveObject}} method for a delayed array.}

\item{metadata}{Named list of metadata for this object, see \code{\link{readObject}} for more details.}

\item{delayed_array.reload.args}{Named list of arguments to be passed to \code{\link{reloadDelayedObject}}.}

\item{...}{Further arguments, ignored.}
}
\value{
A multi-dimensional array-like object.
}
\description{
Read a delayed high-dimensional array from its on-disk representation.
This is usually not directly called by users, but is instead called by dispatch in \code{\link{readObject}}.
}
\examples{
arr <- array(rpois(10000, 10), c(50, 20, 10))
dimnames(arr) <- list(
   paste0("GENE_", seq_len(nrow(arr))),
   letters[1:20],
   NULL
)

dir <- tempfile()
saveObject(arr, dir)
readObject(dir)

}
\seealso{
\code{"\link{saveObject,DelayedArray-method}"}, to create the directory and its contents.

\code{\link{reloadDelayedObject}}, for the methods to reload each delayed operation.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/saveSparseMatrix.R
\name{saveSparseMatrix}
\alias{saveSparseMatrix}
\alias{saveObject,sparseMatrix-method}
\alias{saveObject,SVT_SparseMatrix-method}
\title{Save a sparse matrix to disk}
\usage{
\S4method{saveObject}{sparseMatrix}(
  x,
  path,
  array.dedup.session = NULL,
  array.dedup.action = "link",
  ...
)

\S4method{saveObject}{SVT_SparseMatrix}(
  x,
  path,
  array.dedup.session = NULL,
  array.dedup.action = "link",
  ...
)
}
\arguments{
\item{x}{A sparse matrix of some kind, typically from either the \pkg{Matrix} or \pkg{SparseArray} packages.}

\item{path}{String containing the path to a directory in which to save \code{x}.}

\item{array.dedup.session, array.dedup.action}{Arguments controlling deduplication of \code{x}, see \code{?"\link{saveObject,array-method}"} for details.}

\item{...}{Further arguments, currently ignored.}
}
\value{
\code{x} is saved to \code{path} and \code{NULL} is invisibly returned.
}
\description{
Save a sparse matrix to its on-disk representations.
}
\examples{
mat <- Matrix::rsparsematrix(100, 200, density=0.2)
rownames(mat) <- paste0("GENE_", seq_len(nrow(mat)))

dir <- tempfile()
saveObject(mat, dir)
list.files(dir)

}
\seealso{
\code{\link{readSparseMatrix}}, to read the directory contents back into the R session.
}
\author{
Aaron Lun
}
